/* This program is part of a correctly-rounded, division-free l2
   vector norm with avoidance of spurious underflow and overflow.

   Author: Nicolas Louvet,

           Université Claude Bernard Lyon 1

   implementing the algorithm described in the paper

   Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
   Laurence Rideau. Accurate calculation of Euclidean Norms using
   Double-word arithmetic. 2021

   This program is

   Copyright (C) 2021 Université Claude Bernard Lyon 1

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <http://www.gnu.org/licenses/>.

*/

#include <stdlib.h>
#include <stdio.h>
#include "norms.h"
#include "randomvectors.h"

void safeFree(void *ptr) {
  free(ptr);
}

void *safeCalloc(size_t nmemb, size_t size) {
  void *ptr;

  ptr = calloc(nmemb, size);
  if (ptr == NULL) {
    fprintf(stderr, "Could not allocate enough memory.\n");
    exit(2);
  }

  return ptr;
}

double log2(double);

int testNorm(double (*norm)(const double *, unsigned int), char *normName, unsigned int n, int k, int displayErrors, int testCase) {
    int overallFaithfulResults, underflowCases, overflowCases;
    int i;
    double *x, y;
    double ratioFaithful, ratioUnderflow, ratioOverflow;
    mpfr_t eps, ulpErr;
    double epsDouble, ulpErrDouble;
    char testCaseName[25];

    getNameOfTestCase(testCaseName, testCase);

    x = malloc(n*sizeof(double));
    mpfr_init2(eps, 53);
    mpfr_set_si(eps, 0, GMP_RNDN);
    mpfr_init2(ulpErr, 53);
    mpfr_set_si(ulpErr, 0, GMP_RNDN);

    overallFaithfulResults = 0;
    underflowCases = 0;
    overflowCases = 0;

    for (i=0; i<k; i++) {
        makeRandomVector(x, n, testCase);
        y = norm(x, n);
        if (isFaithfulNorm(y, x, n, displayErrors, &underflowCases, &overflowCases, eps, ulpErr))
            overallFaithfulResults ++;
    }

    ratioFaithful = ((double) overallFaithfulResults) / ((double) k);
    ratioUnderflow = ((double) underflowCases) / ((double) k);
    ratioOverflow = ((double) overflowCases) / ((double) k);
    epsDouble = mpfr_get_d(eps, GMP_RNDN);
    ulpErrDouble = mpfr_get_d(ulpErr, GMP_RNDN);

    printf("%15s, ", normName);
    printf("%20s, ", testCaseName);
    printf("udr=%.2f%%, ", ratioUnderflow * 100.0);
    printf("ovr=%.2f%%, ", ratioOverflow * 100.0);
    printf("fth=%.2f%%, ", ratioFaithful * 100.0);
    printf("maxrel=%1.5e, ", epsDouble);
    printf("maxulp=%1.5e\n", ulpErrDouble);

    free(x);
    mpfr_clear(eps);
    mpfr_clear(ulpErr);

    return (overallFaithfulResults == k);
}

int doAllTestsNorm(double (*norm)(const double *, unsigned int), char *normName, unsigned int n, int k, int displayErrors) {
    int i, res, okay = 1;
    static int numTestCases = 9;
    static int testCases[] = { NO_OVERFLOW,
                               FULL_EXPONENT_RANGE,
                               AROUND_ONE,
                               REALLY_SMALL,
                               NORM_GRADUALLY_UNDERFLOWS,
                               HALF_ULP_CASES,
                               MONOTONE_INCREASING,
                               MONOTONE_DECREASING,
                               NETLIB_SPURIOUS_UNDERFLOW
                             };
    for(i = 0; i < numTestCases; i++) {
      res = testNorm(norm, normName, n, k, displayErrors, testCases[i]);
      okay = okay && res;
    }

    return okay;
}

void printUsage(char *str) {
  fprintf(stderr, "Usage: %s <vector length> <number of tests>\n",str);
  fprintf(stderr, "\n");
  fprintf(stderr, "Beware, for some tests and some functions tested,\n");
  fprintf(stderr, "<vector length> must be a multiple of the length\n");
  fprintf(stderr, "of the simd vectors used in the implementation\n");
  fprintf(stderr, "(SSE2 2, AVX2 4, AVX512 8,...\n");
  fprintf(stderr, "\n");
}

int readArg(int *r, char *str, int min) {
  int res;
  char *end;

  if (str[0] == '\0') return 0;
  res = strtol(str, &end, 10);
  if (end[0] != '\0') return 0;

  if (res < min) return 0;

  *r = res;
  return 1;
}

int main(int argc, char **argv) {
  int sn, k;
  unsigned int n;

  if (argc < 3) {
    printUsage(argv[0]);
    return 2;
  }

  if (!readArg(&sn, argv[1], 1)) {
    printUsage(argv[0]);
    return 2;
  }
  n = sn;

  if (!readArg(&k, argv[2], 1)) {
    printUsage(argv[0]);
    return 2;
  }

  if(!doAllTestsNorm(accurateNorm, "accurate norm", n, k, 1)) {
    printf("WARNING: ERRORS FOUND WHILE TESTING accurateNorm!!!\n");
    return -1;
  }
  else
    printf("ALL TESTS WITH accurateNorm PASSED...\n");

  if(!doAllTestsNorm(faithfulNorm, "faithful norm", n, k, 1)) {
    printf(" WARNING: ERRORS FOUND WHILE TESTING accurateNorm!!!\n");
    return -1;
  }
  else
    printf("ALL TESTS WITH faithfulNorm PASSED...\n");

  doAllTestsNorm(naiveNorm, "naive norm", n, k, 0);
  doAllTestsNorm(vecNaiveNorm, "vec naive", n, k, 0);
  //testNorm(vecNaiveNorm, "vecNaiveNorm", n, k, 0, AROUND_ONE);
  //testNorm(vecScalNaiveNorm, "vecScalNaiveNorm", n, k, 0, AROUND_ONE);
  return 0;
}



