## Data entries.
##
## This is column names as found in CSV files.

#datadir = "data"
datadir = "datarev1"

if (datadir == "data") {
    entries = c("naive", "netlib", "faithful", "accurate")

} else if (datadir == "datarev1") {
    entries = c("naive", "vecnaive", "netlib", "faithful", "accurate")

}


## Load a CSV file.

myreadtable <- function(host, case, n) {
    filename = sprintf("%s_%s_%d.csv", host, case, n)
    pathname = file.path(datadir, filename)
    read.table(c(pathname),
               header=TRUE,
               sep=",")
}


## Smart limit finder.
##
## 1. AVOID EXTRANEOUS VALUES.
##
##    `minTol` and `maxTol` is positions in histogram to take limit values:
##
##         minTol=0.0 means take the minimum of all values
##         minTol=0.1 means take the minimum below 90% of all values
##
##    Use minTol=0.0 and maxTol=1.0 by default. If you have trouble with
##    extraneous values, increate minTol or decrease maxTol a little.
## 

smartlim <- function(T, step, minTol=0.0, maxTol=1.0, mar=1, digits=0) {

    ## Find min and max in data, trying to avoid extraneous values.

    mi = Inf
    ma = -Inf
    for (name in entries) {
        mi = min(mi, quantile(T[[name]], probs=minTol))
        ma = max(ma, quantile(T[[name]], probs=maxTol))
    }

    ## Now we increase limits in order to get nice labels, e.g. "5.1" instead
    ## of "5.0942678324…".

    ## First we add some maring
    mi = mi - mar * step
    ma = ma + mar * step

    limi = floor(mi / step) * step
    lima = ceiling(ma / step) * step

    c(limi, lima)
}


## Histogram creation.

myhist <- function(data, xbreaks) {
    h <- hist(data, plot=FALSE, breaks=xbreaks)
    h$density <- h$counts / sum(h$counts) * 100
    h
}

## Graph creation.

myplot <- function(
        host, case, n,   ## Parameter of the filename.
        time_unit=1,     ## Scale data by this factor.
        hist_step=100,   ## Desired step size.
        ymax=100,        ## Top bound of the printed area.
        xticks=NA,       ## Ticks.
        colors, density, angle, algonames, forpaper=FALSE) {

    ## Read data into table.
    T <- myreadtable(host, case, n)

    T = T / time_unit

    ## Build histogram data.
    histlims = smartlim(T, hist_step, minTol=0.0, maxTol=1.0, mar=0)
    cat("Histogram limits: from", histlims[1], "to", histlims[2], "by", hist_step, fill=TRUE)

    histbreaks = seq(from=histlims[1], to=histlims[2], by=hist_step)
    cat("Number of histogram breaks:", length(histbreaks), fill=TRUE)

    if (typeof(xticks) == "logical" && is.na(xticks)) {
        ## Find min and max values.
        plotxnum = 10
        plotxlim = smartlim(T, plotxnum, minTol=0.02, maxTol=0.98)
        xticks <- seq(from=plotxlim[1], to=plotxlim[2],
                        by=round((plotxlim[2] - plotxlim[1]) / 6, digits=0))
    } else {
        plotxlim = c(xticks[1], xticks[length(xticks)])
    }
    cat("Plot ticks at:", xticks, fill=TRUE)

    ## Create histograms and scale them.
    h = list()
    for (name in entries) { h[[name]] <- myhist(T[[name]], histbreaks) }

    ## Legend text.
    l = list()
    for (name in entries) {
        if (forpaper) {
            l[name] <- sprintf("%s", algonames[[name]])
        } else {
            l[name] <- sprintf("%s (mean = %.1f, sd = %1.f)", algonames[[name]], mean(T[[name]]), sd(T[[name]]))
        }
    }

    ## Margins.
    if (forpaper) {
        par(mar=c(4,4,1,1))
    }

    ## Create plots.
    plot(NULL, xlim=plotxlim, ylim=c(0, ymax), xlab="microseconds", ylab="frequency in %")
    #axis(1, at=xticks)
    for (name in entries) {
        plot(h[[name]],
             freq = FALSE,
             density = density[[name]],
             angle = angle[[name]],
             col = colors[[name]],
             border = (if (forpaper) NULL else colors[[name]]),
             add = TRUE)
    }
    legend("topright", unlist(l), fill=colors, border=colors, density=density, angle=angle)

    ## Title.
    if (!forpaper) {
        title(sprintf("%s, %s, n = %d", host, case, n))
    }

}


## Command-line arguments.

parse_args <- function() {

    args = commandArgs(trailingOnly=TRUE)

    if (length(args) <= 1) {
        stop("usage: Rscript --vanilla hist.R <host> <n>")
    }

    list(host=args[1], n=strtoi(args[2]))
}
