/* This program is part of a faithfully-rounded, division-free l2
   vector norm with avoidance of spurious underflow and overflow.
   
   Author: Christoph Lauter,

           Université Pierre et Marie Curie Paris 6, UPMC, LIP6,
	   PEQUAN team.

   implementing the algorithm described in the paper

   Graillat, Lauter, Tang, Yamanaka and Oishi: Efficient calculations
   of faithfully rounded l2-norms of n-vectors.

   This program is

   Copyright (C) 2014 Université Pierre et Marie Curie Paris 6, UPMC,
   LIP6, PEQUAN team.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <http://www.gnu.org/licenses/>. 

*/

#include <stdint.h>
#include "commons.h"
#include "vec.h"

/* A non-SSE twoAdd
   
   Computes u + v = a + b exactly, where u = RN(a + b) 
   
*/
STATIC INLINE void twoAdd(double * RESTRICT u, double * RESTRICT v, double a, double b) {
  double u1, u2, u3, u4, s, r;

  s = a + b;
  u1 = s - a;
  u2 = s - u1;
  u3 = b - u1;
  u4 = a - u2;
  r = u4 + u3;

  *u = s;
  *v = r;
}

/* A non-SSE fastTwoAdd 

   Computes u + v = a + b exactly, where u = RN(a + b)

   UNDER THE CONDITION THAT abs(a) >= abs(b)

*/
STATIC INLINE void fastTwoAdd(double * RESTRICT u, double * RESTRICT v, double a, double b) {
  double z, s, r;

  s = a + b;
  z = s - a;
  r = b - z;

  *u = s;
  *v = r;
}

/* A non-SSE double-double add 

   Computes u + v approximating ah + al + bh + bl

*/
STATIC INLINE void doubleDoubleAdd(double * RESTRICT u, double * RESTRICT v, 
                                   double ah, double al, double bh, double bl) {
  double th, tl, r;

  twoAdd(&th, &tl, ah, bh);
  r = tl + (al + bl);
  fastTwoAdd(u, v, th, r);
}

/* A non-SSE twoSquare 

   Computes u + v = a^2 exactly, where u = RN(a^2) 

*/
STATIC INLINE void twoSquare(double * RESTRICT u, double * RESTRICT v, double a) {
  STATIC CONST double c = 134217729.0;
  double ap, a1, a2, p, r;

  ap = a*c;
  a1 = (a-ap)+ap;
  a2 = a-a1;

  p = a*a;
  r = (((a1*a1-p)+(a1*a2))+(a2*a1))+(a2*a2);

  *u = p;
  *v = r;
}

STATIC INLINE double scaledSquareRoot(int m, double S) {
  STATIC CONST double sqrtGammaToM[] = { 5.26013590154837350724098988288012866555033980282317e210, /* gamma^-2/2 = (2^-350)^-2 */
                                         2.29349861599007151161082089530208694079656498916828e105, /* gamma^-1/2 = (2^-350)^-1 */
                                         1.0,                                                      /* gamma^0/2  = (2^-350)^0  */
                                         4.3601508761683463371878950054372442606045150559341e-106, /* gamma^1/2  = (2^-350)^1  */
                                         1.9010915662951598235150724058351031092648712063735e-211  /* gamma^2/2  = (2^-350)^2  */ };
  double Z, y;

  Z = sqrt(S);
  y = sqrtGammaToM[m - (-2)] * Z;

  return y;
}

STATIC INLINE double faithfulFinal(int k, double U, double u, double V, double v) {
  STATIC CONST double bound1    = 3.08469742733169167070816004443201143863168634177267e-179; /* beta_lo^2 / eps^3 = 2^-593 */
  STATIC CONST double bound2    = 1.4396524142538228424993723224595141948383030778566e163;   /* beta_hi^2 * eps^2 = 2^542  */
  STATIC CONST double gamma     = 1.9010915662951598235150724058351031092648712063735e-211;  /* gamma             = 2^-700 */
  STATIC CONST double gammaToM1 = 5.26013590154837350724098988288012866555033980282317e210;  /* gamma^-1          = 2^700  */
  int m;
  double S, s;
  
  if ((k == 2) || ((U >= bound1) || (V <= bound2))) /* First two branches of the article pseudo-code */
    return scaledSquareRoot(k,U); 
  
  if (fabs(v) <= bound2) v = 0.0;
  U *= gammaToM1;
  u *= gammaToM1;
  V *= gamma;
  v *= gamma;
  m = k + 1;
  
  doubleDoubleAdd(&S,&s,U,u,V,v);
  
  return scaledSquareRoot(m, S);
}

/* Our faithful norm with 2 bins

   Attention: the number of elements needs to be even!

*/
double faithfulNorm(CONST double * RESTRICT x, unsigned int n) {
  unsigned int i;
  vec_t xChunk, xAbs, xA, xSMid, xC, boundA, boundB;
  mask_t maskA, maskB, maskC, maskT, maskCP;
  mask_t discardExt, bigOccured, bigOccuredNew;
  vec_t scaleA, scaleC, scaleExt, xExt, xSExt, two700, twoM700;
  vec_t sMid, rMid, sExt, rExt;
  vec_t pMidh, pMidl, sMidh, sMidl, uMid1, uMid2, uMid3, uMid4, tMid;
  vec_t pExth, pExtl, sExth, sExtl, uExt1, uExt2, uExt3, uExt4, tExt;

  if (n % vec_len != 0) return 0.0;

  /* Load two bounds, 2^326 and 2^-374 */
  boundA = vec_set1(0x1p326);    /* 2^326 */
  boundB = vec_set1(0x1p-374);   /* 2^-374 */

  /* Load two scaling factors 2^-700 and 2^700 */
  twoM700 = vec_set1(0x1p-700);  /* 2^-700 */  
  two700 = vec_set1(0x1p700);   /* 2^700 */
  
  /* Initialize the accumulators to zero */
  sMid = vec_setzero();
  rMid = vec_setzero();
  sExt = vec_setzero();
  rExt = vec_setzero();

  /* A flag if we ever saw a large (A) bin value */
  bigOccured = mask_setzero();
  
  /* Main loop over the vector */
  for (i=0; i<=n-vec_len; i+=vec_len) {
    xChunk = vec_load(&(x[i]));
    xAbs = vec_abs(xChunk);

    maskA = vec_cmpge(xAbs, boundA);
    maskC = vec_cmplt(xAbs, boundB);
    maskT = vec_cmpge(xAbs, boundB);
    maskB = mask_andnot(maskA, maskT);
    xA = vec_select_vorz(maskA, xAbs);
    xSMid = vec_select_vorz(maskB, xAbs);

    bigOccuredNew = mask_or(bigOccured, maskA);
    
    maskCP = mask_andnot(bigOccuredNew, maskC);

    xC = vec_select_vorz(maskCP, xAbs);
    xExt = vec_or(xA, xC);

    scaleA = vec_select_vorz(maskA, twoM700);
    scaleC = vec_select_vorz(maskCP, two700);
    scaleExt = vec_or(scaleA, scaleC);
    xSExt = vec_mul(xExt, scaleExt);
   
    discardExt = mask_xor(bigOccuredNew, bigOccured);

    bigOccured = bigOccuredNew;

    sExt = vec_select_zorv(discardExt, sExt);
    rExt = vec_select_zorv(discardExt, rExt);

    vec_twosqr(&pMidh, &pMidl, xSMid);

    sMidh = vec_add(sMid, pMidh);
    uMid1 = vec_sub(sMidh, sMid);
    uMid2 = vec_sub(sMidh, uMid1);
    uMid3 = vec_sub(pMidh, uMid1);
    uMid4 = vec_sub(sMid, uMid2);
    sMidl = vec_add(uMid3, uMid4);
    tMid = vec_add(rMid, vec_add(pMidl, sMidl));
    sMid = vec_add(sMidh, tMid);
    rMid = vec_sub(tMid, vec_sub(sMid, sMidh));

    vec_twosqr(&pExth, &pExtl, xSExt);
    
    sExth = vec_add(sExt, pExth);
    uExt1 = vec_sub(sExth, sExt);
    uExt2 = vec_sub(sExth, uExt1);
    uExt3 = vec_sub(pExth, uExt1);
    uExt4 = vec_sub(sExt, uExt2);
    sExtl = vec_add(uExt3, uExt4);
    tExt = vec_add(rExt, vec_add(pExtl, sExtl));

    sExt = vec_add(sExth, tExt);
    rExt = vec_sub(tExt, vec_sub(sExt, sExth));
  }

  vec_t sA, rA, sB, rB, sC, rC;
  sB = sMid;
  rB = rMid;

  sA = vec_select_vorz(bigOccured, sExt);
  rA = vec_select_vorz(bigOccured, rExt);

  sC = vec_select_zorv(bigOccured, sExt);
  rC = vec_select_zorv(bigOccured, rExt);

  double sfA, rfA, sfB, rfB, sfC, rfC;
  double SA[vec_len], RA[vec_len];
  double SB[vec_len], RB[vec_len];
  double SC[vec_len], RC[vec_len];
  vec_store(SA, sA);
  vec_store(RA, rA);
  vec_store(SB, sB);
  vec_store(RB, rB);
  vec_store(SC, sC);
  vec_store(RC, rC);
  doubleDoubleAdd(&sfA, &rfA, SA[0], RA[0], SA[1], RA[1]);
  doubleDoubleAdd(&sfB, &rfB, SB[0], RB[0], SB[1], RB[1]);
  doubleDoubleAdd(&sfC, &rfC, SC[0], RC[0], SC[1], RC[1]);
  for(i = 2; i < vec_len; i++) {
    doubleDoubleAdd(&sfA, &rfA, sfA, rfA, SA[i], RA[i]);
    doubleDoubleAdd(&sfB, &rfB, sfB, rfB, SB[i], RB[i]);
    doubleDoubleAdd(&sfC, &rfC, sfC, rfC, SC[i], RC[i]);
  }
  
  if (sfA != 0.0) return faithfulFinal(-2, sfA, rfA, sfB, rfB);
  if (sfB != 0.0) return faithfulFinal(0, sfB, rfB, sfC, rfC);
  return faithfulFinal(2, sfC, rfC, 0.0, 0.0);
}
