/* This program is part of a correctly-rounded, division-free l2
   vector norm with avoidance of spurious underflow and overflow.

   Author: Nicolas Louvet,

           Université Claude Bernard Lyon 1

   implementing the algorithm described in the paper

   Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
   Laurence Rideau. Accurate calculation of Euclidean Norms using
   Double-word arithmetic. 2021

   This program is

   Copyright (C) 2021 Université Claude Bernard Lyon 1

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <http://www.gnu.org/licenses/>.

*/

#include <stdint.h>
#include <math.h>
#include <stdio.h>

#include "vec.h"

#define PREC      53
#define GAMMA    700
#define EMAXMED  485
#define EMINMED -484
#define EBND1 (2*EMAXMED - 2*PREC)
#define EBND2 (2*EMAXMED - 2*GAMMA + 3*PREC)
#define BND1 (ldexp(1.0, EBND1))
#define BND2 (ldexp(1.0, EBND2))
#define EBND3 (2*EMINMED + 2*GAMMA - 2*PREC)
#define EBND4 (2*EMINMED + 3*PREC)
#define BND3 (ldexp(1.0, EBND3))
#define BND4 (ldexp(1.0, EBND4))

/* **************************************************************************************************** */

/* non-SSE twoSum: computes u + v = a + b exactly, where u = RN(a + b) */
static inline void twoSum(double *u, double *v, double a, double b) {
  double u1, u2, u3, u4, s, r;

  s = a + b;
  u1 = s - a;
  u2 = s - u1;
  u3 = b - u1;
  u4 = a - u2;
  r = u4 + u3;

  *u = s;
  *v = r;
}

/* non-SSE fastTwoSum: computes u + v = a + b exactly, where u = RN(a + b),
   UNDER THE CONDITION THAT abs(a) >= abs(b) */
static inline void fastTwoSum(double *u, double *v, double a, double b) {
  double z, s, r;
  s = a + b;
  z = s - a;
  r = b - z;
  *u = s;
  *v = r;
}

/* non-SSE double-double add: computes u + v approximating ah + al + bh + bl
   corresponds to Algorithm 7 – SloppyDWPlusDW */
static inline void sloppyDDPlusDD(double *u, double *v, double ah, double al, double bh, double bl) {
  double th, tl, r;

  twoSum(&th, &tl, ah, bh);
  r = tl + (al + bl);
  fastTwoSum(u, v, th, r);
}

/* non-SSE double-double add: computes u + v approximating ah + al + b
   corresponds to Algorithm 6 – DWPlusFP */
static inline void DDPlusD(double *zh, double *zl, double ah, double al, double b) {
  double th, tl, r;

  twoSum(&th, &tl, ah, b);
  r = al + tl;
  fastTwoSum(zh, zl, th, r);
}

#if defined(__FP_FAST_FMA) // See "Common Predefined Macros" in GNU CPP documentation

#pragma message("Using fma")

static inline void twoSquare(double *u, double *v, double a) {
  *u = a * a;
  *v = fma(a, a, -(*u));
}

#else

/* non-SSE twoSquare: computes u + v = a^2 exactly, where u = RN(a^2) */
static inline void twoSquare(double *u, double *v, double a) {
  static const double c = 134217729.0;
  double ap, a1, a2;

  ap = a*c;
  a1 = (a-ap)+ap;
  a2 = a-a1;

  *u = a*a;
  *v = (((a1*a1-(*u))+(a1*a2))+(a2*a1))+(a2*a2);
}

#endif

/* **************************************************************************************************** */

/* SIMD version of DDPlusD */
static inline void vec_DDPlusD(vec_t *zh, vec_t *zl, vec_t ah, vec_t al, vec_t b) {
  vec_t th, tl, r;

  vec_twosum(&th, &tl, ah, b);
  r = vec_add(al, tl);
  vec_fast_twosum(zh, zl, th, r);
}

/* **************************************************************************************************** */

static inline void SQRTDW(double *s_hi, double *s_lo, double x_hi, double x_lo) {
  double r1, r2, sh, sl;
  *s_hi = sqrt(x_hi);
  twoSquare(&sh, &sl, *s_hi);
  r1 = x_hi - sh - sl; /* r1 = - fma(*s_hi, *s_hi, - x_hi); */
  r2 = x_lo + r1;
  *s_lo = r2 / (2.0 * (*s_hi));
}

static inline double SQRTDWtoFP(double x_hi, double x_lo) {
  double s_hi, s_lo;
  SQRTDW(&s_hi, &s_lo, x_hi, x_lo);
  return s_hi + s_lo;
}

double SQRT2DW_big(double mid_hi, double mid_lo, double big_hi, double big_lo) {
  double s_hi, s_lo;
  double t_hi, t_lo;

  if(mid_hi < BND1 || big_hi > BND2) {
    SQRTDW(&s_hi, &s_lo, big_hi, big_lo);
    return scalbln(s_hi, GAMMA) + scalbln(s_lo, GAMMA);
  }
  else {
    big_hi = scalbln(big_hi,  GAMMA); big_lo = scalbln(big_lo,  GAMMA);
    mid_hi = scalbln(mid_hi, -GAMMA); mid_lo = scalbln(mid_lo, -GAMMA);
    sloppyDDPlusDD(&t_hi, &t_lo, big_hi, big_lo, mid_hi, mid_lo);
    SQRTDW(&s_hi, &s_lo, t_hi, t_lo);
    return scalbln(s_hi, GAMMA >> 1) + scalbln(s_lo, GAMMA >> 1);
  }
}

double SQRT2DW_med(double mid_hi, double mid_lo, double tiny_hi, double tiny_lo) {
  double s_hi, s_lo;
  double t_hi, t_lo;

  if(tiny_hi < BND3 || mid_hi > BND4) {
    SQRTDW(&s_hi, &s_lo, mid_hi, mid_lo);
    return s_hi + s_lo;
  }
  else {
    mid_hi  = scalbln(mid_hi,   GAMMA);  mid_lo = scalbln(mid_lo,   GAMMA);
    tiny_hi = scalbln(tiny_hi, -GAMMA); tiny_lo = scalbln(tiny_lo, -GAMMA);
    sloppyDDPlusDD(&t_hi, &t_lo, mid_hi, mid_lo, tiny_hi, tiny_lo);
    SQRTDW(&s_hi, &s_lo, t_hi, t_lo);
    return scalbln(s_hi, - (GAMMA >> 1)) + scalbln(s_lo, - (GAMMA >> 1));
  }
}

/* **************************************************************************************************** */

static void SOS(double *S_mid_hi, double *S_mid_lo, double *Sigma_mid,
                int *Ext_is_big,
                double *S_ext_hi, double *S_ext_lo, double *Sigma_ext,
                const double *x, unsigned int n) {

  double minmed = ldexp(1.0, EMINMED);
  double maxmed = ldexp(1.0, EMAXMED);
  double t_big = ldexp(1.0, -GAMMA);
  double t_tiny = ldexp(1.0, GAMMA);

  *S_mid_hi = *S_mid_lo = *Sigma_mid = 0.0;
  *S_ext_hi = *S_ext_lo = *Sigma_ext = 0.0;
  double a_hi, a_lo;
  int ext_is_big = 0;

  for(int i = 0; i < n; i++) {
      double x_abs = fabs(x[i]);
      int is_big = x_abs >= maxmed;
      int is_tiny = x_abs < minmed;
          
      if(is_big) {
          if(!ext_is_big) {
              ext_is_big = 1;
              *S_ext_hi = *S_ext_lo = *Sigma_ext = 0.0;
          }
          twoSquare(&a_hi, &a_lo, t_big * x[i]);
          DDPlusD(S_ext_hi, S_ext_lo, *S_ext_hi, *S_ext_lo, a_hi);
          *Sigma_ext += a_lo;
      }
      else if(is_tiny){
          if(!ext_is_big) {
              twoSquare(&a_hi, &a_lo, t_tiny * x[i]);
              DDPlusD(S_ext_hi, S_ext_lo, *S_ext_hi, *S_ext_lo, a_hi);
              *Sigma_ext += a_lo;
          }
      }
      else {
          twoSquare(&a_hi, &a_lo, x[i]);
          DDPlusD(S_mid_hi, S_mid_lo, *S_mid_hi, *S_mid_lo, a_hi);
          *Sigma_mid += a_lo;
      }
  }
  *Ext_is_big = ext_is_big;

}

/* **************************************************************************************************** */

/* SIMD version of SOS ; returns the number of elements in x that have been taken
   into account, that is floor(n/vec_len) * vec_len.
*/
unsigned int vec_SOS(vec_t *S_mid_hi, vec_t *S_mid_lo, vec_t *Sigma_mid,
                     mask_t *Ext_is_big,
                     vec_t *S_ext_hi, vec_t *S_ext_lo, vec_t *Sigma_ext,
                     const double *x, unsigned int n) {
  
  unsigned int i;
  
  vec_t x_vec, x_abs, x_mid, x_ext;
  vec_t s_mid_hi, s_mid_lo;
  vec_t s_ext_hi, s_ext_lo;
  mask_t is_tiny, is_big, is_med;
  mask_t is_ext, is_first_big, ext_is_big;
  vec_t minmed, maxmed, t_tiny, t_big, scale;

  vec_t p_hi, p_lo, s_hi, s_lo, sigma_mid, sigma_ext;
  vec_t v;

  /* lower bound for the medium range  */
  minmed = vec_set1(ldexp(1.0, EMINMED));
  /* upper bound for the medium range  */
  maxmed = vec_set1(ldexp(1.0, EMAXMED));

  /* scaling factor for the big elements */
  t_big = vec_set1(ldexp(1.0, -GAMMA));
  /* scaling factor for the tiny elements */
  t_tiny = vec_set1(ldexp(1.0, GAMMA));

  /* initializing the accumulators to zero */
  s_mid_hi  = vec_setzero();
  s_mid_lo  = vec_setzero();
  sigma_mid = vec_setzero();

  s_ext_hi  = vec_setzero();
  s_ext_lo  = vec_setzero();
  sigma_ext = vec_setzero();

  /* ext_is_big[j] = 0ull */
  ext_is_big = mask_setzero();

  for (i = 0; i < n-vec_len; i += vec_len) {
    /* load vec_len consecutive elements, x[i] and x[i+1], into x_vec */
    x_vec = vec_load(x+i);
    /* compute absolute values. */
    x_abs = vec_abs(x_vec);
    
    /* is_big[j]  = x_abs[j] ≥ maxmed[j] */
    is_big  = vec_cmpge(x_abs, maxmed);
    /* is_tiny[j] = x_abs[j] < minmed[j] */
    is_tiny = vec_cmplt(x_abs, minmed);
    /* is_med[j]  = ~is_big[j] & ~is_tiny[j] */
    is_med  = mask_andnot(is_big, mask_not(is_tiny));

    /* is_ext[j] = is_big[j] | (~ext_is_big[j] & is_tiny[j]) */
    is_ext = mask_or(is_big, mask_andnot(ext_is_big, is_tiny));
    
    /* is_first_big[j] = ~ext_is_big[j] & is_big[j] */
    is_first_big = mask_andnot(ext_is_big, is_big);
    /* s_ext[j] = (is_first_big[j] ? 0 : s_ext[j]) */
    s_ext_hi = vec_select_zorv(is_first_big, s_ext_hi);
    s_ext_lo = vec_select_zorv(is_first_big, s_ext_lo);
    sigma_ext = vec_select_zorv(is_first_big, sigma_ext);
    /* ext_is_big[j] = ext_is_big[j] | is_big[j] */
    ext_is_big = mask_or(ext_is_big, is_big);

    /* scale[j] = is_big[j] ? t_big[j] : t_tiny[j] */
    scale = vec_select(is_big, t_big, t_tiny);
    /* scale[j] = is_ext[j] ? scale[j] : 0.0 */
    scale = vec_select_vorz(is_ext, scale);

    /* x_mid[j] = is_med[j] ? x_abs[j] : 0 */
    x_mid = vec_select_vorz(is_med, x_abs);
    /* x_ext[j] = is_ext[j] ? x_abs[j] : 0 */
    x_ext = vec_select_vorz(is_ext, x_abs);

    /* accumulates the square of x_mid[j] into (s_mid_hi[j], s_mid_lo[j]) */
    vec_twosqr(&p_hi, &p_lo, x_mid);
    sigma_mid = vec_add(sigma_mid, p_lo);
    /* (s_mid_hi[j], s_mid_lo[j]) = DWPlusFP((s_mid_hi[j], s_mid_lo[j]), p_hi) */ 
    vec_twosum(&s_hi, &s_lo, s_mid_hi, p_hi);
    v = vec_add(s_mid_lo, s_lo);
    vec_fast_twosum(&s_mid_hi, &s_mid_lo, s_hi, v);

    /* scaling of x_ext */
    x_ext = vec_mul(x_ext, scale);
    /* accumulates the square of x_ext[j] into (s_ext_hi[j], s_ext_lo[j]) */
    vec_twosqr(&p_hi, &p_lo, x_ext);
    sigma_ext = vec_add(sigma_ext, p_lo);
    /* (s_mid_hi[j], s_mid_lo[j]) = DWPlusFP((s_mid_hi[j], s_mid_lo[j]), p_hi) */ 
    vec_twosum(&s_hi, &s_lo, s_ext_hi, p_hi);
    v = vec_add(s_ext_lo, s_lo);
    vec_fast_twosum(&s_ext_hi, &s_ext_lo, s_hi, v);
  }

  *S_mid_hi = s_mid_hi;
  *S_mid_lo = s_mid_lo;
  *Sigma_mid = sigma_mid;
  *Ext_is_big = ext_is_big;
  *S_ext_hi = s_ext_hi;
  *S_ext_lo = s_ext_lo;
  *Sigma_ext = sigma_ext;

  return i;
}

/* **************************************************************************************************** */

double accurateNorm0(const double *x, unsigned int n) {
  double S_mid_hi, S_mid_lo, Sigma_mid;
  double S_ext_hi, S_ext_lo, Sigma_ext;
  int Ext_is_big = 0;
  double r;

  SOS(&S_mid_hi, &S_mid_lo, &Sigma_mid, &Ext_is_big, &S_ext_hi, &S_ext_lo, &Sigma_ext, x, n);
  DDPlusD(&S_mid_hi, &S_mid_lo, S_mid_hi, S_mid_lo, Sigma_mid);
  DDPlusD(&S_ext_hi, &S_ext_lo, S_ext_hi, S_ext_lo, Sigma_ext);

  if(Ext_is_big) {
    r = SQRT2DW_big(S_mid_hi, S_mid_lo, S_ext_hi, S_ext_lo);
  }
  else {
    if(S_mid_hi == 0.0 && S_mid_lo == 0.0) {
      r = scalbln(SQRTDWtoFP(S_ext_hi, S_ext_lo), -GAMMA);
    }
    else {
      r =  SQRT2DW_med(S_mid_hi, S_mid_lo, S_ext_hi, S_ext_lo);
    }
  }
  
  return r;
}

/* **************************************************************************************************** */

double accurateNorm1(const double *x, unsigned int n) {
  unsigned int i;
  
  vec_t x_vec, x_abs, x_mid, x_ext;
  vec_t s_mid_hi, s_mid_lo;
  vec_t s_ext_hi, s_ext_lo;
  mask_t is_tiny, is_big, is_med;
  mask_t is_ext, is_first_big, ext_is_big;
  vec_t minmed, maxmed, t_tiny, t_big, scale;

  vec_t p_hi, p_lo, s_hi, s_lo, sigma_mid, sigma_ext;
  vec_t v;

  if (n%vec_len != 0) return 0.0;

  /* lower bound for the medium range  */
  minmed = vec_set1(ldexp(1.0, EMINMED));
  /* upper bound for the medium range  */
  maxmed = vec_set1(ldexp(1.0, EMAXMED));

  /* scaling factor for the big elements */
  t_big = vec_set1(ldexp(1.0, -GAMMA));
  /* scaling factor for the tiny elements */
  t_tiny = vec_set1(ldexp(1.0, GAMMA));

  /* initializing the accumulators to zero */
  s_mid_hi  = vec_setzero();
  s_mid_lo  = vec_setzero();
  sigma_mid = vec_setzero();

  s_ext_hi  = vec_setzero();
  s_ext_lo  = vec_setzero();
  sigma_ext = vec_setzero();

  /* ext_is_big[j] = 0ull */
  ext_is_big = mask_setzero();

  for (i=0; i<=n-vec_len; i+=vec_len) {
    /* load vec_len consecutive elements, x[i] and x[i+1], into x_vec */
    x_vec = vec_load(x+i);
    /* compute absolute values. */
    x_abs = vec_abs(x_vec);
    
    /* is_big[j]  = x_abs[j] ≥ maxmed[j] */
    is_big  = vec_cmpge(x_abs, maxmed);
    /* is_tiny[j] = x_abs[j] < minmed[j] */
    is_tiny = vec_cmplt(x_abs, minmed);
    /* is_med[j]  = ~is_big[j] & ~is_tiny[j] */
    is_med  = mask_andnot(is_big, mask_not(is_tiny));

    /* is_ext[j] = is_big[j] | (~ext_is_big[j] & is_tiny[j]) */
    is_ext = mask_or(is_big, mask_andnot(ext_is_big, is_tiny));
    
    /* is_first_big[j] = ~ext_is_big[j] & is_big[j] */
    is_first_big = mask_andnot(ext_is_big, is_big);
    /* s_ext[j] = (is_first_big[j] ? 0 : s_ext[j]) */
    s_ext_hi = vec_select_zorv(is_first_big, s_ext_hi);
    s_ext_lo = vec_select_zorv(is_first_big, s_ext_lo);
    sigma_ext = vec_select_zorv(is_first_big, sigma_ext);
    /* ext_is_big[j] = ext_is_big[j] | is_big[j] */
    ext_is_big = mask_or(ext_is_big, is_big);

    /* scale[j] = is_big[j] ? t_big[j] : t_tiny[j] */
    scale = vec_select(is_big, t_big, t_tiny);
    /* scale[j] = is_ext[j] ? scale[j] : 0.0 */
    scale = vec_select_vorz(is_ext, scale);

    /* x_mid[j] = is_med[j] ? x_abs[j] : 0 */
    x_mid = vec_select_vorz(is_med, x_abs);
    /* x_ext[j] = is_ext[j] ? x_abs[j] : 0 */
    x_ext = vec_select_vorz(is_ext, x_abs);

    /* accumulates the square of x_mid[j] into (s_mid_hi[j], s_mid_lo[j]) */
    vec_twosqr(&p_hi, &p_lo, x_mid);
    sigma_mid = vec_add(sigma_mid, p_lo);
    /* (s_mid_hi[j], s_mid_lo[j]) = DWPlusFP((s_mid_hi[j], s_mid_lo[j]), p_hi) */ 
    vec_twosum(&s_hi, &s_lo, s_mid_hi, p_hi);
    v = vec_add(s_mid_lo, s_lo);
    vec_fast_twosum(&s_mid_hi, &s_mid_lo, s_hi, v);

    /* scaling of x_ext */
    x_ext = vec_mul(x_ext, scale);
    /* accumulates the square of x_ext[j] into (s_ext_hi[j], s_ext_lo[j]) */
    vec_twosqr(&p_hi, &p_lo, x_ext);
    sigma_ext = vec_add(sigma_ext, p_lo);
    /* (s_mid_hi[j], s_mid_lo[j]) = DWPlusFP((s_mid_hi[j], s_mid_lo[j]), p_hi) */ 
    vec_twosum(&s_hi, &s_lo, s_ext_hi, p_hi);
    v = vec_add(s_ext_lo, s_lo);
    vec_fast_twosum(&s_ext_hi, &s_ext_lo, s_hi, v);
  }
  
  vec_twosum(&s_hi, &s_lo, s_mid_hi, sigma_mid);
  v = vec_add(s_mid_lo, s_lo);
  vec_fast_twosum(&s_mid_hi, &s_mid_lo, s_hi, v);

  vec_twosum(&s_hi, &s_lo, s_ext_hi, sigma_ext);
  v = vec_add(s_ext_lo, s_lo);
  vec_fast_twosum(&s_ext_hi, &s_ext_lo, s_hi, v);
  
  int ext_is_big_array[vec_len];
  mask_store(ext_is_big_array, ext_is_big);
  
  double VEC_ALGN s_mid_array_hi[vec_len];
  double VEC_ALGN s_mid_array_lo[vec_len];
  vec_store_algn(s_mid_array_hi, s_mid_hi);
  vec_store_algn(s_mid_array_lo, s_mid_lo);
  
  double VEC_ALGN s_ext_array_hi[vec_len];
  double VEC_ALGN s_ext_array_lo[vec_len];
  vec_store_algn(s_ext_array_hi, s_ext_hi);
  vec_store_algn(s_ext_array_lo, s_ext_lo);
  
  double S_mid_hi, S_mid_lo, Sigma_mid;
  double S_ext_hi, S_ext_lo, Sigma_ext;
  S_mid_hi = S_mid_lo = Sigma_mid = 0.0;
  S_ext_hi = S_ext_lo = Sigma_ext = 0.0;
  /*
  int Ext_is_big = 0;
  for(i = 0; i < vec_len; i++) {
    if(ext_is_big_array[i]) {
      if(!Ext_is_big) {
        S_ext_hi = S_ext_lo = Sigma_ext = 0.0;
        Ext_is_big = 1;
      }
      DDPlusD(&S_ext_hi, &S_ext_lo, S_ext_hi, S_ext_lo, s_ext_array_hi[i]);
      Sigma_ext += s_ext_array_lo[i];
    }
    else {
      if(!Ext_is_big) {
        DDPlusD(&S_ext_hi, &S_ext_lo, S_ext_hi, S_ext_lo, s_ext_array_hi[i]);
        Sigma_ext += s_ext_array_lo[i];
      }
    }
    DDPlusD(&S_mid_hi, &S_mid_lo, S_mid_hi, S_mid_lo, s_mid_array_hi[i]);
    Sigma_mid += s_mid_array_lo[i];
  }

  DDPlusD(&S_mid_hi, &S_mid_lo, S_mid_hi, S_mid_lo, Sigma_mid);
  DDPlusD(&S_ext_hi, &S_ext_lo, S_ext_hi, S_ext_lo, Sigma_ext);
  */

  int Is_first_big, Is_tiny, Ext_is_big = 0;
  double X_ext_hi, X_ext_lo;
  for(i = 0; i < vec_len; i++) {
    Is_first_big  = !Ext_is_big && ext_is_big_array[i];
    S_ext_hi = (Is_first_big ? 0.0 : S_ext_hi);
    S_ext_lo = (Is_first_big ? 0.0 : S_ext_lo);
    Sigma_ext = (Is_first_big ? 0.0 : Sigma_ext);
    Ext_is_big = Ext_is_big || ext_is_big_array[i];
    Is_tiny = !ext_is_big_array[i] && Ext_is_big;
    X_ext_hi = (Is_tiny ? 0.0 : s_ext_array_hi[i]);
    X_ext_lo = (Is_tiny ? 0.0 : s_ext_array_lo[i]);
    DDPlusD(&S_ext_hi, &S_ext_lo, S_ext_hi, S_ext_lo, X_ext_hi);
    Sigma_ext += X_ext_lo;
    DDPlusD(&S_mid_hi, &S_mid_lo, S_mid_hi, S_mid_lo, s_mid_array_hi[i]);
    Sigma_mid += s_mid_array_lo[i];
  }

  DDPlusD(&S_mid_hi, &S_mid_lo, S_mid_hi, S_mid_lo, Sigma_mid);
  DDPlusD(&S_ext_hi, &S_ext_lo, S_ext_hi, S_ext_lo, Sigma_ext);
  
  double r;
  if(Ext_is_big) {
    r = SQRT2DW_big(S_mid_hi, S_mid_lo, S_ext_hi, S_ext_lo);
  }
  else {
    if(S_mid_hi == 0.0 && S_mid_lo == 0.0) {
      r = scalbln(SQRTDWtoFP(S_ext_hi, S_ext_lo), -GAMMA);
    }
    else {
      r =  SQRT2DW_med(S_mid_hi, S_mid_lo, S_ext_hi, S_ext_lo);
    }
  }
  
  return r;
}

/* **************************************************************************************************** */

double accurateNorm(const double *x, unsigned int n) {
  vec_t s_mid_hi, s_mid_lo, sigma_mid;
  vec_t s_ext_hi, s_ext_lo, sigma_ext;
  mask_t ext_is_big;

  unsigned int i = vec_SOS(&s_mid_hi, &s_mid_lo, &sigma_mid,
                           &ext_is_big,
                           &s_ext_hi, &s_ext_lo, &sigma_ext,
                           x, n);

  vec_DDPlusD(&s_mid_hi, &s_mid_lo, s_mid_hi, s_mid_lo, sigma_mid);
  vec_DDPlusD(&s_ext_hi, &s_ext_lo, s_ext_hi, s_ext_lo, sigma_ext);

  double S_mid_hi, S_mid_lo, Sigma_mid;
  double S_ext_hi, S_ext_lo, Sigma_ext;
  int Ext_is_big;

  SOS(&S_mid_hi, &S_mid_lo, &Sigma_mid, &Ext_is_big, &S_ext_hi, &S_ext_lo, &Sigma_ext, x+i, n-i);
  
  int ext_is_big_array[vec_len];
  mask_store(ext_is_big_array, ext_is_big);
  
  double VEC_ALGN s_mid_array_hi[vec_len];
  double VEC_ALGN s_mid_array_lo[vec_len];
  vec_store_algn(s_mid_array_hi, s_mid_hi);
  vec_store_algn(s_mid_array_lo, s_mid_lo);
  
  double VEC_ALGN s_ext_array_hi[vec_len];
  double VEC_ALGN s_ext_array_lo[vec_len];
  vec_store_algn(s_ext_array_hi, s_ext_hi);
  vec_store_algn(s_ext_array_lo, s_ext_lo);

  for(i = 0; i < vec_len; i++) {
    if(ext_is_big_array[i]) {
      if(!Ext_is_big) {
        S_ext_hi = S_ext_lo = Sigma_ext = 0.0;
        Ext_is_big = 1;
      }
      DDPlusD(&S_ext_hi, &S_ext_lo, S_ext_hi, S_ext_lo, s_ext_array_hi[i]);
      Sigma_ext += s_ext_array_lo[i];
    }
    else {
      if(!Ext_is_big) {
        DDPlusD(&S_ext_hi, &S_ext_lo, S_ext_hi, S_ext_lo, s_ext_array_hi[i]);
        Sigma_ext += s_ext_array_lo[i];
      }
    }
    DDPlusD(&S_mid_hi, &S_mid_lo, S_mid_hi, S_mid_lo, s_mid_array_hi[i]);
    Sigma_mid += s_mid_array_lo[i];
  }

  DDPlusD(&S_mid_hi, &S_mid_lo, S_mid_hi, S_mid_lo, Sigma_mid);
  DDPlusD(&S_ext_hi, &S_ext_lo, S_ext_hi, S_ext_lo, Sigma_ext);

  double r;
  if(Ext_is_big) {
    r = SQRT2DW_big(S_mid_hi, S_mid_lo, S_ext_hi, S_ext_lo);
  }
  else {
    if(S_mid_hi == 0.0 && S_mid_lo == 0.0) {
      r = scalbln(SQRTDWtoFP(S_ext_hi, S_ext_lo), -GAMMA);
    }
    else {
      r =  SQRT2DW_med(S_mid_hi, S_mid_lo, S_ext_hi, S_ext_lo);
    }
  }
  
  return r;
}
