/* This program is part of a faithfully-rounded, division-free l2
   vector norm with avoidance of spurious underflow and overflow.

   Author: Christoph Lauter,

           Université Pierre et Marie Curie Paris 6, UPMC, LIP6,
	   PEQUAN team.

   implementing the algorithm described in the paper

   Graillat, Lauter, Tang, Yamanaka and Oishi: Efficient calculations
   of faithfully rounded l2-norms of n-vectors.

   This program is

   Copyright (C) 2014 Université Pierre et Marie Curie Paris 6, UPMC,
   LIP6, PEQUAN team.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <http://www.gnu.org/licenses/>. 

*/
/* Modified by: Nicolas Louvet,

           Université Claude Bernard Lyon 1

   implementing the algorithm described in the paper

   Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
   Laurence Rideau. Accurate calculation of Euclidean Norms using
   Double-word arithmetic. 2021

*/

#include <gmp.h>
#include <mpfr.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "randomvectors.h"
#include "dblcast.h"

/* Compute a correctly rounded double precision square root of op
   rounded in rounding direction rnd. rop is supposed to have at least
   53 bits of precision.
*/
int mpfr_sqrt_d(mpfr_t rop, mpfr_t op, mp_rnd_t rnd) {
  mp_exp_t oldEmin, oldEmax;
  mp_prec_t oldDefaultPrec;
  int ternaryA, ternaryB;
  mpfr_t temp;

  oldEmin = mpfr_get_emin();
  oldEmax = mpfr_get_emax();
  oldDefaultPrec = mpfr_get_default_prec();

  mpfr_set_default_prec(53);
  mpfr_set_emin(-1073); mpfr_set_emax(1024);
  
  mpfr_init(temp);

  ternaryA = mpfr_sqrt(temp, op, rnd);
  ternaryB = mpfr_subnormalize(temp, ternaryA, rnd);

  mpfr_set(rop, temp, GMP_RNDN); /* exact because prec(rop) >= 53 */

  mpfr_set_emin(oldEmin);
  mpfr_set_emax(oldEmax);
  mpfr_set_default_prec(oldDefaultPrec);

  mpfr_clear(temp);

  return ternaryB;
}

void computeUlp(mpfr_t ulp, mpfr_t yd) {
  dblcast ydb, sdb;

  ydb.d = mpfr_get_d(yd, GMP_RNDN); /* exact */

  if (ydb.i == SIGN_MASK) {
    sdb.i = ydb.i - 1;
    ydb.i = sdb.i - 2;
    ydb.d = sdb.d - ydb.d;
  } else {
    if (!((ydb.i & SIGN_MASK) == SIGN_MASK)) {
      sdb.i = ydb.i + 1;
      if (sdb.i == SIGN_MASK) {
	sdb.i = ydb.i - 1;
	ydb.i = ydb.i - 2;
      }
      ydb.d = sdb.d - ydb.d;
    }
  }

  mpfr_set_d(ulp, ydb.d, GMP_RNDN);

}

int isFaithfulNorm(double y, CONST double * RESTRICT x, int n, int displayErrors, int *underflowCases, int *overflowCases, mpfr_t eps, mpfr_t ulpErr) {
  mpfr_t acc, t, yd, yu, ym, yAcc, temp, ulp;
  int i, res;
  dblcast xdb, ydb, adb, bdb;

  mpfr_init2(acc, (2 * EXPO_MAX) - (2 * EXPO_MIN) + 1 + (2 * PRECISION));
  mpfr_init2(t, (2 * PRECISION));
  mpfr_init2(yd, PRECISION);
  mpfr_init2(yu, PRECISION);
  mpfr_init2(ym, PRECISION);
  mpfr_init2(yAcc, 500);
  mpfr_init2(temp, 500);
  mpfr_init2(ulp, 64);

  mpfr_set_d(ym, y, GMP_RNDN); /* exact */

  mpfr_set_si(acc, 0, GMP_RNDN); /* exact */
  for (i=0; i<n; i++) {
    mpfr_set_d(t, x[i], GMP_RNDN); /* exact */
    mpfr_mul(t, t, t, GMP_RNDN); /* exact */
    mpfr_add(acc, acc, t, GMP_RNDN); /* exact */
  }
  mpfr_sqrt_d(yd, acc, GMP_RNDD); /* one correct rounding DOWN */
  mpfr_sqrt_d(yu, acc, GMP_RNDU); /* one correct rounding UP */
  mpfr_sqrt(yAcc, acc, GMP_RNDN); /* Some error for max. eps estimation only */

  if (!mpfr_zero_p(yAcc)) {
    mpfr_sub(temp, ym, yAcc, GMP_RNDN);
    mpfr_div(temp, temp, yAcc, GMP_RNDN);
    mpfr_abs(temp, temp, GMP_RNDN);
    if (mpfr_cmp(temp, eps) > 0) {
      mpfr_set(eps, temp, GMP_RNDN);
    }
  }

  computeUlp(ulp, yd);
  mpfr_sub(temp, ym, yAcc, GMP_RNDN);
  mpfr_abs(temp, temp, GMP_RNDN);
  mpfr_div(temp, temp, ulp, GMP_RNDN);
  if (mpfr_cmp(temp, ulpErr) > 0) {
    mpfr_set(ulpErr, temp, GMP_RNDN);
  }

  res = (mpfr_equal_p(ym,yd) || mpfr_equal_p(ym, yu));

  if ((!res) && (displayErrors)) {
    ydb.d = y;
    adb.d = mpfr_get_d(yd, GMP_RNDN);
    bdb.d = mpfr_get_d(yu, GMP_RNDN);
    printf("Not faithfully rounded: n = %d, y = %1.50e = 0x%016llx = %a, should be one of 0x%016llx or 0x%016llx\n", n, y, 
	   (long long unsigned int) ydb.i, y, (long long unsigned int) adb.i, (long long unsigned int) bdb.i);
    printf("x = [\n");
    for (i=0; i<n; i++) {
      xdb.d = x[i];
      printf("%1.50e = 0x%016llx = %a\n", x[i], (long long unsigned int) xdb.i, x[i]);
    }
    printf("]\n");
  }

  ydb.d = y;
  if (((ydb.i & SIGN_MASK) >> (PRECISION - 1)) == 0ull) {
    (*underflowCases)++;
  }
  if ((ydb.i & SIGN_MASK) == SIGN_MASK) {
    (*overflowCases)++;
  }


  mpfr_clear(acc);
  mpfr_clear(t);
  mpfr_clear(yd);
  mpfr_clear(yu);
  mpfr_clear(yAcc);
  mpfr_clear(temp);
  mpfr_clear(ulp);

  return res;
}

double buildFPNumber(int sign, int expo, uint64_t mant) {
  dblcast xdb;
  double res;
  
  if (expo + BIAS >= 0) {
    xdb.i = (((uint64_t) (expo + BIAS)) << (PRECISION - 1)) | 
      ((mant >> (64 - PRECISION) & (~(((uint64_t) 1) << (PRECISION - 1)))));
  } else {
    xdb.i = ((mant >> (64 - PRECISION - (expo + BIAS)) & (~(((uint64_t) 1) << (PRECISION - 1 + (expo + BIAS))))));
  }
  res = (sign ? -xdb.d : xdb.d); 
  
  return res;
}

uint64_t randomUint64() {
  uint64_t res;
  int i;
  unsigned long int r;

  res = 0;
  for (i=0; i<8; i++) {
    res <<= 8;
    r = rand();
    res |= (uint64_t) (r & 0xff);
  }

  return res;
}

int randomIntInRange(int min, int max) {
  int res;

  do {
    res = min + ((int) (((double) (max - min + 1)) * (((double) rand()) / ((double) RAND_MAX))));
  } while ((res < min) || (res > max));
  
  return res;
}

double randomDoubleMOneOne() {
  uint64_t mant;
  int expo, sign;

  expo = randomIntInRange(EXPO_MIN, 0);
  sign = (rand() <= (RAND_MAX / 2));
  mant = randomUint64();
  return buildFPNumber(sign, expo, mant);
}

int log2int(int n) {
  int i, l;
  for (i=n, l = 0; i > 0; i >>= 1, l++);
  return l;
}

double sqrt(double);

void getNameOfTestCase(char name[25], int testCase) {
  static char testCaseName[][25] = {"FULL_EXPONENT_RANGE",  /*  0 */
                                    "NO_OVERFLOW",          /*  1 */
                                    "AROUND_ONE",           /*  2 */
                                    "REALLY_SMALL",         /*  3 */
                                    "GRADUALLY_UNDERFLOWS", /*  4 */
                                    "HALF_ULP_CASES",       /*  5 */
                                    "MONOTONE_INCREASING",  /*  6 */
                                    "MONOTONE_DECREASING",  /*  7 */
                                    "SPURIOUS_UNDERFLOW",   /*  8 */
		                    "LOSER :)",             /*  9 */
                                    "NRM_INP",              /* 10 */
                                    "SUB_SQR",              /* 11 */
                                    "SUB_INP",              /* 12 */
                                    "MED_INP",              /* 13 */
                                    "SPUR_OVF",             /* 14 */
                                    "INF_RES"               /* 15 */
                                   };
  strncpy(name, testCaseName[testCase], 25);
}

void makeRandomVector(double *x, int n, int testCase) {
  int i;
  uint64_t mant;
  int expo, sign;
  int minExpo, maxExpo;
  CONST double exactSquare = 1.16415432822132222656641431512980489060282707214355e-10; /* 2^-33 + 2^-53 - 2^-12 * 2^-53 */
  CONST double perturb = 6.617444900424221398971269536559702828526496887207e-24; /* 1/2 * 2^-33 * 2^-53 * 2^10 */
  CONST double two1022 = 4.49423283715578976932326297697256183404494244735577e307; /* 2^1022 */
  CONST double two969 = 4.9896007738367995291409317825920964151686281510886e291; /* 2^969 */
  double valsq;

  if (testCase == HALF_ULP_CASES) {
    if (n > 0) {
      x[0] = 1.0;
      for (i=1; i<n; i++) {
	valsq = exactSquare + perturb * randomDoubleMOneOne();
	x[i] = sqrt(valsq);
      }
    }
  } else {
    if (testCase == MONOTONE_INCREASING) {
      for (i=0;i<n;i++) {
	x[i] = 1.0 + ((double) i) / ((double) n);
      }
    } else {
      if (testCase == MONOTONE_DECREASING) {
	for (i=0;i<n;i++) {
	  x[i] = 1.0 + ((double) (n-i)) / ((double) n);
	}
      } else {
	if (testCase == NETLIB_SPURIOUS_UNDERFLOW) {
	  if (n > 1) {
	    x[0] = two1022;
	    x[1] = two969;
	    minExpo = EXPO_MIN;
	    maxExpo = -1;
	    for (i=2;i<n;i++) {
	      expo = randomIntInRange(minExpo, maxExpo);
	      sign = (rand() <= (RAND_MAX / 2));
	      mant = randomUint64();
	      x[i] = buildFPNumber(sign, expo, mant);
	    }
	  }
	} else {
	  switch (testCase) {
	  case NO_OVERFLOW:
	    minExpo = EXPO_MIN;
	    maxExpo = EXPO_MAX - log2int(n) - 2;
	    break;
	  case AROUND_ONE:
	    minExpo = EXPO_MIN_ONE_RANGE;
	    maxExpo = EXPO_MAX_ONE_RANGE;
	    break;
	  case REALLY_SMALL:
	    minExpo = EXPO_MIN;
	    maxExpo = EXPO_MAX_SMALL_RANGE;
	    break;
	  case NORM_GRADUALLY_UNDERFLOWS:
	    minExpo = EXPO_MIN;
	    maxExpo = EXPO_LEAST_NORMAL - log2int(n) - 2;
	    break;
	  case NRM_INP:
	    minExpo = EXPO_LEAST_NORMAL;
	    maxExpo = EXPO_MAX - PRECISION;
	    break;
	  case SUB_SQR: /* The inputs are normal but we are sure that the squares are in the subnormal range */
	    minExpo = -537; /* 2*minExpo = -1074 = EXPO_MIN          */
	    maxExpo = -512; /* 2*maxExpo = -1024 < EXPO_LEAST_NORMAL */
	    break;
	  case SUB_INP: /* All the inputs are subnormal */
	    minExpo = EXPO_MIN;
	    maxExpo = EXPO_LEAST_NORMAL-1;
	    break;
	  case MED_INP: /* All the inputs are normal fp numbers, and so are their squares */
	    minExpo = EXPO_MIN_MED;
	    maxExpo = EXPO_MAX_MED;
	    break;
	  case SPUR_OVF:
	    minExpo = 512;
	    maxExpo = EXPO_MAX - PRECISION;
	    break;
	  case INF_RES:
	    minExpo = EXPO_MAX - PRECISION;
	    maxExpo = EXPO_MAX;
	    break;
	  case FULL_EXPONENT_RANGE:
	  default:
	    minExpo = EXPO_MIN;
	    maxExpo = EXPO_MAX;
	  }
      	  for (i=0; i<n; i++) {
	    expo = randomIntInRange(minExpo, maxExpo);
	    sign = (rand() <= (RAND_MAX / 2));
	    mant = randomUint64();
	    x[i] = buildFPNumber(sign, expo, mant);
	  }
	}
      }
    }
  }
}
