import csv
import sys

from glob import iglob
from statistics import stdev
from statistics import mean as fmean


ALGOS = ["naive", "vecnaive", "netlib", "faithful", "accurate"]
SIZES = [256, 1024, 4096]
CASES = ["MED_INP", "NRM_INP", "SPUR_OVF"]


INDENT = " " * 4


##
## Data reader
##


def _skipcomments(reader):
    for row in reader:
        if row.startswith("#"):
            continue
        yield row


def _skipheader(reader):
    next(reader)
    yield from reader


def read_one_table(filename):

    with open(filename) as csvfile:
        csvfile = _skipheader(_skipcomments(csvfile))
        data = csv.reader(csvfile, skipinitialspace=True)

        for row in data:
            row = row[:-1]  ## Remove empty cell due to trailing separator ','
            row = [float(cell) for cell in row]
            yield row


def separate_columns(tablereader):

    row = next(tablereader)
    cols = [[cell] for cell in row]

    for row in tablereader:
        for i, cell in enumerate(row):
            cols[i].append(cell)

    return cols


## Data reduction

def do_stats(cols):
    for col in cols:
        yield (fmean(col), stdev(col))


def read_data(host):
    data = {}

    for case in CASES:

        ## Iterate over all found sizes.
        for filename in iglob(f"datarev1/{host}_{case}_*.csv"):

            size = filename.rsplit(".csv")[0].split("_")[-1]
            size = int(size)

            ## Read and reduce data.
            table = read_one_table(filename)
            cols = separate_columns(table)
            stats = do_stats(cols)

            ## Store data.
            for algo, stat in zip(ALGOS, stats):
                data[algo, size, case] = stat

    return data


##
## Data writer
##


fmt_size = lambda size: " & {:4d}".format(size)

STAT_SCALE = 1000.0

def fmt_stat(stat, scale=1.0):
    return " & {:4.1f}({:3.1f})".format(*(x/scale for x in stat))


ALGO_OUT = {
    "naive": "Naive",
    "vecnaive": "Naive (vec)",
    "netlib": "Hammarling",
    "faithful": "Graillat et al.",
    "accurate": "Ours",
}


HOST_OUT = {
    "ninjutsu": r"\emph{Intel Skylake} (AVX2) @3.2~GHz",
    "arunch": r"\emph{AMD Zen2} (AVX2) $\le$3.2~GHz",
    "srunch": r"\emph{Intel Skylake} (AVX512) @3.0~GHz",
    "pyxis-3.lyon.grid5000.fr": r"\emph{ARM ThunderX2} (Neon) $\le$2.2~GHz",
}


def write_data(stream, data):

    _writeln = lambda s: stream.write(INDENT * 2 + s + "\n")
    hosts = list(data.keys())

    ## Write header.
    stream.write(INDENT + r"\begin{tabular}{rr")
    for host in hosts:
        stream.write(r"@{\hspace{4\tabcolsep}}rrr")
    stream.write("}\n")
    _writeln(r"\toprule")
    _writeln(r"&")
    stream.write(INDENT * 2)
    for host in hosts:
        stream.write(r"& \multicolumn{3}{c}{"+HOST_OUT[host]+"}")
    stream.write(r" \\" + "\n")
    _writeln(r"\cmidrule(r{4\tabcolsep}){3-5}\cmidrule{6-8}")
    stream.write(INDENT * 2)
    stream.write(r"Algorithm & {$n$}")
    for host in hosts:
        stream.write(r" & {\sf MED\_INP} & {\sf NRM\_INP} & {\sf SPUR\_OVR}")
    stream.write(r" \\" + "\n")

    for algo in ALGOS:

        ## Insert rule between lines.
        if algo != ALGOS[0]:
            _writeln(r"\midrule")
        else:
            _writeln(r"\midrule[\heavyrulewidth]")

        for size in SIZES:

            ## Print algo only the first time to avoid repetition.
            algo_str = ALGO_OUT[algo]
            algo_str = algo_str if size == SIZES[0] else " " * len(algo_str)
            stream.write(INDENT * 2 + algo_str)

            ## Write size.
            stream.write(fmt_size(size))

            ## Write data.
            for host in hosts:

                for case in CASES:

                    if (algo, size, case) in data[host]:

                        ## Write formatted data.
                        stat = data[host][algo, size, case]
                        stream.write(fmt_stat(stat, STAT_SCALE))

                    else:

                        ## Write an empty entry if data is missing.
                        stream.write(r" & &")

            stream.write(r" \\" + "\n")

    # Write footer.
    _writeln(r"\bottomrule")
    stream.write(INDENT + r"\end{tabular}" + "\n")

##
##
##


def main():

    data = dict()

    for host in sys.argv[1:]:
        print(f"host={host}", file=sys.stderr)
        data[host] = read_data(host)

    write_data(sys.stdout, data)

    return 0


if __name__ == "__main__":
    sys.exit(main())
