/* This program is part of a correctly-rounded, division-free l2
   vector norm with avoidance of spurious underflow and overflow.

   Author: Nicolas Louvet,

           Université Claude Bernard Lyon 1

   implementing the algorithm described in the paper

   Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
   Laurence Rideau. Accurate calculation of Euclidean Norms using
   Double-word arithmetic. 2021

   This program is

   Copyright (C) 2021 Université Claude Bernard Lyon 1

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <http://www.gnu.org/licenses/>.

*/

#include <sys/stat.h>
#include <sys/utsname.h>
#include <sys/time.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <unistd.h>

#include "randomvectors.h"
#include "faithfulnormmpfr.h"
#include "norms.h"

#include "readtime.h"

/* Some symbolic constants */
#define FULL_EXPONENT_RANGE        0
#define NO_OVERFLOW                1
#define AROUND_ONE                 2
#define REALLY_SMALL               3
#define NORM_GRADUALLY_UNDERFLOWS  4
#define HALF_ULP_CASES             5
#define MONOTONE_INCREASING        6
#define MONOTONE_DECREASING        7
#define NETLIB_SPURIOUS_UNDERFLOW  8

void forget(volatile double *var) {}

uint64_t timeNorm(double (*norm)(CONST double * RESTRICT, unsigned int), CONST double * RESTRICT x, unsigned int n, int runs) {
  int i;
  uint64_t before, after, time, overallTime;
  double res;
  volatile double sink;
  
  overallTime = 0;
  for (i=0; i<runs; i++) {
    do {
      READ_TIME_COUNTER(&before);
      res = norm(x, n);
      READ_TIME_COUNTER(&after);
      sink = res;
    } while (before >= after);
    time = after - before;
    overallTime += time;
  }
  forget(&sink);
  return overallTime / runs;
}

uint64_t timeCallOverHead(double (*norm)(CONST double * RESTRICT, unsigned int), CONST double * RESTRICT x, unsigned int n, int runs) {
  int i;
  uint64_t before, after, time, minTime;
  double res;
  volatile double sink;
  
  minTime = -1ull;
  for (i=0; i<runs; i++) {
    do {
      READ_TIME_COUNTER(&before);
      res = norm(x, n);
      READ_TIME_COUNTER(&after);
      sink = res;
    } while (before >= after);
    time = after - before;
    if(time < minTime) minTime = time;
  }
  forget(&sink);
  return minTime;
}

void print(FILE *fd, const char * pString, ...){
    va_list args;
    va_start(args, pString);
    vprintf(pString, args);
    va_end(args);
    va_start(args, pString);
    vfprintf(fd, pString, args);
    va_end(args);
}

void printUsage(char *str) {
  fprintf(stderr, "Usage: %s <vector length> <number of tests>\n", str);
  fprintf(stderr, "\n");
}

int readArg(int *r, char *str, int min) {
  int res;
  char *end;
  
  if (str[0] == '\0') return 0;
  res = strtol(str, &end, 10);
  if (end[0] != '\0') return 0;

  if (res < min) return 0;

  *r = res;
  return 1;
}

void doTimeNorm(int testCase, unsigned int n, int k) {
  char *name[] = {"naive", "vecnaive", "netlib", "faithful", "accurate" };
  double (*norm[])(const double *, unsigned int) = {naiveNorm, vecNaiveNorm, netlibNorm, faithfulNorm, accurateNorm};
  
  char testCaseName[25];
  getNameOfTestCase(testCaseName, testCase);
  
  const int nbnorms = 5;
  double *x, *timing[nbnorms];
  
  uint64_t callOverHead;
  double avgTime, stdDev;
  
  struct utsname buffer;
  char filename[128];
  FILE *fd;
  int i, j;

  if(uname(&buffer) == 0)
    snprintf(filename, 128, "datarev1/%s_%s_%d.csv", buffer.nodename, testCaseName, n);
  else
    snprintf(filename, 128, "datarev1/%s_%s_%d.csv", "unknown", testCaseName, n);

  for(j = 0; j < nbnorms; j++) timing[j] = malloc(k * sizeof(double));
  x = (double *)malloc(n * sizeof(double));
  
  callOverHead = timeCallOverHead(dummyFunc, x, n, 10*k);

  struct stat st = {0};
  if (stat("data", &st) == -1) {
      mkdir("data", 0700);
  }

  fd = fopen(filename, "w+");
  print(fd, "# --------------------------------------------------\n");
  print(fd, "# filename     = %s\n", filename);
  print(fd, "# testCaseName = %s\n", testCaseName);
  print(fd, "# length       = %d\n", n);
  print(fd, "# num samples  = %d\n", k);
  if(uname(&buffer) == 0) {
    print(fd, "# node name    = %s\n", buffer.nodename);
    print(fd, "# system name  = %s\n", buffer.sysname);
    print(fd, "# release      = %s\n", buffer.release);
    print(fd, "# version      = %s\n", buffer.version);
    print(fd, "# machine      = %s\n", buffer.machine);
  }
  print(fd, "# callOverHead = %5.1f\n", (double)callOverHead);
  print(fd, "# columns      = ");
  for(j = 0; j < nbnorms; j++) print(fd, "%s ", name[j]);
  print(fd, "\n");
  fclose(fd);
  
  for (i = 0; i < k; i++) {
    makeRandomVector(x, n, testCase);
    for(j = 0; j < nbnorms; j++)
      timing[j][i] = (double)(timeNorm(norm[j], x, n, 1) - callOverHead);
  }

  fd = fopen(filename, "a");
  for(j = 0; j < nbnorms; j++) {
    avgTime = 0.0;
    for(i = 0; i < k; i++) avgTime += timing[j][i];
    avgTime /= (double) k;
    stdDev = 0.0;
    for(i = 0; i < k; i++) stdDev += (timing[j][i] - avgTime) * (timing[j][i] - avgTime);
    stdDev = sqrt(stdDev / k);
    print(fd, "# %13s -> %10.1f (%8.1f)\n", name[j], avgTime, stdDev);
  }
  for(j = 0; j < nbnorms; j++) fprintf(fd, "%15s, ", name[j]);
  fprintf(fd, "\n");
  for (i = 0; i < k; i++) {
    for(j = 0; j < nbnorms; j++) fprintf(fd, "%15.1f, ", timing[j][i]);
    fprintf(fd, "\n");
  }
  fclose(fd);

  for(j = 0; j < nbnorms; j++) free(timing[j]);
  free(x);
}

int main(int argc, char **argv) {
  int k;
  unsigned int n;
  int sn;

  if (argc < 3) {
    printUsage(argv[0]);
    return 2;
  }

  if (!readArg(&sn, argv[1], 1)) {
    printUsage(argv[0]);
    return 2;
  }
  n = sn;

  if (!readArg(&k, argv[2], 1)) {
    printUsage(argv[0]);
    return 2;
  }

  doTimeNorm(AROUND_ONE, n, k);
  doTimeNorm(NRM_INP, n, k);
  doTimeNorm(MED_INP, n, k);
  doTimeNorm(SUB_SQR, n, k);
  doTimeNorm(SUB_INP, n, k);
  doTimeNorm(SPUR_OVF, n, k);
  doTimeNorm(INF_RES, n, k);

  return 0;
}
