/* This program is part of a correctly-rounded, division-free l2
   vector norm with avoidance of spurious underflow and overflow.

   Author: Nicolas Louvet,

           Université Claude Bernard Lyon 1

   implementing the algorithm described in the paper

   Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
   Laurence Rideau. Accurate calculation of Euclidean Norms using
   Double-word arithmetic. 2021

   This program is

   Copyright (C) 2021 Université Claude Bernard Lyon 1

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <http://www.gnu.org/licenses/>.

*/

///////////////
#ifndef __VEC__
#define __VEC__
///////////////

/*
inline vec_t vec_load(const double *p); // returns the vector starting at address p (p does not have to be aligned)
inline void vec_store(double *p, vec_t v); // stores v at address p (p does not have to be aligned)

inline vec_t vec_load_algn(const double *p); // returns the vector starting at address p (p must be aligned)
inline void vec_store_algn(double *p, vec_t v); // stores v at address p (p must be aligned)

inline vec_t vec_set1(double x); // returns a vector whose all components equal to x
inline vec_t vec_setzero(void); // returns a vector whose all components equal to 0.0

inline vec_t vec_add(vec_t v1, vec_t v2); // returns v1 + v2 componentwise
inline vec_t vec_sub(vec_t v1, vec_t v2); // returns v1 - v2 componentwise
inline vec_t vec_mul(vec_t v1, vec_t v2); // returns v1 * v2 componentwise
inline vec_t vec_div(vec_t v1, vec_t v2); // returns v1 / v2 componentwise

inline vec_t vec_fma(vec_t v1, vec_t v2, vec_t v3); // returns v1 * v2 + v3 componentwise
inline vec_t vec_fms(vec_t v1, vec_t v2, vec_t v3); // returns v1 * v2 - v3 componentwise

inline vec_t vec_abs(vec_t v); // returns fabs(v) componentwise

inline vec_t vec_or(vec_t v1, vec_t v2); // returns v1 or v2 componentwise

inline mask_t mask_setzero(void); // mask whose all components are 0

inline mask_t mask_not(mask_t v); // returns not(v) componentwise
inline mask_t mask_and(mask_t v1, mask_t v2); // returns v1 and v2 componentwise
inline mask_t mask_or(mask_t v1, mask_t v2);  // returns v1 or v2 componentwise
inline mask_t mask_xor(mask_t v1, mask_t v2); // returns v1 xor v2 componentwise
inline mask_t mask_andnot(mask_t v1, mask_t v2); // returns not(v1) and v2 componentwise

inline void mask_store(int mtab[vec_len], mask_t m); // stores the components of m in mtab

inline mask_t vec_cmpge(vec_t a, vec_t b); // returns a >= b componentwise 
inline mask_t vec_cmplt(vec_t a, vec_t b); // returns a < b componentwise 

inline vec_t vec_select_vorz(mask_t m, vec_t v); // returns (m ? v : 0.0) componentwise
inline vec_t vec_select_zorv(mask_t m, vec_t v); // returns (m ? 0.0 : v) componentwise
inline vec_t vec_select(mask_t m, vec_t v1, vec_t v2); // returns (m ? v1 : v2) componentwise

inline void vec_twosqr(vec_t *p_hi, vec_t *p_lo, vec_t a);
inline void vec_twosum(vec_t *s_hi, vec_t *s_lo, vec_t a, vec_t b);
inline void vec_fast_twosum(vec_t *s_hi, vec_t *s_lo, vec_t a, vec_t b);
*/

////////////////////////
#if defined(__AVX512F__)
////////////////////////

#pragma message("Using AVX512")

#include <immintrin.h>

typedef __m512d vec_t;
typedef __mmask8 mask_t;

#define vec_len   8
#define vec_algn 64
#define VEC_ALGN __attribute__ ((aligned(vec_algn)))

inline vec_t vec_load(const double *p) { return _mm512_loadu_pd(p); }
inline void vec_store(double *p, vec_t v) { _mm512_storeu_pd(p, v); }
inline vec_t vec_load_algn(const double *p) { return _mm512_load_pd(p); }
inline void vec_store_algn(double *p, vec_t v) { _mm512_store_pd(p, v); }

inline vec_t vec_set1(double x) { return _mm512_set1_pd(x); }
inline vec_t vec_setzero(void) { return _mm512_setzero_pd(); }

inline vec_t vec_set1int64(long long v) { return _mm512_castsi512_pd(_mm512_set1_epi64(v)); }

inline vec_t vec_add(vec_t v1, vec_t v2) { return _mm512_add_pd(v1, v2); }
inline vec_t vec_sub(vec_t v1, vec_t v2) { return _mm512_sub_pd(v1, v2); }
inline vec_t vec_mul(vec_t v1, vec_t v2) { return _mm512_mul_pd(v1, v2); }
inline vec_t vec_div(vec_t v1, vec_t v2) { return _mm512_div_pd(v1, v2); }

inline vec_t vec_fma(vec_t v1, vec_t v2, vec_t v3) { return _mm512_fmadd_pd(v1, v2, v3); }
inline vec_t vec_fms(vec_t v1, vec_t v2, vec_t v3) { return _mm512_fmsub_pd(v1, v2, v3); }

inline vec_t vec_abs(vec_t a) { return _mm512_abs_pd(a); }
inline vec_t vec_max(vec_t v1, vec_t v2) {  return _mm512_max_pd(v1, v2); }
inline vec_t vec_min(vec_t v1, vec_t v2) {  return _mm512_min_pd(v1, v2); }
inline vec_t vec_maxabs(vec_t v1, vec_t v2) { return vec_max(vec_abs(v1), vec_abs(v2)); }
inline vec_t vec_minabs(vec_t v1, vec_t v2) { return vec_min(vec_abs(v1), vec_abs(v2)); }
//line vec_t vec_maxabs(vec_t v1, vec_t v2) {  return _mm512_range_pd(v1, v2, 0b10); }

inline vec_t vec_or(vec_t v1, vec_t v2) { return _mm512_or_pd(v1, v2); }

inline mask_t mask_setzero(void) { return _cvtu32_mask8(0u); }
inline mask_t mask_not(mask_t m) { return _knot_mask8(m); }
inline mask_t mask_and(mask_t m, mask_t n) { return _kand_mask8(m, n); }
inline mask_t mask_or(mask_t m, mask_t n) { return _kor_mask8(m, n); }
inline mask_t mask_xor(mask_t m, mask_t n) { return _kxor_mask8(m, n); }
inline mask_t mask_andnot(mask_t m, mask_t n) { return _kandn_mask8(m, n); }

inline void mask_store(int mtab[vec_len], mask_t m) {
  unsigned int n = _cvtmask8_u32(m);
  for(int i = 0; i < vec_len; n >>= 1, i++) mtab[i] = (n & 1u);
}

inline mask_t vec_cmpge(vec_t a, vec_t b) { return _mm512_cmp_pd_mask(a, b, _CMP_GE_OS); }
inline mask_t vec_cmplt(vec_t a, vec_t b) { return _mm512_cmp_pd_mask(a, b, _CMP_LT_OS); }

inline vec_t vec_select_vorz(mask_t m, vec_t a) { return _mm512_mask_blend_pd(m, _mm512_setzero_pd(), a); }
inline vec_t vec_select_zorv(mask_t m, vec_t a) { return _mm512_mask_blend_pd(m, a, _mm512_setzero_pd()); }
inline vec_t vec_select(mask_t m, vec_t a, vec_t b) { return _mm512_mask_blend_pd(m, b, a); }

inline void vec_twosqr(vec_t *p_hi, vec_t *p_lo, vec_t a) {
  *p_hi = _mm512_mul_pd(a, a);
  *p_lo = _mm512_fmsub_pd(a, a, *p_hi);
}

inline void vec_twosum(vec_t *s_hi, vec_t *s_lo, vec_t a, vec_t b) {
  *s_hi = _mm512_add_pd(a, b);
  __m512i sgnmask = _mm512_set1_epi64(0x7fffffffffffffffLL);
  __m512d abs1 = _mm512_castsi512_pd(_mm512_and_epi64(_mm512_castpd_si512(a), sgnmask));
  __m512d abs2 = _mm512_castsi512_pd(_mm512_and_epi64(_mm512_castpd_si512(a), sgnmask));
  __mmask8 mask = _mm512_cmp_pd_mask(abs1, abs2, _CMP_GE_OS);
  __m512d maxabs = _mm512_mask_blend_pd(mask, b, a);
  __m512d minabs = _mm512_mask_blend_pd(mask, a, b);
  *s_lo = _mm512_add_pd(_mm512_sub_pd(maxabs, *s_hi), minabs);
}

inline void vec_fast_twosum(vec_t *s_hi, vec_t *s_lo, vec_t a, vec_t b) {
  *s_hi = _mm512_add_pd(a, b);
  *s_lo = _mm512_sub_pd(b, _mm512_sub_pd(*s_hi, a));
}

///////////////////////
#elif defined(__AVX2__)
///////////////////////

#pragma message("Using AVX2")

#include <immintrin.h>

typedef __m256d vec_t;
typedef __m256d mask_t;

#define vec_len   4
#define vec_algn 32
#define VEC_ALGN __attribute__ ((aligned(vec_algn)))

inline vec_t vec_set1int64(long long v) {
  return _mm256_castsi256_pd(_mm256_set1_epi64x(v));
}

inline vec_t vec_load(const double *p) { return _mm256_loadu_pd(p); }
inline void vec_store(double *p, vec_t v) { _mm256_storeu_pd(p, v); }

inline vec_t vec_load_algn(const double *p) { return _mm256_load_pd(p); }
inline void vec_store_algn(double *p, vec_t v) { _mm256_store_pd(p, v); }

inline vec_t vec_set1(double x) { return _mm256_set1_pd(x); }
inline vec_t vec_setzero(void) { return _mm256_setzero_pd(); }

inline vec_t vec_add(vec_t v1, vec_t v2) { return _mm256_add_pd(v1, v2); }
inline vec_t vec_sub(vec_t v1, vec_t v2) { return _mm256_sub_pd(v1, v2); }
inline vec_t vec_mul(vec_t v1, vec_t v2) { return _mm256_mul_pd(v1, v2); }
inline vec_t vec_div(vec_t v1, vec_t v2) { return _mm256_div_pd(v1, v2); }

inline vec_t vec_fma(vec_t v1, vec_t v2, vec_t v3) { return _mm256_fmadd_pd(v1, v2, v3); }
inline vec_t vec_fms(vec_t v1, vec_t v2, vec_t v3) { return _mm256_fmsub_pd(v1, v2, v3); }

inline vec_t vec_abs(vec_t a) { return _mm256_and_pd(a, _mm256_castsi256_pd(_mm256_set1_epi64x(0x7fffffffffffffffull))); }
inline vec_t vec_max(vec_t v1, vec_t v2) { return _mm256_max_pd(v1, v2); }
inline vec_t vec_min(vec_t v1, vec_t v2) { return _mm256_min_pd(v1, v2); }
inline vec_t vec_maxabs(vec_t v1, vec_t v2) { return vec_max(vec_abs(v1), vec_abs(v2)); }
inline vec_t vec_minabs(vec_t v1, vec_t v2) { return vec_min(vec_abs(v1), vec_abs(v2)); }

inline vec_t vec_or(vec_t v1, vec_t v2) { return _mm256_or_pd(v1, v2); }

inline mask_t mask_setzero(void) { return _mm256_set1_pd(0.0); }

inline mask_t mask_not(mask_t a) { return _mm256_xor_pd(a, _mm256_castsi256_pd(_mm256_set1_epi64x(0xffffffffffffffffull))); }
inline mask_t mask_and(mask_t v1, mask_t v2) { return _mm256_and_pd(v1, v2); }
inline mask_t mask_or(mask_t v1, mask_t v2) { return _mm256_or_pd(v1, v2); }
inline mask_t mask_xor(mask_t v1, mask_t v2) { return _mm256_xor_pd(v1, v2); }
inline mask_t mask_andnot(mask_t v1, mask_t v2) { return _mm256_andnot_pd(v1, v2); }

inline void mask_store(int mtab[vec_len], mask_t m) {
  double __attribute__ ((aligned(vec_algn))) t[vec_len];
  _mm256_storeu_pd(t, m);
  for(int i = 0; i < vec_len; i++) mtab[i] = (t[i] != 0.0);
}

inline mask_t vec_cmpge(vec_t a, vec_t b) { return _mm256_cmp_pd(a, b, _CMP_GE_OS); }
inline mask_t vec_cmplt(vec_t a, vec_t b) { return _mm256_cmp_pd(a, b, _CMP_LT_OS); }

inline vec_t vec_select_vorz(mask_t m, vec_t a) { return _mm256_and_pd(m, a); }
inline vec_t vec_select_zorv(mask_t m, vec_t a) { return _mm256_andnot_pd(m, a); }
inline vec_t vec_select(mask_t m, vec_t a, vec_t b) { return _mm256_blendv_pd(b, a, m); }

inline void vec_twosqr(vec_t *p_hi, vec_t *p_lo, vec_t a) {
  *p_hi = _mm256_mul_pd(a, a);
  *p_lo = _mm256_fmsub_pd(a, a, *p_hi);
}

inline void vec_twosum(vec_t *s_hi, vec_t *s_lo, vec_t a, vec_t b) {
  *s_hi = _mm256_add_pd(a, b);
  __m256d vmin = _mm256_min_pd(a, b);
  __m256d vmax = _mm256_max_pd(a, b);
  __m256d vminmag = _mm256_blendv_pd(vmin, vmax, *s_hi);
  __m256d vmaxmag = _mm256_blendv_pd(vmax, vmin, *s_hi);
  __m256d vt = _mm256_sub_pd(*s_hi, vmaxmag);
  *s_lo = _mm256_sub_pd(vminmag, vt);
}

inline void vec_fast_twosum(vec_t *s_hi, vec_t *s_lo, vec_t a, vec_t b) {
    *s_hi = _mm256_add_pd(a, b);
    *s_lo = _mm256_sub_pd(b, _mm256_sub_pd(*s_hi, a));
}

/////////////////////////////////////////////////////////////////////
#elif defined(__SSE2__) && defined(__SSE4_1__) && defined(__SSE4_2__)
/////////////////////////////////////////////////////////////////////

#pragma message("Using SSE")

#include <immintrin.h>

typedef __m128d vec_t;
typedef __m128d mask_t;

#define vec_len   2
#define vec_algn 16
#define VEC_ALGN __attribute__ ((aligned(vec_algn)))

inline vec_t vec_load(const double *p) { return _mm_loadu_pd(p); }
inline void vec_store(double *p, vec_t v) { _mm_storeu_pd(p, v); }

inline vec_t vec_load_algn(const double *p) { return _mm_load_pd(p); }
inline void vec_store_algn(double *p, vec_t v) { _mm_store_pd(p, v); }

inline vec_t vec_set1(double x) { return _mm_set1_pd(x); }
inline vec_t vec_setzero(void) { return _mm_setzero_pd(); }

inline vec_t vec_add(vec_t v1, vec_t v2) { return _mm_add_pd(v1, v2); }
inline vec_t vec_sub(vec_t v1, vec_t v2) { return _mm_sub_pd(v1, v2); }
inline vec_t vec_mul(vec_t v1, vec_t v2) { return _mm_mul_pd(v1, v2); }
inline vec_t vec_div(vec_t v1, vec_t v2) { return _mm_div_pd(v1, v2); }

inline vec_t vec_abs(vec_t a) { return _mm_and_pd(a, _mm_castsi128_pd(_mm_set1_epi64x(0x7fffffffffffffffull))); }
inline vec_t vec_max(vec_t v1, vec_t v2) { return _mm128_max_pd(v1, v2); }
inline vec_t vec_min(vec_t v1, vec_t v2) { return _mm128_min_pd(v1, v2); }
inline vec_t vec_maxabs(vec_t v1, vec_t v2) { return vec_max(vec_abs(v1), vec_abs(v2)); }
inline vec_t vec_minabs(vec_t v1, vec_t v2) { return vec_min(vec_abs(v1), vec_abs(v2)); }

inline vec_t vec_or(vec_t v1, vec_t v2) { return _mm_or_pd(v1, v2); }

// inline vec_t vec_max(vec_t v1, vec_t v2) {  }

inline vec_t vec_set1int64(long long v) { return _mm_castsi128_pd(_mm_set1_epi64x(v)); }

inline mask_t mask_setzero(void) { return _mm_set1_pd(0.0); }
inline mask_t mask_not(mask_t a) { return _mm_xor_pd(a, _mm_castsi128_pd(_mm_set1_epi64x(0xffffffffffffffffull))); }

inline vec_t mask_and(vec_t v1, vec_t v2) { return _mm_and_pd(v1, v2); }
inline vec_t mask_or(vec_t v1, vec_t v2) { return _mm_or_pd(v1, v2); }
inline vec_t mask_xor(vec_t v1, vec_t v2) { return _mm_xor_pd(v1, v2); }
inline vec_t mask_andnot(vec_t v1, vec_t v2) { return _mm_andnot_pd(v1, v2); }

inline void mask_store(int mtab[vec_len], vec_t m) {
  double __attribute__ ((aligned(vec_algn))) t[vec_len];
  _mm_storeu_pd(t, m);
  for(int i = 0; i < vec_len; i++) mtab[i] = (t[i] != 0.0);
}

inline mask_t vec_cmpge(vec_t a, vec_t b) { return _mm_cmpge_pd(a, b); }
inline mask_t vec_cmplt(vec_t a, vec_t b) { return _mm_cmplt_pd(a, b); }

inline vec_t vec_select_vorz(mask_t m, vec_t a) { return _mm_and_pd(m, a); }
inline vec_t vec_select_zorv(mask_t m, vec_t a) { return _mm_andnot_pd(m, a); }
inline vec_t vec_select(mask_t m, vec_t a, vec_t b) { return _mm_or_pd(_mm_and_pd(m, a), _mm_andnot_pd(m, b)); }

static inline void vec_twosqr(vec_t *p_hi, vec_t *p_lo, vec_t a) {
  __m128d splitter = _mm_set1_pd(134217729.0);
  __m128d a_hi, a_lo;
  __m128d t1, t2, t3;

  /* Veltkamp splitting */
  t1 = _mm_mul_pd(a, splitter);
  a_hi = _mm_add_pd(_mm_sub_pd(a, t1), t1);
  a_lo = _mm_sub_pd(a, a_hi);
  /* Dekker's algorithm */
  *p_hi = _mm_mul_pd(a, a);
  t1 = _mm_mul_pd(a_hi, a_hi);
  t2 = _mm_mul_pd(a_hi, a_lo);
  t3 = _mm_mul_pd(a_lo, a_lo);
  *p_lo = _mm_add_pd(_mm_add_pd(_mm_add_pd(_mm_sub_pd(t1, *p_hi), t2), t2), t3);
}

static inline void vec_twosum(vec_t *s_hi, vec_t *s_lo, vec_t a, vec_t b) {
    __m128d t1, t2, t3, t4;
    *s_hi = _mm_add_pd(a, b);
    t1 = _mm_sub_pd(*s_hi, a);
    t2 = _mm_sub_pd(*s_hi, t1);
    t3 = _mm_sub_pd(b, t1);
    t4 = _mm_sub_pd(a, t2);
    *s_lo = _mm_add_pd(t3, t4);
}

static inline void vec_fast_twosum(vec_t *s_hi, vec_t *s_lo, vec_t a, vec_t b) {
    *s_hi = _mm_add_pd(a, b);
    *s_lo = _mm_sub_pd(b, _mm_sub_pd(*s_hi, a));
}

///////////////////////
#elif defined(__ARM_NEON)
///////////////////////

#pragma message("Using Arm NEON")

#include <arm_neon.h>

typedef float64x2_t vec_t;
typedef uint64x2_t mask_t;

#define vec_len   2
#define vec_algn 16
#define VEC_ALGN __attribute__ ((aligned(vec_algn)))

inline vec_t vec_load(const double *p) { return vld1q_f64(p); }
inline void vec_store(double *p, vec_t v) { vst1q_f64(p, v); }

inline vec_t vec_load_algn(const double *p) { return vld1q_f64(p); }
inline void vec_store_algn(double *p, vec_t v) { vst1q_f64(p, v); }

//inline vec_t vec_set1(double x) { r = vld1q_dup_f64(&x); }
inline vec_t vec_set1(double x) { return vdupq_n_f64(x); }
inline vec_t vec_setzero(void) { return vdupq_n_f64(0.0); }

inline vec_t vec_add(vec_t v1, vec_t v2) { return vaddq_f64(v1, v2); }
inline vec_t vec_sub(vec_t v1, vec_t v2) { return vsubq_f64(v1, v2); }
inline vec_t vec_mul(vec_t v1, vec_t v2) { return vmulq_f64(v1, v2); }
inline vec_t vec_div(vec_t v1, vec_t v2) { return vdivq_f64(v1, v2); }

inline vec_t vec_fma(vec_t v1, vec_t v2, vec_t v3) { return vfmaq_f64(v3, v1, v2); }
inline vec_t vec_fms(vec_t v1, vec_t v2, vec_t v3) { return vnegq_f64(vfmsq_f64(v3, v1, v2)); }

inline vec_t vec_abs(vec_t v) { return vabsq_f64(v); }

inline vec_t vec_or(vec_t v1, vec_t v2) { return vreinterpretq_f64_u64(vorrq_u64(vreinterpretq_u64_f64(v1), vreinterpretq_u64_f64(v2))); }

// inline vec_t vec_max(vec_t v1, vec_t v2) {  }

inline mask_t mask_setzero(void) { return vdupq_n_u64(0ull); }

inline void mask_store(int mtab[vec_len], mask_t m) {
  uint64_t t[vec_len];
  vst1q_u64(t, m);
  for(int i = 0; i < vec_len; i++)
    mtab[i] = (t[i] != 0ull);
}

inline mask_t mask_not(mask_t v1) { return veorq_u64(v1, vdupq_n_u64(0xffffffffffffffffull)); }
inline mask_t mask_and(mask_t v1, mask_t v2) { return vandq_u64(v1, v2); }
inline mask_t mask_or(mask_t v1, mask_t v2) { return vorrq_u64(v1, v2); }
inline mask_t mask_xor(mask_t v1, mask_t v2) { return veorq_u64(v1, v2); }
inline mask_t mask_andnot(mask_t v1, mask_t v2) { return vandq_u64(vornq_u64(vdupq_n_u64(0ull), v1), v2); }

inline mask_t vec_cmpge(vec_t v1, vec_t v2) { return vcgeq_f64(v1, v2); }
inline mask_t vec_cmplt(vec_t v1, vec_t v2) { return vcltq_f64(v1, v2); }

inline vec_t vec_select(mask_t m, vec_t v1, vec_t v2) { return vbslq_f64(m, v1, v2); }
inline vec_t vec_select_vorz(mask_t m, vec_t v) { return vbslq_f64(m, v, vdupq_n_f64(0.0)); }
inline vec_t vec_select_zorv(mask_t m, vec_t v) { return vbslq_f64(m, vdupq_n_f64(0.0), v); }

inline void vec_twosqr(vec_t *p_hi, vec_t *p_lo, vec_t v) {
  *p_hi = vmulq_f64(v, v);
  *p_lo = vnegq_f64(vfmsq_f64(*p_hi, v, v));;
}

inline void vec_twosum(vec_t *s_hi, vec_t *s_lo, vec_t v1, vec_t v2) {
  *s_hi = vaddq_f64(v1, v2);
  float64x2_t maxabs, minabs;
  uint64x2_t tau = vcageq_f64(v1, v2);
  maxabs = vbslq_f64(tau, v1, v2);
  minabs = vbslq_f64(tau, v2, v1);
  *s_lo = vaddq_f64(vsubq_f64(maxabs, *s_hi), minabs);
}

inline void vec_fast_twosum(vec_t *s_hi, vec_t *s_lo, vec_t v1, vec_t v2) {
  *s_hi = vaddq_f64(v1, v2);
  *s_lo = vsubq_f64(v2, vsubq_f64(*s_hi, v1));
}

/////
#else
/////


#define VEC_ALIGN_CST 1

#error "All of this won't work with a non-simd isa..."

//////
#endif
//////

////////////////////////
#endif //#ifndef __VEC__
////////////////////////














/*
typedef __m512d vec_t;
typedef __m512d mask_t;

#define vec_len   8
#define vec_algn 64

_LDST(double, 512, pd)
_SET1(double, 512, pd)

inline vec_t vec_set1int64(long long v) {
  return _mm512_castsi512_pd(_mm512_set1_epi64(v));
}

_SETZERO(512, pd)
_ARITH(double, 512, pd)
_LOGIC(double, 512, pd)

static inline __m512d mask_not(__m512d a) { return _mm512_xor_pd(a, _mm512_castsi512_pd(_mm512_set1_epi64(0xffffffffffffffffull))); }

static inline __m512d vec_abs(__m512d a) { return _mm512_and_pd(a, _mm512_castsi512_pd(_mm512_set1_epi64(0x7fffffffffffffffull))); }

_FMA(double, 512, pd)

static inline __m512d vec_cmpge(__m512d a, __m512d b) {
  __m512d ones = _mm512_castsi512_pd(_mm512_set1_epi64(0xffffffffffffffffLL));
  __mmask8 m = _mm512_cmp_pd_mask(a, b, _CMP_GE_OS);
  return _mm512_maskz_expand_pd(m, ones);
}

static inline __m512d vec_cmplt(__m512d a, __m512d b) {
  __m512d ones = _mm512_castsi512_pd(_mm512_set1_epi64(0xffffffffffffffffLL));
  __mmask8 m = _mm512_cmp_pd_mask(a, b, _CMP_LT_OS);
  return _mm512_maskz_expand_pd(m, ones);
}

static inline __m512d vec_select_vorz(__m512d m, __m512d a) { return _mm512_and_pd(m, a); }
static inline __m512d vec_select_zorv(__m512d m, __m512d a) { return _mm512_andnot_pd(m, a); }
static inline __m512d vec_select(__m512d m, __m512d a, __m512d b) { return _mm512_or_pd(_mm512_and_pd(m, a), _mm512_andnot_pd(m, b)); }

static inline void vec_twosqr(__m512d *p_hi, __m512d *p_lo, __m512d a) {
  *p_hi = _mm512_mul_pd(a, a);
  *p_lo = _mm512_fmsub_pd(a, a, *p_hi);
}

static inline void vec_twosum(__m512d *s_hi, __m512d *s_lo, __m512d a, __m512d b) {
  *s_hi = _mm512_add_pd(a, b);
  __m512i sgnmask = _mm512_set1_epi64(0x7fffffffffffffffLL);
  __m512d abs1 = _mm512_castsi512_pd(_mm512_and_epi64(_mm512_castpd_si512(a), sgnmask));
  __m512d abs2 = _mm512_castsi512_pd(_mm512_and_epi64(_mm512_castpd_si512(a), sgnmask));
  __mmask8 mask = _mm512_cmp_pd_mask(abs1, abs2, _CMP_GE_OS);
  __m512d maxabs = _mm512_mask_blend_pd(mask, b, a);
  __m512d minabs = _mm512_mask_blend_pd(mask, a, b);
  *s_lo = _mm512_add_pd(_mm512_sub_pd(maxabs, *s_hi), minabs);
}

static inline void vec_fast_twosum(__m512d *s_hi, __m512d *s_lo, __m512d a, __m512d b) {
  *s_hi = _mm512_add_pd(a, b);
  *s_lo = _mm512_sub_pd(b, _mm512_sub_pd(*s_hi, a));
}
*/

/*
// ST scalar type manipulated: float or double
// BL bit length of the vectors: void for sse2, 256 for AVX2
// IT instruction type: ps or pd

#define _LDST(ST, BL, IT) \
inline vec_t vec_load(const ST *p) { return _mm ## BL ## _loadu_ ## IT(p); }        \
inline void vec_store(ST *p, const vec_t v) { _mm ## BL ## _storeu_ ## IT (p, v); }

#define _LDST_ALGN(ST, BL, IT) \
inline vec_t vec_load_algn(const ST *p) { return _mm ## BL ## _load_ ## IT(p); }        \
inline void vec_store_algn(ST *p, const vec_t v) { _mm ## BL ## _store_ ## IT (p, v); }

#define _SET1(ST, BL, IT) \
inline vec_t vec_set1(ST x) { return _mm ## BL ## _set1_ ## IT(x); }

#define _SETZERO(BL, IT) \
inline vec_t vec_setzero(void) { return _mm ## BL ## _setzero_ ## IT(); }

#define _ARITH(ST, BL, IT) \
inline vec_t vec_add(vec_t v1, vec_t v2) { return _mm ## BL ## _add_ ## IT(v1, v2); } \
inline vec_t vec_sub(vec_t v1, vec_t v2) { return _mm ## BL ## _sub_ ## IT(v1, v2); } \
inline vec_t vec_mul(vec_t v1, vec_t v2) { return _mm ## BL ## _mul_ ## IT(v1, v2); } \
inline vec_t vec_div(vec_t v1, vec_t v2) { return _mm ## BL ## _div_ ## IT(v1, v2); }

#define _MASK(ST, BL, IT) \
inline vec_t mask_and(vec_t v1, vec_t v2) { return _mm ## BL ## _and_ ## IT(v1, v2); }       \
inline vec_t mask_or(vec_t v1, vec_t v2) { return _mm ## BL ## _or_ ## IT(v1, v2); }         \
inline vec_t mask_xor(vec_t v1, vec_t v2) { return _mm ## BL ## _xor_ ## IT(v1, v2); }       \
inline vec_t mask_andnot(vec_t v1, vec_t v2) { return _mm ## BL ## _andnot_ ## IT(v1, v2); }

#define _FMA(ST, BL, IT) \
inline vec_t vec_fma(vec_t v1, vec_t v2, vec_t v3) { return _mm ## BL ## _fmadd_ ## IT(v1, v2, v3); } \
inline vec_t vec_fms(vec_t v1, vec_t v2, vec_t v3) { return _mm ## BL ## _fmsub_ ## IT(v1, v2, v3); }

*/
