## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

using Printf
using Random
using Statistics

using EuNms


ref_func = Dict()
ref_func[:plusfp] = a -> foldl(+, big.(a), init=zero(BigFloat))
ref_func[:sloppy] = a -> foldl(+, big.(a), init=zero(BigFloat))


function test_precision(func, input, func_ref, nstat, εmax=missing)

    εs = BigFloat[]

    for _ in 1:nstat

        args = input()
        s = func(args...)
        s0 = func_ref(args...)

        ε = rel_error(big(s), s0)
        append!(εs, ε)

        if ismissing(εmax)
            εmax = max(εmax, (ε, args, s))
        end

    end

    εs, εmax
end


function test(algo; size, nstat, rng, seed, unit)

    setprecision(BigFloat, 256)

    Tables.title(string("DWSum/", algo); size=2^size, nstat=nstat, rng=string(rng), seed=seed, unit=unit)
    Tables.head(("Format", 8), ("size", 4))
    Tables.vals([(n, 8) for n in ["min", "Q_1", "med", "Q_3", "max"]]...; color=:yellow)
    Tables.vals(("bound", 10); color=:yellow)
    println()

    I(F, size) = () -> (rng(DW{F}, 2^size), )

    for (S, F) in Iterators.product(4:size,
                                    [[Float32, Float64]; @isdefined(Float128) ? [Float128] : []])

        u = Tables.getUnit(unit, F)
        bound = known_rel_bnds[(DWSum, algo)](getFPrdoff(F), 2^S)

        Random.seed!(seed)

        εmax = (big(0.0), (zero(DW{F}), zero(F)), zero(F))
        εs, εmax = test_precision(y -> DWSum(Val(algo), y), I(F, S), ref_func[algo], nstat, εmax)

        stats = quantile(εs, [0.0, 0.25, 0.5, 0.75, 1.0])

        Tables.head((string(F), 8), (string(2^S), 4))
        Tables.vals([(x / u, 8) for x in stats]...)
        Tables.check((bound / u, 8), stats[end] ≤ bound)
        println()

        εmax[1] > bound && @warn @sprintf "args of max: %s" εmax[2]
    end

end


#######################################################################
## Main.
#######################################################################


begin
    size = 11
    size = 8
    nstat = 5040
    seed = get(ENV, "SEED", 123456789)
    unit = :u²

    test(:plusfp; size=size, nstat=nstat, rng=rand, seed=seed, unit=unit)
    test(:sloppy; size=size, nstat=nstat, rng=rand, seed=seed, unit=unit)

    test(:plusfp; size=size, nstat=nstat, rng=hyprand, seed=seed, unit=unit)
    test(:sloppy; size=size, nstat=nstat, rng=hyprand, seed=seed, unit=unit)
end
