## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

## Compare versions of error bounds
using Printf
using Random
using Statistics

using EuNms


ref_func(args...) = begin
    sq(x) = x^2
    big.(args[1]) .|> sq |> sum |> sqrt
end


hyprandT(F, n) = hyprand(F, (exponent(floatmin(F))+precision(F), exponent(floatmax(F))-8), n)


function test_precision(func, input_gen, func_ref, nstat)

    εs = BigFloat[]
    faithfuls, corrects = 0, 0

    for _ in 1:nstat

        args = input_gen()
        s = func(args...)
        s0 = func_ref(args...)

        ε = rel_error(s, s0)
        append!(εs, ε)

        is_faithful(s, s0) && (faithfuls += 1)
        is_correct(s, s0) && (corrects += 1)
    end

    εs, faithfuls, corrects
end


function test(algo, sum_algo, sq_algo, prm; size, nstat, seed, unit)

    setprecision(BigFloat, 256)

    Tables.title(string("EuNm/",algo,"/",sum_algo,"/",sq_algo,"/",prm);
                 size=2^size, nstat=nstat, seed=seed, unit=string(unit))
    Tables.head(("RNG", 8), ("Format", 8), ("size", 4))
    Tables.vals([(n, 8) for n in ["min", "Q_1", "med", "Q_3", "max"]]...; color=:yellow)
    Tables.vals(("bound", 10); color=:yellow)
    Tables.head_rounds()
    println()

    for (S, F, R) in Iterators.product(4:size,
                                       #[Float32, Float64],
                                       [Float64],
                                       #[rand, hyprandT],
                                       [hyprandT],
                                      )

        u = Tables.getUnit(unit, F)

        Tables.head((string(R), 8), (string(F), 8), (string(2^S), 4))
        bound = if (EuNm, algo, sum_algo, sq_algo, prm) in keys(known_rel_bnds)
            known_rel_bnds[(EuNm, algo, sum_algo, sq_algo, prm)](getFPrdoff(F), 2^S)
        else
            convert(F, NaN)
        end

        func = a -> EuNm(Val(algo), Val(sum_algo), Val(sq_algo), ENParam(Val(prm), F; validate=:no), a)
        input_gen = () -> (R(F, 2^S), )
        εs, faithfuls, corrects = test_precision(func, input_gen, ref_func, nstat)

        stats = quantile(εs, [0.0, 0.25, 0.5, 0.75, 1.0])
        Tables.vals([(x ./ u, 8) for x in stats]...)
        Tables.check((bound / u, 8), stats[end] ≤ bound)
        Tables.rounds(faithfuls, corrects)
        println()
    end

end


#######################################################################
## Main.
#######################################################################


if abspath(PROGRAM_FILE) == @__FILE__

    size = 13
    nstat = 5040
    rng = rand
    #rng = hyprand
    seed = get(ENV, "SEED", 123456789)
    unit = :ε

    #test(:masks, :plusfp, :muller, :muller; size=size, nstat=nstat, seed=seed, unit=unit)
    #test(:onep1, :plusfp, :muller, :muller; size=size, nstat=nstat, seed=seed, unit=unit)
    #test(:onep2, :plusfp, :muller, :muller; size=size, nstat=nstat, seed=seed, unit=unit)
    #test(:onep3, :plusfp, :muller, :muller; size=size, nstat=nstat, seed=seed, unit=unit)
    #test(:onep2, :sloppy, :graill, :graill; size=size, nstat=nstat, seed=seed, unit=unit)
    test(:netlib, :sloppy, :graill, :graill; size=size, nstat=nstat, seed=seed, unit=unit)
end
