## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

## Make result for the paper.
using Printf
using Random
using Statistics
using Quadmath
using Arblib
using ProgressBars

using EuNms


#######################################################################


sq(x) = x^2
eunm_exact(args...) = begin
    Arb.(args[1]) .|> sq |> sum |> sqrt
end


function NRM_INP(rng, F, n)
    (_, p, (emin, emax)) = getFPparam(F)
    hyprand(rng, F, (emin, emax-p), n)
end
NRM_INP(F, n) = NRM_INP(Random.GLOBAL_RNG, F, n)

MED_INP(rng, F, n) = hyprand(rng, F, getENparam_muller(F), n)
MED_INP(F, n) = MED_INP(Random.GLOBAL_RNG, F, n)


#######################################################################


## Parameter sets for each algorithm.
ALGOS = Dict()
ALGOS["Naive"] = ((:naive, :plusfp, :muller), :muller)
ALGOS["Netlib"] = ((:netlib, :plusfp, :muller), :muller)
ALGOS["Louvet"] = ((:louvet, :plusfp, :muller), :muller)
ALGOS["Graillat"] = ((:onep2, :sloppy, :graill), :graill)
ALGOS["Ours"] = ((:onep2, :plusfp, :muller), :muller)


function run1!(res, res0, func, arg_iterator, func_ref=eunm_exact)

    for args in arg_iterator
        s = func(args...)
        s0 = func_ref(args...)
        append!(res, s)
        append!(res0, s0)
    end

    res, res0
end


function get_arb_ulp(F::Type{<:AbstractFloat}, x::Arb)
    Arb(eps(convert(F, ubound(x))))
end


function run2(
        (algo, prm),
        seed, nstat,
        (F, R, S),
        err_type)

    rng = Random.seed!(seed)  ## Initialize a new random series
    setprecision(Arb, 2048)   ## Set precision of reference values

    runs = (F[], Arb[])       ## Tuple of (values, reference values)
    arg_iterator = [(R(rng, F, S),) for _ in 1:nstat]

    ## Fancy progress bars.
    arg_iterator = ProgressBar(arg_iterator, printing_delay=0.1)
    set_description(arg_iterator, string(F,"/",R, "/",S,"/",algo[1]))

    ## Select algorithm.
    func = a -> EuNm(Val.(algo)..., ENParam(Val(prm), F; validate=:no), a)

    ## Actually to the test.
    runs = run1!(runs..., func, arg_iterator)

    ## Check under/overflows.
    finite = isfinite.(runs[1])
    !all(finite) && @warn "There was some under/overflows"

    ## Check that exact result yield a unique floating-point number.
    good_reference = arb_is_unique.(F, runs[2])
    !all(good_reference) && @warn "Some reference are not accurate enough"

    spurious =    ifelse.(good_reference, .!finite, missing)
    valid = finite .&& good_reference

    εs = if err_type == :ulp
        ifelse.(valid, ulp_error.(Arb.(runs[1]), runs[2], get_arb_ulp.(F, runs[2])), missing)
    else
        ifelse.(valid, rel_error.(Arb.(runs[1]), runs[2]), missing)
    end
    faithfuls =   ifelse.(good_reference, is_faithful.(runs[1], runs[2]), missing)
    corrects =    ifelse.(good_reference, is_correct.(runs[1], runs[2]), missing)

    (εs, spurious, faithfuls, corrects)
end


function print_stats(io::IO, name, size, rnd, F, εs, spurious, faithfuls, corrects, err_type)

    εs2 = filter(!ismissing, εs)
    ntest = length(εs2)
    stats = ntest > 0 ? quantile(εs2, [0.0, 0.25, 0.5, 0.75, 1.0]) : fill(Arb(NaN), 5)

    unit = (err_type == :ulp) ? one(F) : eps(F) / 2

    #line = [name,
    #        @sprintf("%3d", precision(F)),
    #        @sprintf("\\SI{%3d}{\\percent}", div(100 * sum(spurious), length(spurious) )),
    #        @sprintf("%.4f", stats[3] / unit),
    #        @sprintf("%.4f", ubound(stats[5] / unit)),
    #        @sprintf("\\SI{%3d}{\\percent}", div(100 * sum(faithfuls), max(1, ntest) )),
    #        @sprintf("\\SI{%3d}{\\percent}", div(100 * sum(corrects), max(1, ntest) )),
    #       ]
    #println(io, " " ^ 4, join(line, " & "), " \\\\")

    line = [@sprintf("%8s", name),
            @sprintf("%6s", string(rnd)),
            @sprintf("%3d", precision(F)),
            @sprintf("%8d", size),
            @sprintf("%3d%%", div(100 * sum(spurious), length(spurious) )),
            @sprintf("%.4f", stats[3] / unit),
            @sprintf("%.4f", ubound(stats[5] / unit)),
            @sprintf("%3d%%", div(100 * sum(faithfuls), max(1, ntest) )),
            @sprintf("%3d%%", div(100 * sum(corrects), max(1, ntest) )),
           ]
    println(io, " " ^ 2, join(line, " "))

end
print_stats(args...) = print_stats(stdout, args...)


function print_histogram(io::IO, name, size, rnd, F, εs, spurious, faithfuls, corrects, err_type)

    unit = (err_type == :ulp) ? one(F) : eps(F) / 2

    εs2 = filter(!ismissing, εs)

    w0 = ldexp(F(2.0), -8)
    ws = w0:w0:ldexp(w0, 8)

    histogram = hist(convert.(F, εs2 / unit), ws)
    for (x, y) in zip(ws, histogram)
        println(io, x, " ", y)
    end
    println(io)
    println(io)

end
print_histogram(args...) = print_histogram(stdout, args...)


#######################################################################
## Main.
#######################################################################


if abspath(PROGRAM_FILE) == @__FILE__

    seed = get(ENV, "SEED", 123456789)
    err_type = :rel  ## How to compute errors: :ulp or :rel

    runs = Iterators.product(
                             [2^5, 2^6],
                             ["Naive", "Netlib", "Louvet", "Graillat", "Ours"],
                             [NRM_INP, MED_INP],
                             [Float32, Float64],
                            )

    io_stat = open(string("stat.dat"), "w")
    for (size, name, rnd, F) in runs
        (algo, prm) = ALGOS[name]

        ## Do the computation
        result = run2((algo, prm),
                       seed, 10^5,
                       (F, rnd, size),
                       err_type)

        ## Flatten arguments for printing
        result = (name, size, rnd, F, result...)

        print_stats(stdout, result..., err_type)
        print_stats(io_stat, result..., err_type)
        flush(io_stat)
        io = open(string("hist", "-", name, "-", F, ".dat"), "w")
        print_histogram(io, result..., err_type)
        close(io)
    end

    close(io_stat)
end
