## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

using Printf
using Random
using Statistics

using Quadmath

using EuNms


ref_func = Dict()
ref_func[:muller]  = (x,) -> sqrt(big(x))
ref_func[:muller2] = (x,) -> sqrt(big(x))
ref_func[:graill]  = (x,) -> sqrt(big(x))


input_gen = Dict()
input_gen[:muller]  = (rng, F) -> () -> (rng(DW{F}), )
input_gen[:muller2] = (rng, F) -> () -> (rng(DW{F}), )
input_gen[:graill]  = (rng, F) -> () -> (rng(DW{F}), )


function test_precision(algo, input, func_ref, nstat, εmax)

    εs = BigFloat[]
    for _ in 1:nstat

        args = input()
        s = SQRTDW(Val(:noovf), Val(algo), args...)

        args0 = big.(args)
        s0 = func_ref(args0...)

        begin
            ε = rel_error(big(s), s0)
            append!(εs, ε)
            εmax = max(εmax, (ε, args, s))
        end
    end

    return εs, εmax
end


function test(algo; nstat, seed, unit)

    setprecision(BigFloat, 256)

    Tables.title(string("SQRTDW/noovf/",algo);
                 nstat=nstat, seed=seed, unit=string(unit))
    Tables.head(("Rand", 8), ("Format", 8))
    Tables.vals([(n, 8) for n in ["min", "Q_1", "med", "Q_3", "max"]]...; color=:yellow)
    Tables.vals(("bound", 10); color=:yellow)
    #Tables.head_rounds()
    println()

    for (I, F) in Iterators.product([rand, hyprand, hyprand2],
                                    [Float32, Float64, Float128])

        bound = known_rel_bnds[(SQRTDW, :noovf, algo)](getFPrdoff(F))

        Random.seed!(seed)
        εmax = (big(0.0), (zero(DW{F}), ), zero(DW{F}))
        εs, εmax = test_precision(algo, input_gen[algo](I, F), ref_func[algo], nstat, εmax)

        stats = quantile(εs, [0.0, 0.25, 0.5, 0.75, 1.0])
        #εmax[1] > bound && @warn @sprintf "args of max: %s" εmax[2]

        u = Tables.getUnit(unit, F)
        Tables.head((string(I), 8), (string(F), 8))
        Tables.vals([(x ./ u, 8) for x in stats]...)
        Tables.check((bound / u, 8), stats[end] ≤ bound)
        println()
    end

end


#######################################################################
## Main.
#######################################################################


begin
    nstat = 25555*17
    nstat = 10
    seed = get(ENV, "SEED", 123456789)

    test(:graill,  nstat=nstat, seed=seed, unit=:u)
    test(:muller,  nstat=nstat, seed=seed, unit=:u)
    test(:muller2, nstat=nstat, seed=seed, unit=:u²)
end
