import asyncio
import csv
import enum
import sys
import textwrap


TABLES = [
    ("NRM_INP", "24"),
    ("NRM_INP", "53"),
    ("MED_INP", "24"),
    ("MED_INP", "53"),
    ]

PAD = " " * 8

ALGO_NAMES = {
    "Naive": "Naive",
    "Netlib": "Hammarling",
    "Graillat": "Graillat et al.",
    "Ours": "Ours",
    }


class StatDialect:
    delimiter = ' '
    quotechar = '"'
    excapechar = None
    doublequote = False
    skipinitialspace = True
    lineterminator = '\r\n'
    quoting = csv.QUOTE_MINIMAL


class Columns(enum.IntEnum):
    KEY = 0
    PROFILE = 1
    PRECISION = 2
    SIZE = 3
    SPURIOUS_OVERFLOW = 4
    ERROR_MEDIAN = 5
    ERROR_MAX = 6
    ROUNDING_FAITHFUL = 7
    ROUNDING_CORRECT = 8


TABLE_HEADER = object()
TABLE_FOOTER = object()

escape_latex = lambda s: s.replace("%", r"~\%").replace("_", r"\_")


async def processor(rows, printq):
    """Turn data rows into latex lines

    Also includes header and footer
    """

    ## Get information from first row.
    row = await rows.get()

    ## Write table header.
    header = textwrap.dedent(
       r"""
       \toprule
       \multicolumn{6}{c}{Accuracy tests for $p=%s$ and {\sf %s}} \\
       \midrule
       Algorithm & {$n$} & Spurious & {Maximum} & \multicolumn{2}{c}{Rounding} \\
       &                 & overflow & {error / $u$} & Faithful & Correct \\"""
       % (row[Columns.PRECISION], escape_latex(row[Columns.PROFILE])))
    await printq.put(textwrap.indent(header, PAD))

    last_key = TABLE_HEADER

    while True:

        if row == TABLE_FOOTER:
            break

        if last_key == row[Columns.KEY]:
            name = " " * len(ALGO_NAMES[row[Columns.KEY]])
        elif last_key ==  TABLE_HEADER:
            await printq.put(PAD + r"\midrule[\heavyrulewidth]")
            name = ALGO_NAMES[row[Columns.KEY]]
        else:
            await printq.put(PAD + r"\midrule")
            name = ALGO_NAMES[row[Columns.KEY]]

        row2 = [
            name,
            format(row[Columns.SIZE], ">4s"),
            format(row[Columns.SPURIOUS_OVERFLOW], ">4s"),
            format(row[Columns.ERROR_MAX], ">7s"),
            format(row[Columns.ROUNDING_FAITHFUL], ">4s"),
            format(row[Columns.ROUNDING_CORRECT], ">4s"),
            ]
        row3 = [escape_latex(c) for c in row2]
        await printq.put(PAD + " & ".join(row3) + r" \\")

        last_key = row[Columns.KEY]

        rows.task_done()
        row = await rows.get()

    ## Table footer.
    await printq.put(PAD + r"\bottomrule")
    rows.task_done()


async def printer(prnts):
    for printq in prnts:
        while True:
            line = await printq.get()
            print(line)
            printq.task_done()
            if printq.empty():
                break


async def main(
        inputfile="stat.dat",
        ):

    ## Create one task for each table.
    procs, tasks, prnts = dict(), dict(), dict()
    for coef in TABLES:
        procs[coef] = asyncio.Queue()
        prnts[coef] = asyncio.Queue()
        tasks[coef] = asyncio.create_task(
                processor(procs[coef], prnts[coef]))

    ## Create printer task.
    ptask = printer([prnts[coef] for coef in TABLES])

    with open(inputfile) as fp:
        csv_reader = csv.reader(fp, dialect=StatDialect)

        for row in csv_reader:
            coef = (row[Columns.PROFILE], row[Columns.PRECISION])
            await procs[coef].put(row)

    for coef in TABLES:
        await procs[coef].put(TABLE_FOOTER)
    await ptask

    for coef in TABLES:
        await procs[coef].join()
        await prnts[coef].join()
        tasks[coef].cancel()

    return 0


if __name__ == "__main__":
    status = asyncio.run(main())
    sys.exit(status)
