## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

using Printf
using Random
using Statistics

using BenchmarkTools

using EuNms


function test(; seed)

    Random.seed!(seed)

    Tables.title("DWAdd", seed=seed)

    bench_suite = BenchmarkGroup()

    bench_suite["DWPlusFP"] = BenchmarkGroup()
    for F in [[Float32, Float64]; @isdefined(Float128) ? [Float128] : []]
        args = (rand(DW{F}), rand(F))
        bench_suite["DWPlusFP"][F] = @benchmarkable DWPlusFP($args...) samples=1000000
    end

    bench_suite["SloppyDWPlusDW"] = BenchmarkGroup()
    for F in [[Float32, Float64]; @isdefined(Float128) ? [Float128] : []]
        args = (rand(DW{F}), rand(DW{F}))
        bench_suite["SloppyDWPlusDW"][F] = @benchmarkable SloppyDWPlusDW($args...) samples=1000000
    end

    tune!(bench_suite)
    println(bench_suite)
    results = run(bench_suite, seconds=2)

    Tables.head(("Algorithm", 14), ("Format", 8))
    Tables.vals(("Time / ns", 9), ("Mem / b", 9), ("Allocs", 9); color=:yellow)
    println()

    for F in [[Float32, Float64]; @isdefined(Float128) ? [Float128] : []]
        Tables.head(("DWPlusFP", 14), (string(F), 8))
        Tables.vals((@sprintf("%g", time(results["DWPlusFP"][F])), 9),
                    (@sprintf("%d", memory(results["DWPlusFP"][F])), 9),
                    (@sprintf("%d", allocs(results["DWPlusFP"][F])), 9))
        println()
    end
    for F in [[Float32, Float64]; @isdefined(Float128) ? [Float128] : []]
        Tables.head(("SloppyDWPlusDW", 14), (string(F), 8))
        Tables.vals((@sprintf("%g", time(results["SloppyDWPlusDW"][F])), 9),
                    (@sprintf("%d", memory(results["SloppyDWPlusDW"][F])), 9),
                    (@sprintf("%d", allocs(results["SloppyDWPlusDW"][F])), 9))
        println()
    end

end


#######################################################################
## Main.
#######################################################################


begin
    seed = get(ENV, "SEED", rand(UInt))
    test(seed=seed)
end
