## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

using Printf
using Random
using Statistics

using BenchmarkTools

using EuNms


function test(; seed, size)

    Random.seed!(seed)

    Tables.title("DWSum", seed=seed, size=2^size)

    algs = [DWSum_plusfp, DWSum_sloppy]
    fmts = [[Float32, Float64]; @isdefined(Float128) ? [Float128] : []]

    bench_suite = BenchmarkGroup()
    for A in algs
        bench_suite[A] = BenchmarkGroup()
        for F in fmts
            bench_suite[A][F] = BenchmarkGroup()
            for S in 1:size
                bench_suite[A][F][S] = BenchmarkGroup()
            end
        end
    end

    for (A, F, S) in Iterators.product(algs, fmts, 1:size)
        args = rand(DW{F}, 2^S)
        bench_suite[A][F][S] = @benchmarkable $A($args)
    end

    tune!(bench_suite)
    results = run(bench_suite, seconds=2)

    Tables.head(("Format", 8), ("Algorithm", 20), ("Size", 4))
    Tables.vals(("Time / ns", 9), ("Mem / b", 9), ("Allocs", 9); color=:yellow)
    println()

    for (S, A, F) in Iterators.product(1:size, algs, fmts)
        Tables.head((string(F), 8), (string(A), 20), (string(2^S), 4))
        Tables.vals((@sprintf("%7.1f", time(results[A][F][S])), 9),
                    (@sprintf("%d", memory(results[A][F][S])), 9),
                    (@sprintf("%d", allocs(results[A][F][S])), 9))
        println()
    end

end


#######################################################################
## Main.
#######################################################################


begin
    seed = get(ENV, "SEED", rand(UInt))
    size = 8
    test(seed=seed, size=size)
end
