## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

using Printf
using Random

using EuNms


"""
    int2bin(d, [pad])

return bitstring of integer ``d`` with ``pad`` digits at least.
"""
function int2bin(d::Int, pad::Int=0)
    q, r = divrem(d, 2)
    s = ""
    while pad > 0 || q > 0
        s = string(r, s)
        pad -= 1
        q, r = divrem(q, 2)
    end
    s
end


#######################################################################
## Coordinates and co.
#######################################################################


"""
    fp_bin_coords(nbit, x)

return _coordinates_ of number ``x``.

Coordinates are in the form ``(s, e)`` where ``e`` is the magnitude of ``x``,
and ``s`` is the first ``nbit`` of the significant, packed as an integer.

# Example

With ``nbit=3``, the number 0b1.1011e-6 has coordinates (0b110, -5)

"""
function fp_bin_coords(nbit, x::F) where {F<:AbstractFloat}
    s = significand(x)
    s = round(s, RoundToZero, digits=nbit, base=2)
    s = convert(Int, ldexp(s, nbit))
    s = copysign(mod(s, 2^nbit), s)
    (s, exponent(x))
end


function fp_bin_coords(nbit, x::DW{F}) where {F<:AbstractFloat}
    il = fp_bin_coords(nbit, lo(x))
    ih = fp_bin_coords(nbit, hi(x))
    (il[1], il[2], ih[1], ih[2])
end


function coords_to_index(i::NTuple{2,Int}; emax)
    ( 1+i[1], 1+emax-i[2] )
end


function coords_to_index(i::NTuple{4,Int}; emax, prec, nbit)
    ( 1+nbit+1+i[1], 1+emax-prec+i[4]-i[2], 1+i[3], 1+emax-i[4] )
end


function clamp2(is::NTuple{N,Int}, sizes::NTuple{N,Int}) where {N}
    Tuple(clamp(is[i], 1, sizes[i]) for i in eachindex(is))
end


#######################################################################
## Collect random numbers.
#######################################################################


function collect(emin, emax, nbit, xs::Array{F}) where {F<:AbstractFloat}
    bins = zeros(Int, 2^nbit, emax-emin+2)

    cs = fp_bin_coords.(nbit, xs)
    is = coords_to_index.(cs; emax=emax)

    for i in is
        i = clamp2(i, size(bins))
        bins[CartesianIndex(i)] += 1
    end

    bins
end


function collect(emin, emax, nbit, xs::Array{DW{F}}) where {F<:AbstractFloat}
    bins = zeros(Int, 2*2^nbit, emax-emin+2, 2^nbit, emax-emin+2)

    cs = fp_bin_coords.(nbit, xs)
    is = coords_to_index.(cs; emax=emax, prec=precision(F), nbit=nbit)

    for i in is
        i = clamp2(i, size(bins))
        bins[CartesianIndex(i)] += 1
    end

    bins
end


function printbins((emin, emax), nbit, bins::Array{Int, 2})

    for e in emax:-1:emin-1
        for s in (2^nbit)-1:-1:0
            head  = (s == 0) ? (@sprintf " %+2d " e) : (@sprintf " %2s " " ")
            head *= "1." * int2bin(s, nbit) * " |"
            printstyled(head; color=:yellow)

            data = lpad(bins[1+s, 1+emax-e], 6)
            printstyled(data; color=
                        (emin<=e<emax) ? :white : :light_black)
            println()
        end
        println()
    end

end


function printbins((emin, emax, prec), nbit, bins::Array{Int, 4})

    head1, head2 = (@sprintf "%10s" " "), (@sprintf "%10s" " ")
    for e in emax-prec:-1:emin-prec-1
        for s in (2^nbit)-1:-1:0
            head1 *= lpad((s == 0) ? (@sprintf " %+d" e) : " ", 6)
            head1 *= (s == 0) ? " |" : "  "
            head2 *= lpad("  1." * int2bin(s, nbit), 6)
            head2 *= (s == 0) ? " |" : "  "
        end
    end
    printstyled(head1; color=:yellow); println()
    printstyled(head2; color=:yellow); println()

    for eh in emax:-1:emin-1
        for sh in (2^nbit)-1:-1:0
            head  = (sh == 0) ? (@sprintf " %+2d " eh) : (@sprintf " %2s " " ")
            head *= "1." * int2bin(sh, nbit) * " |"
            printstyled(head; color=:yellow)

            for el in emax:-1:emin-1
                for sl in (2^nbit)-1:-1:0

                    data = lpad(bins[1+sl, 1+emax-el, 1+sh, 1+emax-eh], 6)
                    printstyled(data; color=
                                (emin<=eh<emax && emin<=el<emax) ? :white : :light_black)
                    sl > 0 && print("  ")
                end
                print(" |")
            end
            println()
        end
        println()
    end
    
end


#######################################################################
## Main.
#######################################################################


begin
    seed = get(ENV, "SEED", rand(UInt))
    nbit = 2
    emin, emax = -8, 0
    nstat = 2^16

    println("""
            ## This program let you check Random Number Generators by
            ## reviewing PDFs of random series of numbers.
            """)

    println("""
            ## The standard ``rand`` generator should have a uniform pdf in
            ## the range ``[0,1)``. This mean a probability of ``1/2^e``
            ## for numbers in the range ``[2^-e, 2^(-e+1))`` and uniform
            ## inside.
            """)
    printstyled("[RAND]"; bold=true, color=:yellow)
    printstyled(@sprintf " F:%s" Float64; color=:yellow)
    printstyled(@sprintf " seed:0x%x" seed; color=:yellow)
    printstyled(@sprintf " nstat:%d" nstat; color=:yellow)
    println()
    Random.seed!(seed)
    xs = rand(Float64, nstat)
    bins = collect(emin, emax, nbit, xs)
    printbins((emin, emax), nbit, bins)

    println("""
            ## The ``hyprand`` generator is designed to have a uniform
            ## distribution for each range ``[2^-e, 2^(-e+1))`` between
            ## ``emin`` and ``emax``.
            """)
    printstyled("[HYPRAND]"; bold=true, color=:yellow)
    printstyled(@sprintf " F:%s" Float64; color=:yellow)
    printstyled(@sprintf " e:%d…%d" emin emax; color=:yellow)
    printstyled(@sprintf " seed:0x%x" seed; color=:yellow)
    printstyled(@sprintf " nstat:%d" nstat; color=:yellow)
    println()
    Random.seed!(seed)
    xs = hyprand(Float64, (emin, emax), nstat)
    bins = collect(emin, emax, nbit, xs)
    printbins((emin, emax), nbit, bins)

    printstyled("[RAND]"; bold=true, color=:yellow)
    printstyled(@sprintf " F:%s" DW{Float64}; color=:yellow)
    printstyled(@sprintf " seed:0x%x" seed; color=:yellow)
    printstyled(@sprintf " nstat:%d" nstat; color=:yellow)
    println()
    Random.seed!(seed)
    xs = rand(DW{Float64}, nstat)
    bins = collect(emin, emax, nbit, xs)
    printbins((emin, emax, precision(Float64)), nbit, bins)
end
