## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

using Random
using Test

using EuNms

include("faithfulnormtools.jl")
include("accuracy/accuracy_EuNm.jl")

#const two1022 = β^1022  ## 4.49423283715578976932326297697256183404494244735577e307
#const two969  = β^969   ## 4.9896007738367995291409317825920964151686281510886e291

l2norm(x) = sqrt(sum(big.(x) .^ 2))


function half_up_cases(F, size, nstat, seed, unit)

    setprecision(BigFloat, 256)

    Tables.title("HALF_ULP_CASES"; fmt=F, size=2^size, nstat=nstat, seed=seed, unit=unit)
    Tables.head(("Algo", 5), ("Sum", 6), ("Sqrt", 6), ("Param", 6))
    Tables.vals([(n, 8) for n in ["min", "Q_1", "med", "Q_3", "max"]]...; color=:yellow)
    Tables.vals(("bound", 10); color=:yellow)
    Tables.head_rounds()
    println()

    for (algo, sum_algo, sq_algo, prm) in [[:masks, :plusfp, :muller, :muller],
                                           [:onep2, :sloppy, :muller, :graill]]
        Random.seed!(seed)
        Tables.head((string(algo), 5), (string(sum_algo), 6), (string(sq_algo), 6), (string(prm), 6))
        bound = if (EuNm, algo, sum_algo, sq_algo, prm) in keys(known_rel_bnds)
            known_rel_bnds[(EuNm, algo, sum_algo, sq_algo, prm)](getFPrdoff(F), 2^size)
        else
            convert(F, NaN)
        end

        func = a -> EuNm(Val(algo), Val(sum_algo), Val(sq_algo), ENParam(Val(prm), F; validate=:no), a)
        input_gen = () -> (rand_half_ulp_cases(F, 2^size), )
        εs, faithfuls, corrects = test_precision(func, input_gen, ref_func, nstat)

        u = Tables.getUnit(unit, F)

        stats = quantile(εs, [0.0, 0.25, 0.5, 0.75, 1.0])
        Tables.vals([(x / u, 8) for x in stats]...)
        Tables.check((bound / u, 8), stats[end] ≤ bound)
        Tables.rounds(corrects, faithfuls)
        println()
    end
end


function around_one(F, size, nstat, seed, unit)

    setprecision(BigFloat, 256)

    Tables.title("AROUND_ONE"; fmt=F, size=2^size, nstat=nstat, seed=seed, unit=unit)
    Tables.head(("Algo", 5), ("Sum", 6), ("Sqrt", 6), ("Param", 6))
    Tables.vals([(n, 8) for n in ["min", "Q_1", "med", "Q_3", "max"]]...; color=:yellow)
    Tables.vals(("bound", 10); color=:yellow)
    Tables.head_rounds()
    println()

    for (algo, sum_algo, sq_algo, prm) in [[:masks, :plusfp, :muller, :muller],
                                           [:onep2, :sloppy, :muller, :graill]]
        Random.seed!(seed)
        Tables.head((string(algo), 5), (string(sum_algo), 6), (string(sq_algo), 6), (string(prm), 6))
        bound = if (EuNm, algo, sum_algo, sq_algo, prm) in keys(known_rel_bnds)
            known_rel_bnds[(EuNm, algo, sum_algo, sq_algo, prm)](getFPrdoff(F), 2^size)
        else
            convert(F, NaN)
        end

        func = a -> EuNm(Val(algo), Val(sum_algo), Val(sq_algo), ENParam(Val(prm), F; validate=:no), a)
        input_gen = () -> (rand_half_ulp_cases(F, 2^size), )
        εs, faithfuls, corrects = test_precision(func, input_gen, ref_func, nstat)

        u = Tables.getUnit(unit, F)

        stats = quantile(εs, [0.0, 0.25, 0.5, 0.75, 1.0])
        Tables.vals([(x / u, 8) for x in stats]...)
        Tables.check((bound / u, 8), stats[end] ≤ bound)
        Tables.rounds(corrects, faithfuls)
        println()
    end
end



if abspath(PROGRAM_FILE) == @__FILE__

    F = Float64
    size = 10
    nstat = 5040
    seed = get(ENV, "SEED", 123456789)
    unit = :ε

    #half_up_cases(F, size, nstat, seed, unit)
    around_one(F, size, nstat, seed, unit)

    # @testset "HALF_ULP_CASES" begin

    #     x = [[1.0]; [exactSquare + perturb * randomDoubleMOneOne(typeof(β), (-1074, 0)) for _ in 2:n]]
    #     @testset "$(repr(EuNm))" for EuNm in ALGOS @test EuNm(x) == convert(typeof(β), l2norm(x)) end
    # end

    # @testset "MONOTONE_INCREASING" begin

    #     x = 1.0 .+ [(i-1)/n for i in 1:n]
    #     @testset "$(repr(EuNm))" for EuNm in ALGOS @test EuNm(x) == convert(typeof(β), l2norm(x)) end
    # end

    # @testset "MONOTONE_DECREASING" begin

    #     x = 1.0 .+ [(n-i-1)/n for i in 1:n]
    #     @testset "$(repr(EuNm))" for EuNm in ALGOS @test EuNm(x) == convert(typeof(β), l2norm(x)) end
    # end

    # @testset "NETLIB_SPURIOUS_UNDERFLOW" begin

    #     x = [[two1022, two969]; [exactSquare + perturb * randomDoubleMOneOne(typeof(β), (-1074, -1)) for _ in 3:n]]
    #     @testset "$(repr(EuNm))" for EuNm in ALGOS @test EuNm(x) == convert(typeof(β), l2norm(x)) end
    # end
end
