## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

using Printf
using Quadmath

using EuNms


Base.bitstring(x::Float128) = bitstring(reinterpret(UInt128, x))


function digits(q, scale=0)
    bits = ""
    while q > 0
        (q, r) = divrem(q, 2)
        bits = string((r == 1) ? "1" : "0", bits)
    end
    string("0" ^ scale, bits)
end


function pp(x, scale=0)
    let (m, e, s) = Base.decompose(x)
        string(@sprintf("%+1d %+04d ", s, e+precision(x)-1+scale), digits(m, scale))
    end
end


#begin
#    x1 = reinterpret.(Float128, [
#        0x403970a1aaac0968dd65d1c2df6f2c96,
#        0x4001a3d6e3b8c99c05d885a13b8e53eb,
#         ])
#    #@printf " %a %a\n" x1[1] x1[2]
#
#    setprecision(BigFloat, 128)
#
#    s0 = sqrt(sum(BigFloat.(x1) .^ 2))
#    s1 = EuNm(Val(:onep2), Val(:sloppy), Val(:graill), ENParam(Val(:graill), Float128), x1)
#    s2 = EuNm(Val(:onep2), Val(:plusfp), Val(:muller), ENParam(Val(:muller), Float128), x1)
#    println(pp(s0))
#    println(pp(s1))
#    println(pp(s2))
#
#    @printf "%6.4f\n" (ulp_error(BigFloat(s1), s0, eps(Float128(s0))))
#    @printf "%6.4f\n" (ulp_error(BigFloat(s2), s0, eps(Float128(s0))))
#end


begin
    S1 = (h = reinterpret(Float128, 0x4074096898635c9d21439037afb591cc), 
          l = reinterpret(Float128, 0xc002fef5777d9993c0d5e7b20646ec68))

    println("# Awful input: ", S1)
    println("# Is it a proper DW: ", checkDW(S1...))
    println(pp(S1.h))
    println(pp(S1.l, exponent(S1.h) - exponent(S1.l)))
    println()

    setprecision(BigFloat, 128)

    s0 = sqrt(DWTtoBig(S1))
    s1 = sqrt(S1.h)
    s2 = SQRTDWtoFP(S1)
    ulp = BigFloat(eps(Float128(s0)))

    println("# Comparison of: s0 = sqrt_MPFR(S1), s1 = sqrt(S1.h), s2 - SQRTDWtoFP(S1)")
    println(pp(s0))
    println(pp(s1))
    println(pp(s2))
    println("# ulp error: |s1 - s0| / ulp(s0), |s2 - s0| / ulp(s0)")
    println(@sprintf "%6.4f" ulp_error(BigFloat(s1), s0, ulp))
    println(@sprintf "%6.4f" ulp_error(BigFloat(s2), s0, ulp))
    println()

    #ulp = BigFloat(eps(Float128(s0)))
    #println(pp(ulp))
    #println(pp(BigFloat(s1)-s0))
    #@printf "%6.4f\n" (ulp_error(BigFloat(s1), s0, ulp))
    #@printf "%6.4f\n" (ulp_error(BigFloat(s2), s0, ulp))

    s0 = sqrt(BigFloat(S1.h))
    s1 = sqrt(S1.h)
    ulp = BigFloat(eps(Float128(s0)))

    println("# Comparison of: s0 = sqrt_MPFR(S1.h), s1 = sqrt(S1.h)")
    println(pp(s0))
    println(pp(s1))
    println("# ulp error: |s1 - s0| / ulp(s0)")
    println(@sprintf "%6.4f" ulp_error(BigFloat(s1), s0, ulp))
    println()
end


#begin
#    setprecision(BigFloat, 128)
#    u = Float128(1)
#    for i in 0:201
#        x = u + i*eps(u)
#        println(x," ",sqrt(x)," ",Float128(sqrt(BigFloat(x))))
#    end
#end


begin
    x = reinterpret(Float128, 0x4074096898635c9d21439037afb591cc)

    println("Awful input: ", x)
    println(pp(x))
    println()

    setprecision(BigFloat, 128)

    s0 = sqrt(BigFloat(x))
    s1 = sqrt(x)

    println("Comparison MPFR / Quadmath")
    println(pp(s0))
    println(pp(s1))
    println(ulp_error(BigFloat(s1), s0, BigFloat(eps(Float128(s0)))))
    println()
end
