## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

import Base:finalize

export known_rel_bnds
export Fast2Sum, TwoSum, Fast2Mult
export DWPlusFP, SloppyDWPlusDW, DWAdd
export SQRTDWtoDW, SQRTDWtoFP, SQRTDW
export SQRTDW_big, SQRTDW_med
export SQRTDW2_masks, SQRTDW2_onepass
export DWSum_sloppy, DWSum_plusfp, DWSum
export DWAccu_plusfp, DWAccu_sloppy, finalize, DWAccu
export SqSum_nooverflow_plusfp, SqSum_nooverflow_sloppy, SqSum_nooverflow
export SqSum_masks, SqSum_onepass, SqSum
export EuNm_masks, EuNm_onepass_v1, EuNm_onepass_v2, EuNm


## • This dict stores functions to compute best known relative bound of
##   implemented algorithms.

known_rel_bnds = Dict()


#######################################################################
## Classical building blocs.
#######################################################################


## • Works with both scalar and vector arguments. For vectors, it returns a
##   tuple of arrays, while a broadcast returns an array of tuples.


@inline function Fast2Sum(a, b)
    #@assert abs(a) ≥ abs(b)
    s = a .+ b
    z = s .- a
    t = b .- z
    (h=s, l=t)
end


@inline function TwoSum(a, b)
    s = a + b
    a_ = s - b
    b_ = s - a_
    δa = a - a_
    δb = b - b_
    t = δa + δb
    (h=s, l=t)
end


@inline function Fast2Mult(a, b)
    πh = a .* b
    πl = fma.(a, b, -πh)
    (h=πh, l=πl)
end


#######################################################################
## Double-word addition functions (DWAdd)
#######################################################################


@inline function DWPlusFP((xh, xl), y)
    (sh, sl) = TwoSum(xh, y)
    v = xl .+ sl
    Fast2Sum(sh, v)
end

## • For any input.
#known_rel_bnds[DWPlusFP] = u -> 2u^2 / (1 - 2u)
#
## • For positive input only.
known_rel_bnds[DWPlusFP] = u -> u^2 / (1 + 2u - u^2)


@inline function SloppyDWPlusDW((xh, xl), (yh, yl))
    (sh, sl) = TwoSum(xh, yh)
    v = xl .+ yl
    w = sl .+ v
    Fast2Sum(sh, w)
end

known_rel_bnds[SloppyDWPlusDW] = u -> u^2 * (3 - 8u) / (1 + 10u - 5u^2 - 8u^3)


## • Interface to accept DW

DWPlusFP(x::DW{F}, y) where {F} = convert(DW{F}, DWPlusFP((x.h, x.l), y))
SloppyDWPlusDW(x::DW{F}, y::DW{F}) where {F} = convert(DW{F}, SloppyDWPlusDW((x.h, x.l), (y.h, y.l)))


## • Interface with algorithm parameter

DWAdd(::Val{:plusfp}, (x, y)) = DWPlusFP(x, y)
DWAdd(::Val{:sloppy}, (x, y)) = SloppyDWPlusDW(x, y)

known_rel_bnds[:plusfp] = known_rel_bnds[DWPlusFP]
known_rel_bnds[:sloppy] = known_rel_bnds[SloppyDWPlusDW]


#######################################################################
## Double-word square root function (SQRTDW)
#######################################################################


"""
    SQRTDW((xh, xl))

return ``y₁ + y₂ = √(xh + xl)``.

Warning: ``(y₁, y₂)`` is not necessarily a normal DW number (use SQRTDWtoDW).

Warning: SQRT(0.0) NaNs, you should this special case yourself.
"""
@inline function _SQRTDW_muller((xh, xl))
    sh = sqrt.(xh)
    ρ1 = -fma.(sh, sh, -xh)
    ρ2 = xl .+ ρ1
    sl = ρ2 ./ (2sh)
    (sh, sl)
end


@inline function _SQRTDW_graill((xh, xl))
    sqrt.(xh)
end


"""
    SQRTDWtoDW((xh, xl))

return ``yh + yl = √(xh + xl)``.

Warning: SQRT(0.0) NaNs, you should this special case yourself.
"""
SQRTDWtoDW(x) = Fast2Sum(_SQRTDW_muller(x)...)

known_rel_bnds[SQRTDWtoDW] = u -> 25u^2/8


"""
    SQRTDWtoFP((xh, xl))

return ``y = √(xh + xl)``.

Warning: SQRT(0.0) NaNs, you should this special case yourself.
"""
SQRTDWtoFP(x) = +(_SQRTDW_muller(x)...)

known_rel_bnds[SQRTDWtoFP] = u -> u + 17u^2/8 + 33u^3/8


## • Interface to accept DW

SQRTDWtoDW(x::DW{F}) where {F} = convert(DW{F}, SQRTDWtoDW((x.h, x.l)))
SQRTDWtoFP(x::DW{F}) where {F} = SQRTDWtoFP((x.h, x.l))
_SQRTDW_graill(x::DW{F}) where {F} = _SQRTDW_graill((x.h, x.l))

## Generic


SQRTDW(::Val{:noovf}, ::Val{:muller}, x) = SQRTDWtoFP(x)
SQRTDW(::Val{:noovf}, ::Val{:muller2}, x) = SQRTDWtoDW(x)
SQRTDW(::Val{:noovf}, ::Val{:graill}, x) = _SQRTDW_graill(x)


known_rel_bnds[(SQRTDW, :noovf, :muller)] = known_rel_bnds[SQRTDWtoFP]
known_rel_bnds[(SQRTDW, :noovf, :muller2)] = known_rel_bnds[SQRTDWtoDW]
known_rel_bnds[(SQRTDW, :noovf, :graill)] = u -> 2u  ## Not sure of this bound



# TODO: Test performance improvment proposed.


## Classes TINY, MED, and BIG handle spurious under-/overflows. (see below).
## Here are function to compute the square root with the result of different
## classes.


"""
    SQRTDW_big(Sbig, Smed, γ)

return ``y₁ + y₂ = √(Sbig * 2^-γ + Smed)``.

Warning: ``(y₁, y₂)`` is not necessarily a normal DW number.

"""
function SQRTDW_big(Sbig, Smed, prm::ENParam, ::Val{VER}=Val(:muller)) where {VER}
    (β, p, _) = getFPparam(prm.F)

    if Smed.h < β^(2prm.emax2 - 2p) || Sbig.h > β^(2prm.emax2 - 2prm.γ  + 3p)
        return ldexp.(SQRTDW(Val(:noovf), Val(VER), Sbig), prm.γ)

    else
        χ = SloppyDWPlusDW( (h=ldexp(Sbig.h, prm.γ), l=ldexp(Sbig.l, prm.γ)),
                            (h=ldexp(Smed.h,-prm.γ), l=ldexp(Smed.l,-prm.γ)) )
        return ldexp.(SQRTDW(Val(:noovf), Val(VER), χ), prm.γ >> 1)

    end
end


"""
    SQRTDW_med(Smed, Stiny, γ)

return ``y₁ + y₂ = √(Smed + Stiny * 2^γ)``.

Warning: ``(y₁, y₂)`` is not necessarily a normal DW number.

"""
function SQRTDW_med(Smed, Stiny, prm::ENParam, ::Val{VER}=Val(:muller), ::Val{DBG}=Val(false)) where {VER,DBG}
    (β, p, _) = getFPparam(prm.F)

    DBG && begin
        _dbg_print_section(stderr, "SQRTDW_med")
        print(stderr, " "^4)
        @printf stderr "%a %a\n" Stiny.h β^(2prm.emax2 - 2p)
        print(stderr, " "^4)
        @printf stderr "%a %a\n" Smed.h β^(2prm.emax2 - 2prm.γ  + 3p)
    end

    if Stiny.h < β^(2prm.emin2 +2prm.γ - 2p) || Smed.h > β^(2prm.emin2 + 3p)

        DBG && println(stderr, " Single class case")

        return SQRTDW(Val(:noovf), Val(VER), Smed)
    else

        DBG && println(stderr, " Double class case")

        χ = SloppyDWPlusDW( ldexp(Smed, prm.γ), ldexp(Stiny,-prm.γ) )
        return ldexp.(SQRTDW(Val(:noovf), Val(VER), χ), -(prm.γ >> 1))
    end
end


#######################################################################
## Handle SQRT output.
#######################################################################


function SQRTDW2_masks(Sbig, Smed, Stiny, prm, ::Val{SQ}=Val(:muller)) where {SQ}

    if Sbig.h > zero(prm.F)
        return SQRTDW_big(Sbig, Smed, prm, Val(SQ))

    elseif Smed.h > zero(prm.F)
        return SQRTDW_med(Smed, Stiny, prm, Val(SQ))

    elseif Stiny.h > zero(prm.F)
        return ldexp.(SQRTDW(Val(:noovf), Val(SQ), Stiny),-prm.γ)

    else
        return zero(prm.F)
        ## Will fail if you use SQ=:muller2
    end
end


function SQRTDW2_onepass(Smid, Sext, ext_is_big, prm, ::Val{SQ}=Val(:muller)) where {SQ}

    return ext_is_big           ? SQRTDW_big(Sext, Smid, prm, Val(SQ)) :
           Smid.h > zero(prm.F) ? SQRTDW_med(Smid, Sext, prm, Val(SQ)) :
           Sext.h > zero(prm.F) ? ldexp.(SQRTDW(Val(:noovf), Val(SQ), Sext), -prm.γ) :
           zero(prm.F)
        ## Will fail if you use SQ=:muller2
end


#######################################################################
## Double-word summation functions (DWSum).
#######################################################################


## • There is two way of summing arrays of DW numbers, by using
##   SloppyDWPlusDW (simple) or DWPlusFP (more accurate).
##
## • We will use shorter names: “sloppy” and “plusfp” to refer to the former or
##   the latter.
##
## • Sloppy is the algorithm used by Graillat et al. By using plusfp, we aim
##   for better accuracy and provide correct rounding in almost all cases.


"""
    DWSum_sloppy(([xh1, xh2, …], [xl1, xl2, …]))
prm.
return the sum of it's array argument.

Assume len(y) ≥ 1.
"""
@inline function DWSum_sloppy((yh, yl))
    foldl(SloppyDWPlusDW, zip(yh[2:end], yl[2:end]), init=(yh[1], yl[1]))
end


"""
    DWSum_plusfp(([xh1, xh2, …], [xl1, xl2, …]))

return the sum of it's array argument.

Assume len(y) ≥ 2.
"""
@inline function DWSum_plusfp((yh, yl))
    x = foldl(DWPlusFP, yh[3:end]; init=TwoSum(yh[1], yh[2]))
    σ = foldl(+, yl; init=zero(eltype(yl)))
    DWPlusFP(x, σ)
end

## Equivalent version

#@inline function DWSum_plusfp(y::Array{DW{F}}) where {F<:AbstractFloat}
#  x = zero(DW{F})
#  for i in 1:length(y)
#    yh, yl = hi(y[i]), lo(y[i])
#    x = DWPlusFP(x, yh)
#    σ = σ + yl
#  end
#  DWPlus(x, σ)
#end

known_rel_bnds[DWSum_sloppy] = (u, n) -> 3(n-1) * u^2 / (1 - 3(n-1) * u^2)
#known_rel_bnds[DWSum_plusfp] = (u, n) -> 2(n-1) * u^2 * (1+u)
known_rel_bnds[DWSum_plusfp] = (u, n) -> (2n-1) * u^2 + (n-1) *u^3 + (2n-2) * u^4 + (n-1)*u^5


## • Interface to accept DW

DWSum_plusfp(y::Array{DW{F}}) where {F} = convert(DW{F}, DWSum_plusfp((hi.(y), lo.(y),) ))
DWSum_sloppy(y::Array{DW{F}}) where {F} = convert(DW{F}, DWSum_sloppy((hi.(y), lo.(y),) ))


## • This function allow to easily switch between the two algorithms, without
##   overhead because the dispatch will be done during compilation time.

DWSum(::Val{:plusfp}, y) = DWSum_plusfp(y)
DWSum(::Val{:sloppy}, y) = DWSum_sloppy(y)

known_rel_bnds[(DWSum, :plusfp)] = known_rel_bnds[DWSum_plusfp]
known_rel_bnds[(DWSum, :sloppy)] = known_rel_bnds[DWSum_sloppy]


## DWAccu: accumulator versions.
##
## • The two accumulator must have the same interface to serve as a dropin
##   replacement.


struct DWAccu_plusfp{F}
    x::NamedTuple{(:h, :l), NTuple{2,F}}
    σ::F
end
DWAccu_plusfp(F::Type) = DWAccu_plusfp((h=zero(F), l=zero(F)), zero(F))
Base.:+(acc::DWAccu_plusfp, (yh, yl)) = DWAccu_plusfp(DWPlusFP(acc.x, yh), +(acc.σ, yl))
finalize(acc::DWAccu_plusfp) = DWPlusFP(acc.x, acc.σ)


struct DWAccu_sloppy{F}
    x::NamedTuple{(:h, :l), NTuple{2,F}}
end
DWAccu_sloppy(F::Type) = DWAccu_sloppy((h=zero(F), l=zero(F)))
Base.:+(acc::DWAccu_sloppy, y) = DWAccu_sloppy(SloppyDWPlusDW(acc.x, y))
finalize(acc::DWAccu_sloppy) = acc.x

DWAccu(::Val{:plusfp}, F) = DWAccu_plusfp(F)
DWAccu(::Val{:sloppy}, F) = DWAccu_sloppy(F)

"""
    DWAccu(Val(:plusfp) | Val(:sloppy), F)

create a double-word number accumulator initialized at zero.

Once created double-word numbers can be added with ``accu =+ dw``. The result
souhld be accessed through ``finalize(accu)``.
"""
DWAccu


#######################################################################
## Sum of squares functions (SqSum).
#######################################################################


## • This first par does not handle spurious overflows, they are useful nonetheless.


"""
    SqSum_nooverflow_plusfp([x1, …])

return the sum of it's squared array elements, using DWPlusFP

• Spurious over/underflows are not prevented.

• Assume len(y) ≥ 2.
"""
SqSum_nooverflow_plusfp(a) = DWSum_plusfp(Fast2Mult(a, a))
## ‼ Julia does not seem to figure out the return type.


"""
    SqSum_nooverflow_sloppy([x1, …])

return the sum of it's squared array elements, using SloppyDWPlusDW

• Spurious over/underflows are not prevented.

• Assume len(y) ≥ 1.
"""
SqSum_nooverflow_sloppy(a) = DWSum_sloppy(Fast2Mult(a, a))

known_rel_bnds[SqSum_nooverflow_plusfp] = known_rel_bnds[DWSum_plusfp]
known_rel_bnds[SqSum_nooverflow_sloppy] = known_rel_bnds[DWSum_sloppy]


SqSum_nooverflow(::Val{:plusfp}, a) = SqSum_nooverflow_plusfp(a)
SqSum_nooverflow(::Val{:sloppy}, a) = SqSum_nooverflow_sloppy(a)

known_rel_bnds[(SqSum_nooverflow, :plusfp)] = known_rel_bnds[SqSum_nooverflow_plusfp]
known_rel_bnds[(SqSum_nooverflow, :sloppy)] = known_rel_bnds[SqSum_nooverflow_sloppy]


## • To avoid spurious over- and underflows, vector components are split into
##   tiny, medium, and big numbers in the classes TINY, MED, and BIG.
##
## • Some versions need to store only one class between TINY and BIG. In this
##   case, we use names: MID and EXT for middle and exterior.


function SqSum_masks(::Val{SUM}, prm::ENParam, a::Array{F}) where {SUM,F}

    (β, _, _) = getFPparam(F)
    zeroDW(F) = (h=zero(F), l=zero(F))

    BIG  = abs.(a) .> β ^ prm.emax2
    TINY = abs.(a) .< β ^ prm.emin2
    MED  = true .⊻ BIG .⊻ TINY

    @debug "[MASKS] BIG:", length(a[BIG]), " MED:", length(a[MED]), " TINY:", length(a[TINY])

    a_med = a[MED]
    Smed = !any(MED) ? zeroDW(F) :
           length(a_med) == 1 ? Fast2Mult(a_med[1], a_med[1]) :
           DWSum(Val(SUM), Fast2Mult(a_med, a_med))

    if any(BIG)
        a_big = ldexp.(a[BIG],-prm.γ)
        Sbig = length(a_big) == 1 ? Fast2Mult(a_big[1], a_big[1]) :
        DWSum(Val(SUM), Fast2Mult(a_big, a_big))

        return Sbig, Smed, zeroDW(F)
    end

    a_tiny = ldexp.(a[TINY], prm.γ)
    Stiny = !any(TINY) ? (h=zero(F), l=zero(F)) :
            length(a_tiny) == 1 ? Fast2Mult(a_tiny[1], a_tiny[1]) :
            DWSum(Val(SUM), Fast2Mult(a_tiny, a_tiny))

    if any(MED)

        return zeroDW(F), Smed, Stiny

    else

        return zeroDW(F), zeroDW(F), Stiny

    end

end


"""
    SqSum_onepass([x1, …], γ)

return the sum of it's squared array elements, using DWPlusFP

• Spurious over/underflows are prevented using scale 2^(±γ).

• Assume len(y) ≥ 2.
"""
function SqSum_onepass(::Val{SUM}, prm::ENParam, a::Array{F}) where {SUM,F}

    (β, _, _) = getFPparam(F)

    ## Initial state

    ACmid, ACext = DWAccu(Val(SUM), F), DWAccu(Val(SUM), F)
    ext_is_big = false

    ## By default, EXT represents TINY, once we encounter the first BIG, EXT
    ## is reset and represent BIG and never goes back to TINY.

    for x in a
        x_abs = abs(x)  ## Working on absolute value.

        ## `x` class identification.
        is_big  = x_abs ≥ β ^ prm.emax2
        is_tiny = x_abs < β ^ prm.emin2
        is_med  = ~is_big & ~is_tiny

        ## `is_ext` is true only when we should accumulate to it.
        is_ext = is_big | (~ext_is_big & is_tiny)

        ## Figure out if we have to reset Sext.
        is_first_big = is_big & ~ext_is_big
        is_first_big && (ACext = DWAccu(Val(SUM), F))
        ext_is_big = ext_is_big | is_big

        ## Figure out scale exponent.
        scale = ifelse(is_big, -prm.γ, prm.γ)
        scale = ifelse(is_ext, scale, zero(Int))

        x_mid = ifelse(is_med, x_abs, zero(F))
        x_ext = ifelse(is_ext, x_abs, zero(F))
        x_ext = ldexp(x_ext, scale)

        ACmid += Fast2Mult(x_mid, x_mid)
        ACext += Fast2Mult(x_ext, x_ext)
    end

    Smid = finalize(ACmid)
    Sext = finalize(ACext)

    @debug "[ONEPASS] MID.h:", Smid.h, " EXT.h:", Sext.h, " ext_is_big:", ext_is_big

    return Smid, Sext, ext_is_big
end


SqSum(::Val{:noovf}, ::Val{S}, prm, a) where {S} = SqSum_nooverflow(Val(S), a)
SqSum(::Val{:masks}, ::Val{S}, prm, a) where {S} = SqSum_masks(Val(S), prm, a)
SqSum(::Val{:onep1}, ::Val{S}, prm, a) where {S} = SqSum_onepass(Val(S), prm, a)

for A in [:noovf, :masks, :onep1]
    for S in [:plusfp, :sloppy]
        for P in [:muller, :graill]
            known_rel_bnds[(SqSum, A, S, P)] = known_rel_bnds[(SqSum_nooverflow, S)]
        end
    end
end


#######################################################################
## Euclidean Norm (EuNm)
#######################################################################


"""
    EuNm_masks([x1, …])

return the L2 norm of it's argument, using paper algorithm.

"""
function EuNm_masks(::Val{SUM}, ::Val{SQ}, prm::ENParam, a::Array{F}) where {SUM,SQ,F}

    S = SQRTDW2_masks(SqSum_masks(Val(SUM), prm, a)..., prm, Val(SQ))

    return +(S...)
end


EuNm_rel_bound(u, (m, k)) = (                     u   +
                              (     3k+2m+ 1/8) * u^2 +
                              (    -3k+2m+49/8) * u^3 +
                              ( -51/8k+6m+ 3/8) * u^4 +
                              (-123/8k+6m+99/8) * u^5 +
                              (    -9k   +   9) * u^7 ) /
                            (                     1   +
                              (    -6k-2m+   5) * u^2 +
                              (       -2m+   1) * u^3 +
                              (       -6m+   6) * u^4 +
                              (     4k-6m     ) * u^5 +
                              (     9k   -   9) * u^7 )


known_rel_bnds[(EuNm_masks, :plusfp, :muller, :muller)] = (u, n) -> EuNm_rel_bound(u, (n, 1))
known_rel_bnds[(EuNm_masks, :sloppy, :graill, :graill)] = (u, n) -> 2u


## • Here we get inspiration from Graillat et al algorithm, which does only one
##   pass through the input array.
##
## • We use the custom type DWAccu to iteratively accumulate the terms in the
##   DWPlusFP way, without introducing too many variables.

"""
    EuNm_onepass_v1([x1, …])

return the L2 norm of it's argument, using only one pass on the array.

"""
function EuNm_onepass_v1(::Val{SUM}, ::Val{SQ}, prm::ENParam, a::Array{F}) where {SUM,SQ,F}

    @assert length(a) ≤ F(2)^precision(F)

    return SQRTDW2_onepass(SqSum_onepass(Val(SUM), prm, a)..., prm, Val(SQ))
end


known_rel_bnds[(EuNm_onepass_v1, :plusfp, :muller, :muller)] = (u, n) -> EuNm_rel_bound(u, (n, 1))
known_rel_bnds[(EuNm_onepass_v1, :sloppy, :graill, :graill)] = (u, n) -> 2u


"""
    EuNm_onepass_v2([[x1, …], [y1, …], …])

return the L2 norm of it's argument, using only one pass on the array. Also
sums by blocs.

"""
function EuNm_onepass_v2(::Val{SUM}, ::Val{SQ}, prm::ENParam, a::Vector{Vector{F}}) where {SUM,SQ,F}

    @assert length(a) ≤ F(2)^precision(F)

    ## Initial state

    Smid, Sext = DWAccu(Val(SUM), F), DWAccu(Val(SUM), F)
    ext_is_big = false
    ## By default, EXT represents TINY, once we encounter the first BIG, EXT
    ## is reset and represent BIG and never goes back to TINY.

    class_sums = [SqSum_onepass(Val(SUM), prm, y) for y in a]

    for (x_mid, x_ext, is_big) in class_sums

        if is_big
            if !ext_is_big
                Sext = DWAccu(Val(SUM), F)
                ext_is_big = true
            end
            Sext += x_ext
        else
            if !ext_is_big
                Sext += x_ext
            end
        end
        Smid += x_mid
    end

    Smid = finalize(Smid)
    Sext = finalize(Sext)

    return SQRTDW2_onepass(Smid, Sext, ext_is_big, prm, Val(SQ))
end
function EuNm_onepass_v2(::Val{S}, ::Val{SQ}, prm::ENParam, a::Array{F,1}, k::Int=2) where {S,SQ,F}
    (q, r) = divrem(length(a),k)
    b = [a[1+(j-1)*q+min(j-1,r):j*q+min(j,r)] for j in 1:k]
    EuNm_onepass_v2(Val(S), Val(SQ), prm, b)
end


known_rel_bnds[(EuNm_onepass_v2, :plusfp, :muller, :muller)] = (u, n, k=2) -> EuNm_rel_bound(u, (div(n, k), k))
known_rel_bnds[(EuNm_onepass_v2, :sloppy, :graill, :graill)] = (u, n) -> 2u


#EuNm_graillat(a::Array{F,1}) where {F} = EuNm_onepass_v2(a, Val(:sloppy), ENParam_graill(F), 2)


EuNm(::Val{:masks}, ::Val{S}, ::Val{SQ}, prm::ENParam, a::Array) where {S,SQ} = EuNm_masks(Val(S), Val(SQ), prm, a)

EuNm(::Val{:onep1}, ::Val{S}, ::Val{SQ}, prm::ENParam, a::Array) where {S,SQ} = EuNm_onepass_v1(Val(S), Val(SQ), prm, a)

EuNm(::Val{:onep2}, ::Val{S}, ::Val{SQ}, prm::ENParam, a::Array, k::Int=2) where {S,SQ} = EuNm_onepass_v2(Val(S), Val(SQ), prm, a, k)


for (S, SQ, P) in [(:plusfp, :muller, :muller), (:sloppy, :graill, :graill)]
    known_rel_bnds[(EuNm, :masks, S, SQ, P)] = known_rel_bnds[(EuNm_masks, S, SQ, P)]
    known_rel_bnds[(EuNm, :onep1, S, SQ, P)] = known_rel_bnds[(EuNm_onepass_v1, S, SQ, P)]
    known_rel_bnds[(EuNm, :onep2, S, SQ, P)] = known_rel_bnds[(EuNm_onepass_v2, S, SQ, P)]
end


#######################################################################
##
## Here we add the Netlib norm for comparison


function EuNm_netlib(a::Array{F,1}) where {F}
    scale = zero(F)
    sqsum = one(F)
    for x in a
        iszero(x) && continue

        xabs = abs(x)
        if xabs > scale
            t = scale / xabs
            sqsum = 1 + sqsum * t^2
            scale = xabs
        else
            t = xabs / scale
            sqsum = sqsum + t^2
        end
    end
    return scale * sqrt(sqsum)
end


EuNm(::Val{:netlib}, ::Val{S}, ::Val{SQ}, prm::ENParam, a::Array, k::Int=2) where {S,SQ} = EuNm_netlib(a)


function EuNm_netlib_a_la_louvet(a::Array{F,1}) where {F}
    scale = floatmin(F)
    sqsum = one(F)
    for x in a
        iszero(x) && continue

        current_scale = F(2)^(1+exponent(x))
        if current_scale > scale
            t = abs(x) / current_scale
            sqsum = sqsum * (scale/current_scale)^2 + t^2
            scale = current_scale
        else
            t = abs(x) / scale
            sqsum = sqsum + t^2
        end
    end
    return scale * sqrt(sqsum)
end


EuNm(::Val{:louvet}, ::Val{S}, ::Val{SQ}, prm::ENParam, a::Array, k::Int=2) where {S,SQ} = EuNm_netlib_a_la_louvet(a)


#######################################################################
##
## Here we add the naive norm for comparison


#EuNm_naive(a::Array{F,1}) where {F} = sqrt(sum(a .^ 2))
function EuNm_naive(a::Array{F,1}) where {F}
    sqsum = zero(F)
    for x in a
        sqsum = sqsum + x^2
    end
    return sqrt(sqsum)
end


EuNm(::Val{:naive}, ::Val{S}, ::Val{SQ}, prm::ENParam, a::Array, k::Int=2) where {S,SQ} = EuNm_naive(a)


#######################################################################


function _dbg_print_section(io::IO, name)
    printstyled(io, "="^3 * " "; bold=true)
    printstyled(io, name; bold=true)
    printstyled(io, " " * "="^3; bold=true)
    println(io)
end


function _dbg_print_state(io::IO, name, val)
    print(io, name)
    for j in 1:length(val)
        @printf io "%23a" val[j]
        print(io, "     ")
        j < length(val) && print(io, "// ")
    end
    println(io)
end
