## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

export EuNm_onepass_v3


#######################################################################
## Classical building blocs.
#######################################################################


## • Unfortunately, we fail at providing an implementation that works with the
##   SIMD library.


@inline function Fast2Sum(a::Vec, b::Vec)
    #@assert abs(a) ≥ abs(b)
    s = a + b
    z = s - a
    t = b - z
    (h=s, l=t)
end


@inline function Fast2Mult(a::Vec, b::Vec)
    πh = a * b
    πl = fma(a, b, -πh)
    (h=πh, l=πl)
end


#######################################################################
## Double-word addition functions (DWAdd)
#######################################################################


@inline function DWPlusFP((xh, xl)::Tuple{Vec,Vec}, y::Vec)
    (sh, sl) = TwoSum(xh, y)
    v = xl + sl
    Fast2Sum(sh, v)
end
@inline DWPlusFP(x::NamedTuple{(:h, :l)}, y::Vec) = DWPlusFP((x.h, x.l), y)


#######################################################################
## Sum of squares functions (SqSum).
#######################################################################


"""
    vSqSum_onepass([x1, …], γ, Val(N))

return the sum of it's squared array elements, using DWPlusFP and SIMD
instructions.

• Spurious over/underflows are prevented using scale 2^(±γ).

• Assume len(y) ≥ 2.
"""
function vSqSum_onepass(prm::ENParam, a::Array{F},
                        ::Val{N}=Val(2), ::Val{DBG}=Val(false)) where {F<:AbstractFloat,N,DBG}

    (β, _, _) = getFPparam(F)

    ## Initial state

    (Smid_h, Smid_l), σmid = (zero(Vec{N,F}), zero(Vec{N,F})), zero(Vec{N,F})
    (Sext_h, Sext_l), σext = (zero(Vec{N,F}), zero(Vec{N,F})), zero(Vec{N,F})
    ext_is_big = Vec{N,Bool}(false)
    ## By default, EXT represents TINY, once we encounter the first BIG, EXT
    ## is reset and represent BIG and never goes back to TINY.

    lane = VecRange{N}(0)

    for i in 1:N:length(a)

        DBG && _dbg_print_section(stderr, "vSqSum" * @sprintf("/%04d ", i))

        x_abs = abs(a[lane+i])  ## Working on absolute value.

        ## `x` class identification.
        is_big  = x_abs ≥ β ^ prm.emax2
        is_tiny = x_abs < β ^ prm.emin2
        is_med  = ~is_big & ~is_tiny

        ## Print element information.
        DBG && begin
            print(stderr, " "^4)
            print(stderr, " x:  ")
            for j in 1:N
                @printf stderr "%23a" x_abs[j]
                is_big[j] ? printstyled(stderr, " BIG "; color=:red) :
                is_med[j] ? printstyled(stderr, " MED "; color=:yellow) :
                printstyled(stderr, " TINY"; color=:green)
                j < N && print(stderr, "// ")
            end
            println(stderr)
        end

        ## `is_ext` is true only when we should accumulate to it.
        is_ext = is_big | (~ext_is_big & is_tiny)

        ## Figure out if we have to reset Sext.
        is_first_big = is_big & ~ext_is_big
        Sext_h = vifelse(is_first_big, zero(Vec{N,F}), Sext_h)
        Sext_l = vifelse(is_first_big, zero(Vec{N,F}), Sext_l)
        σext   = vifelse(is_first_big, zero(Vec{N,F}),   σext)
        ext_is_big = ext_is_big | is_big

        ## Figure out scale exponent.
        scale = vifelse(is_big, Vec{N,F}(-prm.γ), Vec{N,F}(prm.γ))
        scale = vifelse(is_ext, scale, zero(Vec{N,F}))

        x_mid = vifelse(is_med, x_abs, zero(F))
        x_ext = vifelse(is_ext, x_abs, zero(F))
        x_ext = x_ext * exp2(scale)

        (y_mid_h, y_mid_l) = Fast2Mult(x_mid, x_mid)
        (y_ext_h, y_ext_l) = Fast2Mult(x_ext, x_ext)

        (Smid_h, Smid_l), σmid = DWPlusFP((Smid_h, Smid_l), y_mid_h), σmid + y_mid_l
        (Sext_h, Sext_l), σext = DWPlusFP((Sext_h, Sext_l), y_ext_h), σext + y_ext_l

        ## Print MID accumulator.
        DBG && begin
            printstyled(stderr, " MID"; bold=true); _dbg_print_state(stderr, " x:  ", x_mid)
            print(stderr, " "^4);                   _dbg_print_state(stderr, " y:  ", y_mid_h)
            print(stderr, " "^4);                   _dbg_print_state(stderr, "     ", y_mid_l)
            print(stderr, " "^4);                   _dbg_print_state(stderr, " S:  ", Smid_h)
            print(stderr, " "^4);                   _dbg_print_state(stderr, "     ", Smid_l)
            print(stderr, " "^4);                   _dbg_print_state(stderr, " σ:  ", σmid)
            print(stderr, )
        end

        ## Print EXT accumulator.
        DBG && begin
            printstyled(stderr, " EXT"; bold=true)
            print(stderr, "      ")
            for j in 1:N
                ext_is_big[j] ? printstyled(stderr, " BIG"; color=:red) :
                printstyled(stderr, " TINY"; color=:green)
                j < N && print(stderr, "// ")
            end
            println(stderr)
            print(stderr, " "^4);                   _dbg_print_state(stderr, " x:  ", x_ext)
            print(stderr, " "^4);                   _dbg_print_state(stderr, " y:  ", y_ext_h)
            print(stderr, " "^4);                   _dbg_print_state(stderr, "     ", y_ext_l)
            print(stderr, " "^4);                   _dbg_print_state(stderr, " S:  ", Sext_h)
            print(stderr, " "^4);                   _dbg_print_state(stderr, "     ", Sext_l)
            print(stderr, " "^4);                   _dbg_print_state(stderr, " σ:  ", σext)
            print(stderr, )
        end
    end

    (Smid_h, Smid_l) = DWPlusFP((Smid_h, Smid_l), σmid)
    (Sext_h, Sext_l) = DWPlusFP((Sext_h, Sext_l), σext)

    DBG && begin
        _dbg_print_section(stderr, "vSqSum/OVER")
        printstyled(stderr, " MID"; bold=true); _dbg_print_state(stderr, " S:  ", Smid_h)
        print(stderr, " "^4);                   _dbg_print_state(stderr, "     ", Smid_l)
        printstyled(stderr, " EXT"; bold=true)
        print(stderr, "      ")
        for j in 1:N
            ext_is_big[j] ? printstyled(stderr, " BIG"; color=:red) :
            printstyled(stderr, " TINY"; color=:green)
            j < N && print(stderr, "// ")
        end
        println(stderr)
        print(stderr, " "^4);                   _dbg_print_state(stderr, " S:  ", Sext_h)
        print(stderr, " "^4);                   _dbg_print_state(stderr, "     ", Sext_l)
    end

    Smid = (NTuple{N,F}(Smid_h), NTuple{N,F}(Smid_l))
    Sext = (NTuple{N,F}(Sext_h), NTuple{N,F}(Sext_l))
    ext_is_big2 = NTuple{N,Bool}(ext_is_big)

    return Smid, Sext, ext_is_big2
end


#######################################################################
## Euclidean Norm (EuNm)
#######################################################################


function EuNm_onepass_v3(::Val{SUM}, ::Val{SQ}, prm::ENParam, a::Array{F},
                         ::Val{N}=Val(2), ::Val{DBG}=Val(false)) where {SUM,SQ,F,N,DBG}

    @assert length(a) ≤ 2^precision(F)

    ## Initial state

    Smid, Sext = DWAccu_plusfp(F), DWAccu_plusfp(F)
    ext_is_big = false
    ## By default, EXT represents TINY, once we encounter the first BIG, EXT
    ## is reset and represent BIG and never goes back to TINY.

    x_mid, x_ext, is_big = vSqSum_onepass(prm, a, Val(N), Val(DBG))

    for i in 1:N

        if is_big[i]
            if !ext_is_big
                Sext = DWAccu_plusfp(F)
                ext_is_big = true
            end
            Sext += (h=x_ext[1][i], l=x_ext[2][i])
        else
            if !ext_is_big
                Sext += (h=x_ext[1][i], l=x_ext[2][i])
            end
        end
        Smid += (h=x_mid[1][i], l=x_mid[2][i])
    end

    Smid = finalize(Smid)
    Sext = finalize(Sext)

    DBG && begin
        _dbg_print_section(stderr, "vSaSum/Final")
        printstyled(stderr, " MID"; bold=true); print(stderr, " S:   ")
        @printf stderr "%23a" Smid.h
        println(stderr)
        print(stderr, " "^4); print(stderr, "      ")
        @printf stderr "%23a" Smid.l
        println(stderr)
        printstyled(stderr, " EXT"; bold=true);
        ext_is_big ? printstyled(stderr, " BIG"; color=:red) :
        printstyled(stderr, " TINY"; color=:green)
        println(stderr)
        print(stderr, " "^4); print(stderr, " S:   ")
        @printf stderr "%23a" Sext.h
        println(stderr)
        print(stderr, " "^4); print(stderr, "      ")
        @printf stderr "%23a" Sext.l
        println(stderr)
    end

    return SQRTDW2_onepass(Smid, Sext, ext_is_big, prm, Val(SQ))
end


EuNm(::Val{:onep3}, ::Val{S}, ::Val{SQ}, prm::ENParam, a::Array,
     ::Val{N}=Val(2), ::Val{DBG}=Val(false)) where {S,SQ,N,DBG} = EuNm_onepass_v3(Val(S), Val(SQ), prm, a, Val(N), Val(DBG))
