## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

import Base:big, convert
import Base:one, zero
import Base:convert
import Base:ldexp
import Base.==
import Base:isless
import Random: AbstractRNG, Sampler, rand

export checkDW, assertDW
export DW, hi, lo
export DWT, DWTtoBig


## There is three ways to implement double-word numbers:
##
##   1. using plain tuples ``(xh, xl)``,
##   2. using named tuples ``(h=xh, l=xl)``,
##   3. using a struct ``x.h, x.l``.
##
## Each approach has pros and cons: using a ``struct`` comes with a new type,
## which helps to write readable code, to extend existing functions, to
## dispatch more easily, and to guaranty well-normalized numbers.
##
## This approach seems to be the best until arrays comes into play: many
## algorithms works faster with separate upper and lower part arrays,
## especially with vectorized functions that works only with arrays of
## floating-point numbers.
##
## Thereforre we decided to make the code compatible with tuples. NamedTuple
## is a readibility compromise.


"""
    checkDW(h, l)

return true or false whether ``(h, l)`` is a valid ``DW``.
"""
checkDW(h, l) = (h + l == h)


"""
    assertDW(h, l)

Raise an error if ``(h, l)`` is not a valid ``DW``.
"""
assertDW(h, l) = checkDW(h, l) ? (h, l) : error(string("xh ≠ RN(x) with ", h, "+", h+l))


#######################################################################
## DW struct
#######################################################################


"""
    DW(xh, xl)

Double world number.
"""
struct DW{F<:AbstractFloat} <: Real
    h::F
    l::F
    ## Uncomment line wether you want to check.
    #DW(h::F, l::F=zero(F)) where {F<:AbstractFloat} = new{F}(h, l)
    DW(h::F, l::F=zero(F)) where {F<:AbstractFloat} = new{F}(assertDW(h, l)...)
end


## Functional access to upper and lower parts.
hi(d::DW) = d.h
lo(d::DW) = d.l


## Very useful in generic code.
zero(d::DW{F}) where {F<:AbstractFloat} = DW(zero(F))
one(d::DW{F}) where {F<:AbstractFloat} = DW(one(F))


## A _normal_ double-world number has a normal lower part.
Base.floatmin(::Type{DW{F}}) where {F<:AbstractFloat} = DW(ldexp(floatmin(F), precision(F)))
Base.floatmax(::Type{DW{F}}) where {F<:AbstractFloat} = DW(floatmax(F), ldexp(floatmax(F), -precision(F)-1))
Base.exponent(d::DW) = exponent(d.h)


## DW equality is the equality of its fields, which is not the current behavior
## for structs.
##
## https://github.com/JuliaLang/julia/issues/4648
==(x::DW, y::DW) = (x.h == y.h) && (x.l == y.l)


isless(x::DW, y::DW) = isless(x.h, y.h)


## BigFloat conversion.

big(d::DW) = big(d.h) + big(d.l)


## Other conversion.

convert(::Type{DW{F}}, x::F) where {F<:AbstractFloat} = DW(x)
convert(::Type{DW{F}}, x::Int) where {F<:AbstractFloat} = DW(convert(F, x))
convert(::Type{DW{F}}, x::Tuple{F,F}) where {F<:AbstractFloat} = DW(x...)
convert(::Type{Tuple{F,F}}, x::DW{F}) where {F<:AbstractFloat} = (x.h, x.l)
convert(::Type{BigFloat}, d::DW) = big(d.h) + big(d.l)


## Scale

ldexp(x::DW, n::Int) = DW(ldexp(x.h, n), ldexp(x.l, n))


## Random numbers

function rand(RNG::AbstractRNG, ::Sampler{DW{F}}) where {F<:AbstractFloat}

    xh = rand(RNG, F)
    iszero(xh) && return DW(xh)
    s = rand([1, -1])
    xl = if ispow2(xh) && s == -1  ## Beware rounding in this case !
        ldexp(rand(RNG, F), exponent(xh) - precision(xh) - 1)
    else
        ldexp(rand(RNG, F), exponent(xh) - precision(xh))
    end
    DW(xh, flipsign(xl, s))
end


#######################################################################
## DW as a named tuple
#######################################################################


const DWT{F<:AbstractFloat} = NamedTuple{(:h, :l), Tuple{F,F}}

## !! Julia does not differentiate ``DWT`` and it's definition. Hence it is not
##    possible to write a ``rand`` method for a general named tuple and for a
##    ``DWT``.


ldexp(x::DWT, n::Int) = (h=ldexp(x.h, n), l=ldexp(x.l, n))


convert(::Type{DWT{F}}, x::DW{F}) where {F<:AbstractFloat} = (h=x.h, l=x.l)
convert(::Type{DW{F}}, x::DWT{F}) where {F<:AbstractFloat} = DW(x.h, x.l)

DWTtoBig(x::DWT) = big(x.h) + big(x.l)
##
## Typically one thing we cannot add in ``convert`` methods
