## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

using Random

export hyprand
export hyprand2


#######################################################################
## Custom random number generator
#######################################################################


## Allen B. Downey, “Generating Pseudo-random Floating-Point Values”, July 25,2007
## https://www.allendowney.com/research/rand/downey07randfloat.pdf


"""
    hyprand([rng=GLOBAL_RNG], S, [(emin=exponent(floatmin(S)), emax=0)], dims...)

random numbers with uniform exponent distribution between 2^emin to 2^emax.
"""
function hyprand(rng::AbstractRNG, ::Type{S}, (Emin, Emax)::Tuple{Int,Int}) where {S<:AbstractFloat}
    β, p = big(2), precision(S)
    e = rand(Emin+1:Emax)
    m = rand(β^(p-1):(β^p)-1)
    ldexp(convert(S, m), e-p)
end
hyprand(rng::AbstractRNG, ::Type{S}, E::Tuple{Int,Int}, dims::Vararg{Integer}) where {S<:AbstractFloat} = hyprand.(rng, S, fill(E, dims...))

hyprand(rng::AbstractRNG, ::Type{S}) where {S<:AbstractFloat} = hyprand(rng, S, (exponent(floatmin(S)), 0))
hyprand(rng::AbstractRNG, ::Type{S}, dims::Vararg{Integer}) where {S<:AbstractFloat} = hyprand(rng, S, (exponent(floatmin(S)), 0), dims...)

hyprand(::Type{S}, E::Tuple{Int,Int}) where {S<:AbstractFloat} = hyprand(Random.GLOBAL_RNG, S, E)
hyprand(::Type{S}, E::Tuple{Int,Int}, dims::Vararg{Integer}) where {S<:AbstractFloat} = hyprand(Random.GLOBAL_RNG, S, E, dims...)
hyprand(::Type{S}) where {S<:AbstractFloat} = hyprand(Random.GLOBAL_RNG, S)
hyprand(::Type{S}, dims::Vararg{Integer}) where {S<:AbstractFloat} = hyprand(Random.GLOBAL_RNG, S, dims...)


## Double-world random number generator


function hyprand(rng::AbstractRNG, ::Type{DW{S}}, (Emin, Emax)::Tuple{Int,Int}) where {S<:AbstractFloat}
    β, p = big(2), precision(S)
    emin = exponent(floatmin(S))

    xh = hyprand(rng, S, (Emin, Emax))

    if exponent(xh)-p ≤ emin+1 ## If xl is a subnormal number.
        xl = zero(S)

    else
        s = rand([-1, 1])
        e = rand(emin+1:exponent(xh)-p) ## TODO: control emin.
        m = rand(β^(p-1):β^p) ## We add β^p to represent zero
        el = e-p
        xl = (m == β^p) ? zero(S) : copysign(ldexp(convert(S, m), el), s)
    end

    DW(xh, xl)
end
hyprand(rng::AbstractRNG, ::Type{DW{S}}) where {S<:AbstractFloat} = hyprand(rng, DW{S}, (exponent(floatmin(DW{S})), 0))

hyprand(::Type{DW{S}}, E::Tuple{Int,Int}) where {S<:AbstractFloat} = hyprand(Random.GLOBAL_RNG, DW{S}, E)
hyprand(::Type{DW{S}}) where {S<:AbstractFloat}  = hyprand(Random.GLOBAL_RNG, DW{S})

hyprand(rng::AbstractRNG, ::Type{DW{S}}, E::Tuple{Int,Int}, dims...) where {S<:AbstractFloat} = hyprand.(rng, DW{S}, fill(E, dims...))
hyprand(rng::AbstractRNG, ::Type{DW{S}}, dims...) where {S<:AbstractFloat} = hyprand.(rng, fill(DW{S}, dims...))
hyprand(::Type{DW{S}}, E::Tuple{Int,Int}, dims...) where {S<:AbstractFloat} = hyprand(Random.GLOBAL_RNG, DW{S}, E, dims...)
hyprand(::Type{DW{S}}, dims...) where {S<:AbstractFloat} = hyprand(Random.GLOBAL_RNG, DW{S}, dims...)


#######################################################################
## Experimental alternative
#######################################################################

## Generated numbers are squares of floating-point numbers


function hyprand2(rng::AbstractRNG, ::Type{S}, (Emin, Emax)::Tuple{Int,Int}) where {S<:AbstractFloat}
    z = hyprand(S, (div(Emin, 2), div(Emax, 2)))
    z^2
end
hyprand2(rng::AbstractRNG, ::Type{S}) where {S<:AbstractFloat} = hyprand2(rng, S, (exponent(floatmin(S)), 0))
function hyprand2(rng::AbstractRNG, ::Type{DW{S}}, (Emin, Emax)::Tuple{Int,Int}) where {S<:AbstractFloat}
    y = hyprand(S, (div(Emin, 2), div(Emax, 2)))
    ## Fast2Mult
    zh = y * y
    zl = fma(y, y, -zh)
    DW(zh, zl)
end
hyprand2(rng::AbstractRNG, ::Type{DW{S}}) where {S<:AbstractFloat} = hyprand2(rng, DW{S}, (exponent(floatmin(DW{S})), 0))

hyprand2(rng::AbstractRNG, ::Type{S}, E::Tuple{Int,Int}, dims...) where {S} = hyprand2.(rng, S, fill(E, dims...))
hyprand2(rng::AbstractRNG, ::Type{S}, dims...) where {S<:AbstractFloat} = hyprand2(rng, S, (exponent(floatmin(S)), 0), dims...)
hyprand2(rng::AbstractRNG, ::Type{DW{S}}, dims...) where {S<:AbstractFloat} = hyprand2(rng, DW{S}, (exponent(floatmin(S)), 0), dims...)

hyprand2(::Type{S}, E::Tuple{Int,Int}) where {S} = hyprand2(Random.GLOBAL_RNG, S, E)
hyprand2(::Type{S}, E::Tuple{Int,Int}, dims...) where {S} = hyprand2(Random.GLOBAL_RNG, S, E, dims...)
hyprand2(::Type{S}) where {S} = hyprand2(Random.GLOBAL_RNG, S)
hyprand2(::Type{S}, dims...) where {S} = hyprand2(Random.GLOBAL_RNG, S, dims...)
