## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

module Tables

using Printf


#######################################################################
## Public building blocs
#######################################################################


function title(name; kwargs...)

    printstyled(@sprintf "# [%s]" name; bold=true, color=:yellow)
    for (key, val) in kwargs
        printstyled(@sprintf " %s:%s" key val; color=:yellow)
    end
    println()
end


lrpad(name, len) = rpad(lpad(name, len+1), len+2)
sep(; color=:white) = printstyled("|"; color=color)
fstring(x) = @sprintf("%6.2e", x)


function head(values::Vararg{Tuple{String,Int}}; color=:yellow)

    print(" ")
    for (name, len) in values
        printstyled(lrpad(name, len); color=color)
        sep(color=color)
    end
    sep(color=color)
end


function head_rounds(; color=:yellow)
    printstyled(@sprintf "| faithful correct |"; color=:yellow)
end


function vals(values::Vararg{Tuple{String,Int}}; color=:white)

    print(" ")
    for (name, len) in values
        printstyled(lrpad(name, len); color=color)
        sep(color=color)
    end
end


function vals(values::Vararg{Tuple{AbstractFloat,Int}}; color=:white)

    values = [(fstring(x), len) for (x, len) in values]
    vals(values...; color=color)
end


function check((name, len)::Tuple{String,Int}, checked::Bool; color=:white)
    print(" ")
    val = string(lrpad(name, len), (checked ? "✔" : "✘"), " ")
    printstyled(val; color=(checked ? :green : :red))
    sep(color=color)
end


function check((x, len)::Tuple{AbstractFloat,Int}, checked::Bool; color=:white)
    check((fstring(x), len), checked; color=color)
end


function rounds(faithfuls, corrects; color=:white)
        sep(color=color)
        printstyled(@sprintf " %7d " faithfuls)
        printstyled(@sprintf " %7d " corrects)
        sep(color=color)
end


getUnit(unit, F) =
    if unit == :u
        eps(F) / 2
    elseif unit == :u²
        (eps(F) / 2) ^ 2
    elseif unit == :ε
        eps(F)
    elseif unit == :ε²
        eps(F) ^ 2
    else
        convert(F, 1)
    end


end
