## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

export getFPparam, getFPeps, getFPrdoff
export getENscale_graillat, getENscale_default
export getENparam_muller, getENparam_graillat
export getENbounds_muller, getENbounds_graillat
export ENParam
export showENparams
export rel_error, ulp_error, is_correct, is_faithful
export hist
export inexp


#######################################################################
## Some backports
#######################################################################


if VERSION < v"1.6"
    ispow2(x::AbstractFloat) = !iszero(x) && frexp(x)[1] == 0.5
end


#######################################################################
## Floating-point number parameters.
#######################################################################


"""
    getFPparam(F<:AbstractFloat)

return radix, precision, and exponent bounds.

The radix type is `F`, which is convenient in some formulas.
"""
function getFPparam(::Type{F}) where {F<:AbstractFloat}
    β = convert(F, 2)
    p = precision(F)
    emax = exponent(floatmax(F))
    emin = exponent(floatmin(F))
    (β, p, (emin, emax))
end


"""
    getFPeps(β, p)
    getFPeps(F<:AbstractFloat)

return ε such as |R(x) - x| < ε·x for any rounding mode
"""
getFPeps(β, p) = β^(1-p)
getFPeps(::Type{F}) where {F<:AbstractFloat} =
    let (β,p,_) = getFPparam(F); getFPeps(β, p) end
##
## This is in fact equivalent to eps(F)


"""
    getFPrdoff(β, p)
    getFPrdoff(F<:AbstractFloat)

return u such as |RN(x) - x| ≤ u·x, i.e. the roundoff unit
"""
getFPrdoff(β, p) = β^-p
getFPrdoff(::Type{F}) where {F<:AbstractFloat} =
    let (β,p,_) = getFPparam(F); getFPrdoff(β, p) end
##
## This is u.


#######################################################################
## Euclidean Norm parameters.
#######################################################################

## These parameters are: the scaling exponent ``γ`` and exponents of the
## classes boundaries ``(emin2, emax2)``. There is two different computation of
## these parameters, depending to the authors.


## Euclidean Norm scaling.


getENscale_graillat(β, p, (emin, emax)) = 2cld(emax - emin + p - 1, 3*2)
getENscale_graillat(::Type{F}) where {F<:AbstractFloat} = getENscale_graillat(getFPparam(F)...)


getENscale_default(F::Type) = string(F) in keys(_DEFAULT_SCALE) ? _DEFAULT_SCALE[string(F)] :
                              error(" no default scale value for this format: $F")
##
## There is no real argument for these default values, there should be one.


## Euclidean Norm class boundaries.


"""
    getENparam_muller(β, p, (emin, emax))
    getENparam_muller(F<:AbstractFloat)

return exponents of minmed and maxmed
"""
getENparam_muller(β, p, (emin, emax)) = (cld(emin+p, 2), fld(emax-p, 2))
getENparam_muller(::Type{F}) where {F<:AbstractFloat} = getENparam_muller(getFPparam(F)...)


"""
    getENparam_graillat(β, p, (emin, emax))
    getENparam_graillat(F<:AbstractFloat)

return exponents of minmed and maxmed
"""
function getENparam_graillat(β, p, (emin, emax))
    γ = getENscale_graillat(β, p, (emin, emax))
    (emax + 1 - 2γ, emax + 1 - γ)
end
getENparam_graillat(::Type{F}) where {F<:AbstractFloat} = getENparam_graillat(getFPparam(F)...)


"""
    getENbounds_muller(β, p, (emin, emax))
    getENbounds_muller(F<:AbstractFloat)
(radix, precision, exponent bounds)

return minmed and maxmed
"""
getENbounds_muller(β, p, (emin, emax)) = β .^ getENparam_muller(β, p, (emin, emax))
getENbounds_muller(::Type{F}) where {F<:AbstractFloat} =
    let (β,p,ebnds) = getFPparam(F); getENbounds_muller(β, p, ebnds) end


"""
    getENbounds_graillat(β, p, (emin, emax))
    getENbounds_graillat(F<:AbstractFloat)
(radix, precision, exponent bounds)

return minmed and maxmed
"""
getENbounds_graillat(β, p, (emin, emax)) = β .^ getENparam_graillat(β, p, (emin, emax))
getENbounds_graillat(::Type{F}) where {F<:AbstractFloat} =
    let (β,p,ebnds) = getFPparam(F); getENbounds_graillat(β, p, ebnds) end


#######################################################################
## Euclidean Parameter struct.
#######################################################################


struct ENParam
    F::Type                ## Floating point type
    emin2::Int             ## minmed exponent
    emax2::Int             ## maxmed exponent
    γ::Union{Int,Missing}  ## scale t exponent
    function ENParam(F, (emin2, emax2), γ; validate=:check)
        msg = if validate == :yes
            error
        elseif   validate == :check
            m -> @error(m)
        elseif   validate == :no
            m -> nothing
        else
            error("invalid argument")
        end

        ## Constructor that checks values.
        (β, p, (emin, emax)) = getFPparam(F)

        emin2 < emax2 || msg("ENParam: emin2 ≮ emax2")
        emin + p ≤ 2emin2 || msg("ENParam: emin2 too small to prevent underflow")
        2emax2 + p ≤ emax || msg("ENParam: emax2 too big to prevent overflow")

        if !ismissing(γ)
            iseven(γ) || msg("ENParam: γ should be even")
            emax - γ ≤ emax2  || msg("ENParam: γ to big to prevent overflow of BIG")
            emin2 ≤ emax2 - γ || msg("ENParam: γ to small to prevent underflow of BIG")
            emin2 + γ ≤ emax2 || msg("ENParam: γ to big to prevent overflow of TINY")
            emin2 ≤ emin + γ  || msg("ENParam: γ to small to prevent underflow of TINY")
        end

        new(F, emin2, emax2, γ)
    end
end


ENParam(::Val{:muller}, F, γ; validate=:check) = ENParam(F, getENparam_muller(F), γ; validate=validate)
ENParam(::Val{:muller}, F; validate=:check) =    ENParam(F, getENparam_muller(F), getENscale_default(F); validate=validate)
ENParam(::Val{:graill}, F; validate=:check) =    ENParam(F, getENparam_graillat(F), getENscale_graillat(F); validate=validate)


#######################################################################
## Parameter diagnostics.
#######################################################################


function showENparams(io::IO, β, p, (emin, emax), (emin2, emax2), γ=missing)

    ε = getFPeps(β, p)
    eη = (emin + p) / 2

    printstyled(io, " p: "; color=:yellow)
    printstyled(io, p)
    printstyled(io, " (emin, emax): "; color=:yellow)
    printstyled(io, (emin, emax))
    println(io)
    printstyled(io, " 1+ε: "; color=:yellow)
    printstyled(io, @sprintf "%a" 1+ε)
    println(io)

    println(io)
    printstyled(io, " log2(η): "; color=:yellow)
    printstyled(io, @sprintf "%5.1f" eη)
    println(io)
    printstyled(io, " (log2(minmed), log2(maxmed)): "; color=:yellow)
    printstyled(io, @sprintf "(%5d, %5d)" emin2 emax2)
    println(io)

    println(io)
    printstyled(io, "      log2(Ω/maxmed): "; color=:yellow)
    printstyled(io, @sprintf "%5d" emax-emax2)
    println(io)
    printstyled(io, "      log2(minmed/α): "; color=:yellow)
    printstyled(io, @sprintf "%5d" emin2-emin+p-1)
    println(io)
    printstyled(io, " log2(maxmed/minmed): "; color=:yellow)
    printstyled(io, @sprintf "%5d" emax2-emin2)
    println(io)
    printstyled(io, "             Card{γ}: "; color=:yellow)
    printstyled(io, @sprintf "%5d" div(emax2-emin2 -emin2+emin-p+1, 2))
    println(io)

    println(io)
    printstyled(io, "  BIG ∈ "; color=:yellow)
    printstyled(io, @sprintf "%6s %5s %5d %5d" "" "" emax2 emax+1)
    printstyled(io, " : "; color=:yellow)
    printstyled(io, @sprintf "%4d" emax+1-emax2)
    println(io)
    printstyled(io, "  MED ∈ "; color=:yellow)
    printstyled(io, @sprintf "%6s %5d %5d %5s" "" emin2 emax2 "" )
    printstyled(io, " : "; color=:yellow)
    printstyled(io, @sprintf "%4d" emax2-emin2)
    println(io)
    printstyled(io, " TINY ∈ "; color=:yellow)
    printstyled(io, @sprintf "%6d %5d %5s %5s" emin emin2 "" ""  )
    printstyled(io, " : "; color=:yellow)
    printstyled(io, @sprintf "%4d" emin2-emin)
    println(io)

    ismissing(γ) && return

    printstyled(io, " t·BIG ∈       "; color=:yellow)
    printstyled(io, @sprintf "%5d %5d" emax2-γ emax+1-γ)
    println(io)
    printstyled(io, "t·TINY ∈       "; color=:yellow)
    printstyled(io, @sprintf "%5d %5d" emin+γ emin2+γ  )
    println(io)

    println(io)
    printstyled(io, " log2(maxmed²u²): "; color=:yellow)
    printstyled(io, @sprintf "%5d" (2emax2 - 2p))
    println(io)
    printstyled(io, " log2(maxmed²t²/u³): "; color=:yellow)
    printstyled(io, @sprintf "%5d" (2emax2 - 2γ  + 3p))
    println(io)
end
showENparams(io, p::ENParam) = showENparams(io, getFPparam(p.F)..., (p.emin2, p.emax2), p.γ)

showENparams(β, p, ebnds, ebnds2, γ=missing) = showENparams(stdout, β, p, ebnds, ebnds2, γ)
showENparams(p::ENParam, γ=missing) = showENparams(stdout, p)


#######################################################################
## Error diagnostics
#######################################################################


"""
    rel_error(value, exact)

compute the error relative to the exact result.
"""
rel_error(value, exact) = iszero(exact) ? iszero(value) ? zero(exact) : typemax(exact) :
                          abs( (value - exact) / exact )

"""
    ulp_error(value, exact)

compute the error relative to the ulp of the exact result.
"""
ulp_error(value, exact, ulp=eps(exact)) = iszero(exact) ? iszero(value) ? zero(exact) : typemax(exact) :
                                          abs( (value - exact) / ulp )

"""
    is_correct(value, exact)

return true if value is correctly rounded to exact.
"""
is_correct(value::F, exact::BigFloat) where {F<:AbstractFloat} = convert(F, exact) == value

"""
    is_faithful(value, exact)

return true if value is faithfully rounded to exact.
"""
function is_faithful(value::F, exact::BigFloat) where {F<:AbstractFloat}
    setprecision(BigFloat, precision(F)) do
        exactD = setrounding(BigFloat, RoundDown) do
            convert(F, exact * 1)
        end
        exactU = setrounding(BigFloat, RoundUp) do
            convert(F, exact * 1)
        end
        exactD ≤ value ≤ exactU
    end
end


#######################################################################
## Histograms
#######################################################################


function hist(v, edges)

    n = length(edges)
    bins = zeros(Int, n)

    for val in v
        i = first(indexin(true, edges[1:n-1] .≥ val))
        i = isnothing(i) ? n : i
        bins[i] += 1
    end

    return bins
end


#######################################################################
## Error functions
#######################################################################


## [Higham, 2002] Lemma 3.1. p. 63
##
## |x·y - fl(x·y)| ≤ γ_n |x·y|
##
## with x,y ∈ ℝ^n and · inner product.

γ_inner(ν) = ν / (1 - ν)


#######################################################################
## Convenient printing
#######################################################################


function myshow(x::F) where {F<:AbstractFloat}
    s = sign(x)
    (m, e) = (significand(x), exponent(x))
    M = let p = precision(F)
        trunc(Int, ldexp(abs(m), p+mod(e, 4)))
    end

    string(s ≥ 0 ? " " : "-", @sprintf "0x%Xp%+d" M 4fld(e, 4))
end
myshow(x::Tuple{F,F}) where {F<:AbstractFloat} = string("(", myshow(x[1]), ", ", myshow(x[2]), ")")


#######################################################################
## https://discourse.julialang.org/t/significand-of-ieee-float-as-an-integer/3698/3


import Base: sign_mask, significand_mask, exponent_mask, Math.significand_bits, Math.exponent_bias


for (F, I) in ((Float32, Int32), (Float64, Int64), (Float128, Int128))
    @eval function inexp(x::$F)
        xu = reinterpret(Unsigned, x)
        xs = xu & ~sign_mask($F)
        xs >= exponent_mask($F) && return throw(DomainError())

        k = $I(xs >> significand_bits($F))
        if k == 0 # x is subnormal
            m = leading_zeros(xs) - exponent_bits($F)
            xs <<= unsigned(m)
            xu = xs | (xu & sign_mask($F))
            k = 1 - m
        end
        k -= exponent_bias($F)
        k -= significand_bits($F)

        xm = xu  & significand_mask($F)
        int = $I(xm | 1 << significand_bits($F))
        return (signbit(x) ? -int : int), k
    end
end
