## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

using Printf
using Random
using Test

using EuNms


## Tweaking the output of a DW when test fails.
Base.show(io::IO, x::Tuple{F,F}) where {F<:AbstractFloat} = @printf io "%a %+a" x[1] x[2]

@testset "DWPlusFP: worst cases" begin

    @testset "Using type $F"  for F in [[Float64, Float32, Float16, BigFloat]
                                        @isdefined(Float128) ? [Float128] : []
                                        @isdefined(BFloat16) ? [BFloat16] : []]
        u = getFPrdoff(F)
        u⁰, u², u³ = u^0, u^2, u^3

        @test (h=u⁰ + 2u, l=0u) == DWPlusFP((u⁰+2u, -u+u²), u)
        @test (h=u⁰ + 2u, l=0u) == DWPlusFP((u⁰, u-u²), u)
    end
end


@testset "SloppyDWPlusDW: worst case" begin

    @testset "Using type $F"  for F in [[Float64, Float32, BigFloat]
                                        @isdefined(Float128) ? [Float128] : []
                                        @isdefined(BFloat16) ? [BFloat16] : []]
        u = getFPrdoff(F)
        u⁰, u², u³ = u^0, u^2, u^3

        @test (h=u⁰ + 10u, l=-8u²) == SloppyDWPlusDW((u⁰+2u, -u+u²), (9u, -6u²-8u³))
    end

    ## Implementation of Float16 have trouble when rounding to nearest, ties to
    ## even.
    ##
    @testset "Using type $F"  for F in [Float16]
        u = getFPrdoff(F)
        u⁰, u², u³ = u^0, u^2, u^3

        @test_broken (h=u⁰ + 10u, l=-8u²) == SloppyDWPlusDW((u⁰+2u, -u+u²), (9u, -6u²-8u³))
    end
end


@testset "DW interface" begin

    @testset "Using type $F"  for F in [[Float64, Float32, Float16, BigFloat]
                                        @isdefined(Float128) ? [Float128] : []
                                        @isdefined(BFloat16) ? [BFloat16] : []]

        u = getFPrdoff(F)
        u⁰, u², u³ = u^0, u^2, u^3

        @test DW(u⁰ + 2u, 0u) == DWPlusFP(DW(u⁰+2u, -u+u²), u)
        @test DW(u⁰ + 2u, 0u) == DWPlusFP(DW(u⁰, u-u²), u)
        F != Float16 && @test DW(u⁰ + 10u, -8u²) == SloppyDWPlusDW(DW(u⁰+2u, -u+u²), DW(9u, -6u²-8u³))
    end
end
