## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

#using Random
using Test

using EuNms


macro mytest(ex)
    op, a, b = ex.args
    quote
        @debug string($(string(ex)),
                      "\n", $(esc(a)) |> bitstring,
                      "\n", $(esc(b)) |> bitstring)
        @test $(esc(ex))
    end
end


@testset "Validate overflow classes" begin
    @testset "Algo: $A/$S/$Q" for (Q,S,A) in Iterators.product([:muller,:graill], [:plusfp, :sloppy], [:masks, :onep1, :onep2, :onep3])

        E(a) = EuNm(Val(A), Val(S), Val(Q), ENParam(Val(:muller), Float64), a)

        ## Cases with one class only.

        ## • MED
        @mytest E(Float64[1,2,3,4]) == sqrt(30)
        ## • BIG
        @mytest E(Float64[1e300,2e300,3e300,4e300]) == sqrt(30) * 1e300
        ## • TINY
        @mytest E(Float64[1e-300,2e-300,3e-300,4e-300]) == sqrt(30) * 1e-300

        ## Cases with two adjacent classes, and the smaller class is not neglected.

        ## • BIG and MED
        @mytest E(Float64[1e146,2e146,3e145,4e145]) == sqrt(525) * 1e145
        ## • MED and TINY
        @mytest E(Float64[1e-145,2e-145,3e-146,4e-146]) == sqrt(525) * 1e-146

        ## Tests targeted to Graillat et al. code (≈ EuNm_v2).
        ## • Floats are close to each other when scaled.

        ## • TINY, then BIG
        @mytest E(Float64[0x1p-700,0x2p-700,0x3p700,0x4p700]) == 5 * 0x1p700
        ## • BIG, then TINY
        @mytest E(Float64[0x1p700,0x2p700,0x3p-700,0x3p-700]) == sqrt(5) * 0x1p700
    end
end


@testset "A fistfull of zero" begin
    @testset "Algo: $A/$S/$Q" for (Q,S,A) in Iterators.product([:muller,:graill], [:plusfp, :sloppy], [:masks, :onep1, :onep2, :onep3])

        E(a) = EuNm(Val(A), Val(S), Val(Q), ENParam(Val(:muller), Float64), a)

        @test E(Float64[]) == zero(Float64)
        @test E(Float64[0, 0, 0, 0]) == zero(Float64)
        @test E(Float64[   1e0, 0, 0, 0]) == 1e0
        @test E(Float64[ 1e300, 0, 0, 0]) == 1e300
        @test E(Float64[1e-300, 0, 0, 0]) == 1e-300

    end
end


@testset "Hard cases" begin
    @testset "Algo: $A/$S/$Q" for (Q,S,A) in Iterators.product([:muller,:graill], [:plusfp, :sloppy], [:masks, :onep1, :onep2, :onep3])

        E(a) = EuNm(Val(A), Val(S), Val(Q), ENParam(Val(:muller), Float64), a)

        ## Tests targeted the EXT behavior of the onepass variant.
        ## • When EXT changes from TINY accumulator to BIG accumulator, its content
        ##   should be reset.
        ## • Not for SIMD versions.

        let x = [0x1.0000002p-500, ## TINY, yield Sext=(0x1.0000004p+400+0x0p0)
                 0x1.0000002p-514, ## TINY, fills the lower part of Sext
                 0x1.0000002p+500, ## TINY, should reset Sext to (0x1.0000004p-400+0x0p0),
                 0x1.0000002p+486, ##       Without reset, this term would be neglected.
                ]                  ## Works only for γ=700
            @mytest E(x) == 0x1.00000028p+500
        end
    end
end


# @testset "Graillat et al. version" begin
# 
#         ## • MED
#         @mytest EuNm_graillat(Float64[1,2,3,4]) == sqrt(30)
#         ## • BIG
#         @mytest EuNm_graillat(Float64[1e300,2e300,3e300,4e300]) == sqrt(30) * 1e300
#         ## • TINY
#         @mytest EuNm_graillat(Float64[1e-300,2e-300,3e-300,4e-300]) == sqrt(30) * 1e-300
# 
# end
