## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

using Random
using Test

using EuNms


@testset "SQRTDWtoDW: even powers exactness" begin

    @testset "Using type $F"  for F in [[Float64, Float32, Float16, BigFloat]
                                        @isdefined(Float128) ? [Float128] : []
                                        @isdefined(BFloat16) ? [BFloat16] : []
                                       ]
        β, p, ebounds = getFPparam(F)
        u = getFPrdoff(β, p)
        (emin2, emax2) = getENparam_muller(β, p, ebounds)

        for k in range(emin2, stop=emax2, step=1+div(emax2-emin2, 256))
            @test (h=β^(k), l=0u) == SQRTDWtoDW((β^(2k), 0u))
        end
    end
end


@testset "SQRTDWtoDW vs sqrt" begin

    @testset "Using type $F"  for F in [[Float64, Float32, Float16, BigFloat]
                                        @isdefined(Float128) ? [Float128] : []
                                        @isdefined(BFloat16) ? [BFloat16] : []
                                        ]
        u = getFPrdoff(F)
        u⁰, u², u³ = u^0, u^2, u^3

        @test SQRTDWtoDW((2.0, 0.0)).h ≈ sqrt(2.0)
        @test SQRTDWtoDW((5.0, 0.0)).h ≈ sqrt(5.0)
        @test SQRTDWtoDW((1.1e-35, 0.0)).h ≈ sqrt(1.1e-35)
        @test SQRTDWtoDW((1.1e35, 0.0)).h ≈ sqrt(1.1e35)
    end
end


@testset "SQRTDWto{FP,DW}: zero" begin

    @testset "Using type $F"  for F in [[Float64, Float32, Float16, BigFloat]
                                        @isdefined(Float128) ? [Float128] : []
                                        @isdefined(BFloat16) ? [BFloat16] : []
                                       ]
        @test_broken SQRTDWtoDW((0.0, 0.0)) == (h=0.0, l=0.0)
        @test_broken SQRTDWtoFP((0.0, 0.0)) == 0.0
    end
end
