## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

using Random
using Test

using EuNms


global β = 2.0


@testset "params" begin
    @testset "getFPparam" begin
        @test getFPparam(Float64) == (β, 53, (-1022, 1023))
        @test getFPparam(Float32) == (β, 24, (-126, 127))
        @test getFPparam(Float16) == (β, 11, (-14, 15))
        @isdefined(Float128) && @test getFPparam(Float128) == (β, 113, (-16382, 16383))
        @isdefined(BFloat16) && @test getFPparam(BFloat16) == (β,   8, (-126, 127))
    end
    @testset "getFPrdoff" begin
        @test getFPrdoff(β, 53) == β^-53
        @test getFPrdoff(β, 24) == β^-24
        @test getFPrdoff(β, 11) == β^-11
        @isdefined(Float128) && @test getFPrdoff(β, 113) == β^-113
        @isdefined(BFloat16) && @test getFPrdoff(β, 8) == β^-8
    end
    @testset "getENparam" begin
        @test getENparam_muller(β, 53, (-1022, 1023)) == (-484, 485)
        @test getENparam_muller(β, 24, (-126, 127)) == (-51, 51)
        @test getENparam_muller(β, 113, (-16382, 16383)) == (-8134, 8135)
        @test getENparam_graillat(β, 53, (-1022, 1023)) == (-376, 324)
        @test getENparam_graillat(β, 24, (-126, 127)) == (-56, 36)
        @test getENparam_graillat(β, 113, (-16382, 16383)) == (-5536, 5424)
    end
end


Random.seed!(1234)
@testset "$(repr(A)): exactness" for (A, r) in [(Fast2Sum, +),
                                                (TwoSum, +),
                                                (Fast2Mult, *)]

    @testset "Using type $F"  for F in [[Float64, Float32, Float16, BigFloat]
                                        @isdefined(Float128) ? [Float128] : []
                                        @isdefined(BFloat16) ? [BFloat16] : []]
        for (x, y) in zip(rand(F, 100), rand(F, 100))
            @test A(x, y).h == r(x, y)
        end
    end
end


@testset "diagnostics" begin

    u = eps(1.0) / 2
    @test is_correct(1.0+4u, big(1.0)+3u) && !is_correct(1.0+2u, big(1.0)+3u)
    @test is_faithful(1.0+4u, big(1.0)+3u) && is_faithful(1.0+2u, big(1.0)+3u)
end
