## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

using Test

using EuNms


@testset "checkDW" begin

    @test checkDW(0x1p0, 0x0p0)

    ## Assume Round to Nearest, ties to even.

    ## exponent(xl) = exponent(xh) - precision(xh) - 1 is always good.
    @test checkDW(0x1.0101010101010p0, 0x1.0101010101010p-54)
    @test checkDW(0x1.0101010101010p0,-0x1.0101010101010p-54)

    ## except when xh is a power of two
    @test checkDW(0x1p0, 0x1.1p-54)
    @test checkDW(0x1p0,-0x1.1p-54) == false
    
    ## Tie-breaking cases
    @test checkDW(0x1.1010101010101p0, 0x1.0p-53) == false  ## Odd last digit
    @test checkDW(0x1.1010101010101p0,-0x1.0p-53) == false
    @test checkDW(0x1.0101010101010p0, 0x1.0p-53)           ## Even last digit
    @test checkDW(0x1.0101010101010p0,-0x1.0p-53)
end


@testset "isequal bugfix" begin

    @test isequal(DW(big(1.0), big(2.0e-80)), DW(big(1.0), big(2.0e-80)))
    @test DW(big(1.0), big(2.0e-80)) == DW(big(1.0), big(2.0e-80))
end


@testset "DWtoBig" begin

    u = eps(Float64) / 2
    U = convert(BigFloat, u)
    u⁰, u² = u^0, u^2
    U⁰, U² = U^0, U^2

    @test big(DW(u⁰+2u, -u+u²)) == U⁰ + 2U - U + U²
end


@testset "conversions" begin
    @test_nowarn convert(DW{Float64}, 1.25)
    @test_nowarn convert(DW{Float64}, 1)
    @test_nowarn convert(DW{Float64}, (1.25, 0.0675e-16))
    @test_nowarn convert(Tuple{Float64,Float64}, DW(1.25, 0.0675e-16))

    @test_nowarn convert(BigFloat, DW(1.25, 0.0675e-16))
    @test_nowarn convert(Array{BigFloat}, [DW(1.25, 0.0675e-16), DW(1.25, 0.0675e-16)])
end
