## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

using Printf
using Random
using Test

using EuNms

const FP_TYPES = [[Float64, Float32, Float16, BigFloat]
                  @isdefined(Float128) ? [Float128] : []]


#######################################################################
## Check random exponents.
#######################################################################


function _collect(nbin, emax, xs)
    bins = zeros(Int, nbin+2)

    index1 = e -> min(max(1, e), size(bins, 1))

    for x in xs
        e = iszero(x) ? exponent(floatmin(x)) : exponent(x)
        bins[index1(2 + emax - e)] += 1
    end

    bins
end


function print_bins(seed, nstat, refs, bins)

    printstyled(@sprintf " nstat=%d seed=%d\n" nstat seed)

    printstyled(@sprintf " %8s |" "Expected"; color=:yellow)
    printstyled(@sprintf "| %4d |" refs[1]; color=:yellow)
    for r in refs[2:end-1]; printstyled(@sprintf " %4d |" r; color=:yellow); end
    printstyled(@sprintf "| %4d |" refs[end]; color=:yellow)
    @printf "\n"

    for (F, vals) in bins
        @printf " %8s |" F
        printstyled(@sprintf "| %4d |" vals[1]; color=:light_black)
        for v in vals[2:end-1] @printf " %4d |" v; end
        printstyled(@sprintf "| %4d |" vals[end]; color=:light_black)
        @printf "\n"
    end
end


function test_std_rand(seed, nstat, nbins)

    refs = vcat([0], [div(nstat, 2^i) for i in 1:nbins+1])

    bins = Dict{Type,Array{Int}}()
    for F in FP_TYPES
        Random.seed!(seed)
        bins[F] = _collect(nbins, -1, rand(F, nstat))
    end

    print_bins(seed, nstat, refs, bins)
end


function test_hyp_rand(seed, nstat, nbins)

    refs = vcat([0], [div(nstat, nbins) for _ in 1:nbins], [0])

    bins = Dict{Type,Array{Int}}()
    for F in FP_TYPES
        Random.seed!(seed)
        bins[F] = _collect(nbins, -1, hyprand(F, (-nbins, 0), nstat))
    end

    print_bins(seed, nstat, refs, bins)
end


#######################################################################
## Double Words Tests
#######################################################################


function _collect_dw(nbin, emax, xs)
    bins = zeros(Int, nbin+2, nbin+2)

    index1 = e -> min(max(1, e), size(bins, 1))
    index2 = e -> min(max(1, e), size(bins, 2))

    for x in xs
        eh = iszero(x.h) ? exponent(floatmin(x.h)) : exponent(x.h)
        el = iszero(x.l) ? exponent(floatmin(x.l)) : exponent(x.l)
        bins[index1(2 + emax - eh), index2(1 - precision(x.h) - (el - eh))] += 1
    end

    bins
end


function print_bins_dw(seed, nstat, nbins, emax, bins)

    printstyled(@sprintf " nstat=%d seed=%d\n" nstat seed)

    for (F, data) in bins

        printstyled(@sprintf " %8s |" F; color=:yellow)
        for il in 1:nbins+1; printstyled(@sprintf " %4d |" 1-precision(F)-il; color=:yellow); end
        @printf "\n"

        printstyled(@sprintf "         |"; color=:yellow)
        printstyled(@sprintf "| %4d |" data[1,1]; color=:light_black)
        for v in data[1,2:end-1] printstyled(@sprintf " %4d |" v; color=:light_black); end
        printstyled(@sprintf "| %4d |" data[1,end]; color=:light_black)
            @printf "\n"
        for ih in 2:size(data,1)-1
            printstyled(@sprintf " eh=%4d |" 2+emax-ih; color=:yellow)
            printstyled(@sprintf "| %4d |" data[ih,1]; color=:light_black)
            for v in data[ih,2:end-1] @printf " %4d |" v; end
            printstyled(@sprintf "| %4d |" data[ih,end]; color=:light_black)
            @printf "\n"
        end
        printstyled(@sprintf "         |"; color=:yellow)
        printstyled(@sprintf "| %4d |" data[end,1]; color=:light_black)
        for v in data[end,2:end-1] printstyled(@sprintf " %4d |" v; color=:light_black); end
        printstyled(@sprintf "| %4d |" data[end,end]; color=:light_black)
        @printf "\n"
    end
end


function test_dw_rand(seed, nstat, nbins)
    TYPES = [[Float64, Float32, BigFloat]
             @isdefined(Float128) ? [Float128] : []]

    bins = Dict{Type,Array{Int}}()
    for F in TYPES
        Random.seed!(seed)
        bins[F] = _collect_dw(nbins, 0, rand(DW{F}, nstat))
    end

    print_bins_dw(seed, nstat, nbins, 0, bins)
end


function test_dwhyp_rand(seed, nstat, nbins)
    TYPES = [[Float64, Float32, BigFloat]
             @isdefined(Float128) ? [Float128] : []]

    bins = Dict{Type,Array{Int}}()
    for F in TYPES
        Random.seed!(seed)
        bins[F] = _collect_dw(nbins, -1, hyprand(DW{F}, (-nbins+1, 0), nstat))
    end

    print_bins_dw(seed, nstat, nbins, -1, bins)
end


#######################################################################
## Main.
#######################################################################


begin
    seed = get(ENV, "SEED", rand(UInt))

    printstyled("# Randomly-generated numbers groubed by their radix-2 magnitude between -16 and 0.\n"; color=:yellow)

    printstyled("[STANDARD RAND]"; color=:bold)
    test_std_rand(seed, 10000, 8)

    printstyled("[HYPERBOLIC RAND]"; color=:bold)
    test_hyp_rand(seed, 10000, 8)

    printstyled("[DOUBLE WORD RAND]"; color=:bold)
    test_dw_rand(seed, 10000, 8)

    printstyled("[DOUBLE WORD HYPERBOLIC RAND]"; color=:bold)
    test_dwhyp_rand(seed, 100000, 8)
end
