## Author: Joris Picot,
##         École Normale Supérieure de Lyon, LIP, AriC Team.
##
## This program is part of an implementation of the algorithm described in the
## paper:
##
##     Vincent Lefèvre, Nicolas Louvet, Jean-Michel Muller, Joris Picot,
##     Laurence Rideau. Accurate calculation of Euclidean Norms using
##     Double-word arithmetic. 2021
##
## Copyright (C) 2021  École Normale Supérieure de Lyon
##
## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
## USA

using Logging
using Random
using Test

using EuNms
include("accuracy/accuracy_EuNm.jl")


function scalar(F, nstat, seed, unit)

    setprecision(BigFloat, 256)
    Random.seed!(seed)

    (_, p, (emin, _)) = getFPparam(F)
    η = round(Int, (emin + p) / 2, RoundUp)

    u = Tables.getUnit(unit, F)

    Tables.title("SCALAR EVIDENCE"; fmt=F, size=2^size, nstat=nstat, seed=seed, unit=unit)
    Tables.head(("exp", 4))
    Tables.vals([(n, 8) for n in ["min", "Q_1", "med", "Q_3", "max"]]...; color=:yellow)
    Tables.vals(("bound", 10); color=:yellow)
    println()

    for e in η+1:-1:η-p-2

        xs = hyprand(F, (e, e+1), nstat)
        s1 = Fast2Mult.(xs, xs)
        s0 = big.(xs) .^ 2
        εs = rel_error.(DWTtoBig.(s1), s0)
        bound = getFPrdoff(F) ^ 2

        stats = quantile(εs, [0.0, 0.25, 0.5, 0.75, 1.0])

        Tables.head((string(e), 4))
        Tables.vals([(x / u, 8) for x in stats]...)
        Tables.check((bound / u, 8), stats[end] ≤ bound)
        println()
    end
end


function sum_of_squares(Algo, Sum, Prm, F, size, nstat, seed, unit)

    setprecision(BigFloat, 256)
    Random.seed!(seed)

    (_, p, (emin, _)) = getFPparam(F)
    η = round(Int, (emin + p) / 2, RoundUp)

    u = Tables.getUnit(unit, F)

    Tables.title("SUM OF SQUARES EVIDENCE"; algo=Algo, sum=Sum, prm=Prm,
                 fmt=F, size=2^size, nstat=nstat, seed=seed, unit=unit)
    Tables.head(("exp", 4))
    Tables.vals([(n, 8) for n in ["min", "Q_1", "med", "Q_3", "max"]]...; color=:yellow)
    Tables.vals(("bound", 10); color=:yellow)
    println()

    εs = BigFloat[]

    for e in η+1:-1:η-p-2

        for _ in 1:nstat

            xs = hyprand(F, (e, e+1), 2^size)
            s1 = SqSum(Val(Algo), Val(Sum), ENParam(Val(Prm), F; validate=:no), xs)
            s0 = sum(big.(xs) .^ 2)
            ε = rel_error(DWTtoBig(s1), s0)
            append!(εs, ε)
        end

        bound = known_rel_bnds[(SqSum, Algo, Sum, Prm)](getFPrdoff(F), 2^size)

        stats = quantile(εs, [0.0, 0.25, 0.5, 0.75, 1.0])

        Tables.head((string(e), 4))
        Tables.vals([(x / u, 8) for x in stats]...)
        Tables.check((bound / u, 8), stats[end] ≤ bound)
        println()
    end
end


function euclidean_norm(Algo, Sum, Sq, Prm, F, size, nstat, seed, unit)

    setprecision(BigFloat, 256)
    Random.seed!(seed)

    (_, p, (emin, _)) = getFPparam(F)
    η = round(Int, (emin + p) / 2, RoundUp)

    u = Tables.getUnit(unit, F)

    Tables.title("EUCLIDEAN NORM EVIDENCE"; algo=Algo, sum=Sum, sq=Sq, prm=Prm,
                 fmt=F, size=2^size, nstat=nstat, seed=seed, unit=unit)
    Tables.head(("exp", 4))
    Tables.vals([(n, 8) for n in ["min", "Q_1", "med", "Q_3", "max"]]...; color=:yellow)
    Tables.vals(("bound", 10); color=:yellow)
    println()

    εs = BigFloat[]

    for e in η+1:-1:η-2p-2

        for _ in 1:nstat

            xs = hyprand(F, (e, e+1), 2^size)
            s1 = EuNm(Val(Algo), Val(Sum), Val(Sq), ENParam(Val(Prm), F; validate=:no), xs)
            s0 = sqrt(sum(big.(xs) .^ 2))
            ε = rel_error(big(s1), s0)
            append!(εs, ε)
        end

        bound = known_rel_bnds[(EuNm, Algo, Sum, Sq, Prm)](getFPrdoff(F), 2^size)

        stats = quantile(εs, [0.0, 0.25, 0.5, 0.75, 1.0])

        Tables.head((string(e), 4))
        Tables.vals([(x / u, 8) for x in stats]...)
        Tables.check((bound / u, 8), stats[end] ≤ bound)
        println()
    end
end



if abspath(PROGRAM_FILE) == @__FILE__

    F = Float32
    size = 6
    nstat = 5040
    seed = get(ENV, "SEED", 123456789)
    unit = :u²

    #scalar(F, nstat, seed, :u²)
    sum_of_squares(:noovf, :sloppy, :graill, F, size, nstat, seed, :u²)
    #sum_of_squares(:masks, :sloppy, :graill, F, size, nstat, seed, :u²)
    euclidean_norm(:onep1, :sloppy, :graill, :graill, F, size, nstat, seed, :ε)
end
